package com.tripgis.mobile;

import android.location.Location;
import android.os.Environment;
import android.util.Log;

import com.tripgis.mobile.data.LocationRecord;
import com.tripgis.mobile.data.TripRecord;
import com.google.android.gms.maps.model.LatLng;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.sql.Date;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.List;

/**
 * References:
 *    https://developer.android.com/guide/topics/data/data-storage.html#filesInternal
 *    https://docs.oracle.com/javase/7/docs/api/java/io/File.html
 *    https://www.mkyong.com/java/how-to-export-data-to-csv-file-java/
 *
 * Created by pk on 7/3/17.
 */

public class FileManager
{
    protected File file;

    // Name to use in log lines
    public static final String TAG = FileManager.class.getSimpleName();

    // Directory to store all filese used by application
    protected static final String BASE_DIR = "tripgis";

    public static final String FILE_EXT = ".csv";

    // Line separator to use for each stored record
    public static final String CSV_SEPARATOR = ",";

    // Text to replace line separators
    public static final String CSV_SUBSTITUTION = "<CMA>";

    /**
     * Prepares the specified for for use. If the file is empty, then it
     * will be created. A subsequent call to read() or write() will actually
     * open the file for editing/reading.
     *
     * NOTE: The file will be stored in DIRECTORY_DOCUMENTS/BASE_DIR.
     *
     * @param fileName - name of file to open
     */
    public FileManager( String fileName ) throws IOException
    {
        Log.d( TAG, "FileManager( " + fileName + " )" );
        setFile( fileName );
    }

    private String getFileName( String name ) { return name + FILE_EXT; }

    protected void setFile( String name ) throws IOException
    {
        Log.d( TAG, "setFile( " + name + " )" );
        file = new File( getAlbumStorageDir().getAbsolutePath(), getFileName( name ) );

        Log.d( TAG, "buildEmptyFile()" );
        // createNewFile() only creates a file if it does not exist
        //    and we don't care if the file existed already at this point.
        if( file.createNewFile() )
            Log.i( TAG, "Created " + file );

//        if( file.length() == 0 )
//        {
//            ArrayList<LocationRecord> l = new ArrayList<>();
//            LocationRecord r = new LocationRecord( LocationRecord.getCurrentTime(), new LatLng(32, 96), "test", null );
//            l.add(r);
//            write(l, true);
//        }

        Log.i( TAG, "Current storage file: " + file + " [" + prettySize( file.length() ) + "]" );
    }

    /**
     * Attempts to open the DIRECTORY_DOCUMENTS public storage area. Creates
     * a directory under it called BASE_DIR.
     *
     * @return a File object containing the absolute path of the directory.
     */
    public static File getAlbumStorageDir() throws IOException
    {
        Log.d( TAG, "getAlbumStorageDir()" );
        File path = new File( Environment.getExternalStoragePublicDirectory( Environment.DIRECTORY_DOCUMENTS ), BASE_DIR );

        if( !path.exists() )        // No need to create the path if it already exists
        {
            if (!path.mkdirs())     // Create all necessary directories (equiv. to mkdir -p)
                throw new IOException("Directory '" + path + "' not created");
        }

        // Return the desired path of the stored file
        Log.i( TAG, "Device media stoage state: " + Environment.getExternalStorageState() );
        Log.d( TAG, "Total: " + prettySize( path.getTotalSpace() ) +
                ", Free: " + prettySize( path.getFreeSpace() ) +
                ", Usable: " + prettySize( path.getUsableSpace() ) );
        return path;
    }

    public static String prettySize( long bytes )
    {
        Log.d( TAG, "prettySize( " + bytes + " )" );
        if( bytes > Math.pow( 1024, 4 ) )
            return String.format( java.util.Locale.US, "%.2f Tb", bytes / Math.pow( 1024, 4 ) );
        else if( bytes > Math.pow( 1024, 3 ) )
            return String.format( java.util.Locale.US, "%.2f Gb", bytes / Math.pow( 1024, 3 ) );
        else if( bytes > Math.pow( 1024, 2 ) )
            return String.format( java.util.Locale.US, "%.2f Mb", bytes / Math.pow( 1024, 2 ) );
        else if( bytes > Math.pow( 1024, 1 ) )
            return String.format( java.util.Locale.US, "%.2f Kb", bytes / Math.pow( 1024, 1 ) );
        else
            return String.format( java.util.Locale.US, "%d bytes", bytes );
    }

    /**
     * Reads the stored file data, parses it, and returns a list of Record found.
     *
     * @return ArrayList of LocationRecord objects, or null if no data or error.
     */
    public ArrayList<LocationRecord> read() throws IOException
    {
        Log.d( TAG, "read()" );

        try
        {
            // Prepare the file for reading
            FileReader fileReader = new FileReader( file );
            BufferedReader lineReader = new BufferedReader( fileReader );

            ArrayList<LocationRecord> list = read( lineReader );
            if( list != null )
                return list;
        }
        catch( FileNotFoundException e )
        {
            throw new IOException( "Failed to read file '" + file + "': " + e );
        }

        // Return null if there was an error reading
        return null;
    }

    /**
     *
     * References:
     *    https://www.mkyong.com/java/how-to-read-and-parse-csv-file-in-java/
     *
     * @param lineReader
     * @return
     */
    protected ArrayList<LocationRecord> read( BufferedReader lineReader ) throws IOException
    {
        Log.d( TAG, "read( <BufferedReader> )" );

        // Initialize the list to return if data is present
        ArrayList<LocationRecord> list = new ArrayList<>();

        // Read in the first line from the file
        String line = lineReader.readLine();

        // Iterate through the data until an empty record is found (EOF)
        while( line != null )
        {
            // Pass the parsed line to LocationRecord to have it properly assigned to object
            LocationRecord record = new LocationRecord( list.size() + 1, line.split( CSV_SEPARATOR ) );
            list.add( record );                 // Add the record to the list
            line = lineReader.readLine();       // Read the next line
        }

        // Return the list if all goes well
        Log.i( TAG, "Read " + list.size() + " lines from '" + file + "'" );
        return list;
    }

    /**
     * Opens the stored file for writing and adds a new line for each record.
     * Data can be appended or the file can be overwritten each time.
     *
     * @param list the list of data to write
     * @param overwrite boolean to indicate existing data should be overwritten
     * @return the number of lines written to the file
     */
    public int write( ArrayList<LocationRecord> list, boolean overwrite ) throws IOException
    {
        Log.d( TAG, "write( " + list.size() + ", " + overwrite + " )" );
        int lineCount = 0;

        FileWriter writer = new FileWriter( file );   // Writer object
        lineCount = write( writer, list );

        writer.flush();
        writer.close();

        // Return the number of lines written
        Log.i( TAG, "Wrote " + lineCount + " lines to '" + file + "'" );
        return lineCount;
    }

    /**
     * References:
     *    https://stackoverflow.com/questions/3585053/in-java-is-it-possible-to-check-if-a-string-is-only-ascii
     *
     * @param writer
     * @param list
     * @return
     * @throws IOException
     */
    protected int write( FileWriter writer, ArrayList<LocationRecord> list ) throws IOException
    {
        int lineCount = 0;      // Number of lines written

        // Iterate through each record in the passed list
        for( LocationRecord record : list )
        {
            lineCount++;                                   // Increment line count (for logging)
            if( lineCount < 5 )
                Log.d( TAG, "Writing " + record );
            String text = record.toString();
            if( !text.matches( "\\A\\p{ASCII}*\\z" ) )
                throw new IOException( "Line " + lineCount + ": Record contains non-ASCII - '" + text + "'" );

            writer.append( record.toString() + "\n" );  // Write out the data
        }

        return lineCount;
    }

    public static String writeString( String text )
    {
        return text.replaceAll( CSV_SEPARATOR, CSV_SUBSTITUTION );
    }

    public static String readString( String text )
    {
        return text.replaceAll( CSV_SUBSTITUTION, CSV_SEPARATOR );
    }

    /**
     * Deletes the file from the filesystem. Unless requested, a new empty file
     * will be attempted to be created.
     *
     * @param dontKeep if set, no new file will be created
     * @return the results of the delete action (true is good)
     */
    public boolean delete( boolean dontKeep )
    {
        // Attempt to delete the file
        boolean rc = file.delete();

        // No need to continue processing if delete failed
        if( rc )
        {
            Log.i( TAG, "Successfully removed '" + file + "'" );

            try
            {
                // Attempt to create a blank file at this point
                // NOTE: This mimics the behavior of the constructor. However,
                //   It is not necessary to verify success. Any future actions
                //   will fail with appropriate notification.
                if( !dontKeep )
                    file.createNewFile();
            }
            catch ( IOException e )
            {
                Log.e( TAG, "Failed to create empty file '" + file + "'" );
            }
        }
        else
            Log.e( TAG, "Failed to delete file '" + file + "'" );

        // Return the results of the delete operation
        return rc;
    }
    public boolean delete() { return delete( false ); }     // Shortcut to default action

    //TODO
    // FUTURE: Check disk space, add to summary - getUsedSpace(), getAvailableSpace()
    //   Add Picture capability (when API figured out)

    //TODO (Connection)
    // Add writeLine to MapsActivity.SaveLocation feature
    // FUTURE: If Overlay selected, add read call to get map_items
    //  If Create Trip selected, add read call to allow selections to port
    //  If Transfer Locations selected, perform same function
    //  If Reset Saved Locations, add clear() - default file only
    //  If Move Locations, add moveData( newFile, ArrayList<Record> points )
    //          -create new FileManager, write points, add removeLine( Record )
    //  If Delete Trip selected, add removeFile() - trip files only
    //  If Copy Trip selected...well, this is very far ahead
    //  Add sort feature - read all lines, sort, write all lines

    /* Checks if external storage is available for read and write */
    public static boolean isExternalStorageWritable()
    {
        String state = Environment.getExternalStorageState();
        if( Environment.MEDIA_MOUNTED.equals( state ) )
            return true;

        return false;
    }

    /* Checks if external storage is available to at least read */
    public static boolean isExternalStorageReadable()
    {
        String state = Environment.getExternalStorageState();
        if( Environment.MEDIA_MOUNTED.equals( state ) ||
                Environment.MEDIA_MOUNTED_READ_ONLY.equals( state ) )
            return true;

        return false;
    }
}
