package com.tripgis.mobile;

import android.util.Log;

import com.tripgis.mobile.data.LocationRecord;
import com.tripgis.mobile.data.TripRecord;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.lang.reflect.Array;
import java.net.FileNameMap;
import java.util.ArrayList;

/**
 * References:
 *    https://developer.android.com/guide/topics/data/data-storage.html#filesInternal
 *    https://docs.oracle.com/javase/7/docs/api/java/io/File.html
 *    https://www.mkyong.com/java/how-to-export-data-to-csv-file-java/
 *
 * Created by pk on 7/3/17.
 */

public class TripFileManager extends FileManager
{
    private TripRecord trip;

    // Name to use in log lines
    public static final String TAG = TripFileManager.class.getSimpleName();

    // Name of the default file used to store data
    public static final String DEFAULT_TRIP_NAME = "TripGIS";

    private static TripFileManager DEFAULT_TRIP = null;

    // "Semaphore" to ensure that default trip is built only once
    private static boolean initializingDefault = false;

    public TripFileManager() throws IOException { this( DEFAULT_TRIP_NAME ); };
    public TripFileManager( String tripName ) throws IOException
    {
        super( tripName );
        Log.d( TAG, "TripFileManager( " + tripName + " )" );

        if( DEFAULT_TRIP == null && !initializingDefault )
        {
            Log.w( TAG, "Generating default trip.." );
            initializingDefault = true;
            DEFAULT_TRIP = new TripFileManager( DEFAULT_TRIP_NAME );
            initializingDefault = false;
        }

        if( tripName.equals( DEFAULT_TRIP_NAME ) && !initializingDefault )
        {
            Log.d( TAG, "Using default trip " + DEFAULT_TRIP.getTrip() );
            this.trip = DEFAULT_TRIP.getTrip();
            trip.setSavedLocations( DEFAULT_TRIP.getSavedLocations() );
        }
        else
        {
            Log.w( TAG, "Building new trip " + tripName );
            // Build a default trip with this information -
            //   It will be overwritten by the data file later
            trip = new TripRecord( tripName );
            populateNewTripFile();

            ArrayList<LocationRecord> list = read();
            Log.i( TAG, "Created Trip object for '" + tripName + "', NumRecords: " + list.size() );
            trip.setSavedLocations( list );
        }
    }

/*    public TripFileManager( TripRecord trip ) throws IOException
    {
        super( trip.getName() + ".csv" );
        Log.d( TAG, "TripFileManager( " + trip + " )" );

        // Copy any additional trip data at this point
        // The name and savedLocations will be obtained from
        this.trip = trip;

        if( file.length() == 0 )
            throw new IOException( "Opened empty file for '" + trip + "'" );
    } */

    /**
     * References:
     *    https://stackoverflow.com/questions/8646984/how-to-list-files-in-an-android-directory
     *    https://stackoverflow.com/questions/924394/how-to-get-the-filename-without-the-extension-in-java
     *
     * @return
     * @throws IOException
     */
    public static ArrayList<String> getAvailableTrips() throws IOException
    {
        Log.d( TAG, "getAvailableTrips()" );
        File path = TripFileManager.getAlbumStorageDir();
        Log.d( TAG, "Reading " + path + "..." );

        ArrayList<String> trips = new ArrayList<>();
        File[] files = path.listFiles();
        if( files != null )
        {
            Log.d( TAG, "Number of files: " + files.length );
            for( int i = 0; i < files.length; i++ )
            {
                String name = files[i].getName();
                if( !name.contains( TripFileManager.DEFAULT_TRIP_NAME ) &&
                        name.contains( FileManager.FILE_EXT ) )
                {
                    Log.d( TAG, "Adding File '" + files[i].getName() + "'" );
                    trips.add( name.replaceFirst("[.][^.]+$", "" ) );
                }
            }
        }

        return trips;
    }

    /**
     * Reads the stored file data, parses it, and returns a list of Record found.
     *
     * References:
     *    https://www.mkyong.com/java/how-to-read-and-parse-csv-file-in-java/
     *
     * @return ArrayList of LocationRecord objects, or null if no data or error.
     */
    public ArrayList<LocationRecord> read() throws IOException
    {
        Log.d( TAG, "read()" );

        try
        {
            // Prepare the file for reading
            Log.d( TAG, "Trip " + trip + " - reading " + file + "..." );
            FileReader fileReader = new FileReader( file );
            BufferedReader lineReader = new BufferedReader( fileReader );

            // Read in the first line from the file
            String line = lineReader.readLine();

            // It is assumed the first line is a TripRecord line, so just treat it as such
            String[] data = line.split( FileManager.CSV_SEPARATOR );
            Log.d( TAG, "Creating trip record from " + line + " ( " + data[0] + " )" );
            trip = new TripRecord( data[0] );

            // Initialize the list to return if data is present
            trip.setSavedLocations( read( lineReader ) );
            return trip.getSavedLocations();
        }
        catch( FileNotFoundException e )
        {
            throw new IOException( "Failed to read file '" + file + "': " + e );
        }
    }

    public TripRecord getTrip()           { return trip; }
    public void setTrip( TripRecord t ) throws IOException
    {
        Log.d( TAG, "setTrip( " + t + " )" );

        if( t == null )
        {
            trip = DEFAULT_TRIP.getTrip();
            Log.d( TAG, "Resetting to default - " + trip );
        }
        else
            trip = t;

        setFile( trip.getName() );
        populateNewTripFile();
        trip.setSavedLocations( read() );
    }

    public ArrayList<LocationRecord> getSavedLocations()
    {
        Log.d( TAG, "getSavedLocations()" );
        return trip.getSavedLocations();
    }

    public ArrayList<LocationRecord> addSavedLocation( LocationRecord record ) throws IOException
    {
        Log.d( TAG, "addSavedLocation( " + record + " )" );
        int init = trip.getItemCount();
        int writeCount = write( record );

        // As long as this is not a multi-threaded app, there's little reason
        // not to just update the list. A lot of I/O goes on if we write
        // then read
        if( init == writeCount - 1 )
        {
            trip.getSavedLocations().add( record );
            return trip.getSavedLocations();
        }

        throw new IOException( "Failed to save " + record + " - wrote " + writeCount +
                ", expected " + init + "!" );
    }

    public ArrayList<LocationRecord> updateSavedLocations( ArrayList<LocationRecord> list ) throws IOException
    {
        Log.d( TAG, "updatedSavedLocations( " + list.size() + " )" );

        int writeCount = write( list, true );
        if( writeCount != list.size() )
            throw new IOException( "Failed to update locations - wrote " + writeCount +
                    ", expected " + list.size() + "!" );

        if( trip.getItemCount() == 0 )
            Log.d( TAG, "Empty List" );
        else
        {
            Log.d( TAG, "First Item: " + trip.getFirstRecord() );
            Log.d( TAG, "Last Item: " + trip.getLastRecord() );
        }
        return trip.getSavedLocations();
    }

    public ArrayList<LocationRecord> refreshSavedLocations() throws IOException
    {
        trip.setSavedLocations( read() );
        if( trip.getItemCount() == 0 )
            Log.d( TAG, "Empty List" );
        else
        {
            Log.d( TAG, "First Item: " + trip.getFirstRecord() );
            Log.d( TAG, "Last Item: " + trip.getLastRecord() );
        }
        return trip.getSavedLocations();
    }

    /**
     * Opens the stored file for writing and adds a new line for each record.
     * Data can be appended or the file can be overwritten each time.
     *
     * @param list the list of data to write
     * @param overwrite boolean to indicate existing data should be overwritten
     * @return the number of lines written to the file
     */
    public int write( ArrayList<LocationRecord> list, boolean overwrite ) throws IOException
    {
        Log.d( TAG, "write( " + list.size() + ", " + overwrite + " )" );
        int lineCount = 0;      // Number of lines written

        if( overwrite )
            trip.setSavedLocations( list );
        else
            trip.getSavedLocations().addAll( list );

        FileWriter writer = new FileWriter( file );   // Writer object
        lineCount = 1;      //

        String text = trip.toString();
        if( !text.matches( "\\A\\p{ASCII}*\\z" ) )
            throw new IOException( " Trip Record data contains non-ASCII - '" + text + "'" );

        writer.append( text + "\n" );

        lineCount += super.write( writer, trip.getSavedLocations() );
        lineCount--;

        writer.close();

        // Return the number of lines written
        Log.i( TAG, "Wrote " + lineCount + " lines to '" + file + "'" );
        return lineCount;
    }

    protected void populateNewTripFile() throws IOException
    {
        Log.d( TAG, "populateNewTripFile()" );

//            if (file.length() == 0)
//                write(null, true);

        Log.d( TAG, "File Size: " + prettySize( file.length() ) );
        if( file.length() == 0 )
        {
            FileWriter writer = new FileWriter(file);   // Writer object
            String text = trip.toString();
            if( !text.matches( "\\A\\p{ASCII}*\\z" ) )
                throw new IOException( " Trip Record data contains non-ASCII - '" + text + "'" );

            writer.append( text + "\n");
            writer.close();
            Log.d(TAG, "wrote " + trip + " to file " + file);
        }
    }

    /**
     * Write a single record to the stored location.
     *
     * @param record the record to save
     * @return 1 if the record was written or 0 if failed.
     */
    public int write( LocationRecord record ) throws IOException
    {
        // Create a temporary list and add the record
        ArrayList<LocationRecord> list = new ArrayList<>();
        list.add( record );

        // Pass the temp list to write() and specify to append
        return write( list, false );
    }
}
