package com.tripgis.mobile.activity;

import android.annotation.TargetApi;
import android.app.Activity;
import android.content.Context;
import android.content.Intent;
import android.os.Bundle;
import android.support.v7.app.AppCompatActivity;
import android.support.v7.widget.Toolbar;
import android.text.Editable;
import android.text.TextWatcher;
import android.util.Log;
import android.view.KeyEvent;
import android.view.Menu;
import android.view.MenuItem;
import android.view.View;
import android.view.inputmethod.InputMethodManager;
import android.widget.Button;
import android.widget.EditText;
import android.widget.TextView;

import com.tripgis.mobile.TripGisApplication;
import com.tripgis.mobile.R;
import com.tripgis.mobile.data.LocationRecord;
import com.tripgis.mobile.data.TripRecord;

import java.io.IOException;

// Added     compile 'com.google.android.gms:play-services-location:11.0.2'

/**
 * This class handles all interactions with the Save Location Form. When
 * the user selects Save Location from MapActivity, data is passed to this
 * class to be processed. Once complete, processing is returned to the main
 * program.
 */
public class TripActivity extends AppCompatActivity implements View.OnClickListener,
        TextWatcher, TextView.OnEditorActionListener
{
    private EditText name;              // Name textfield
    private TextView startTime;         // Start Time textfield (read-only)
    private TextView endTime;           // End Time textfield (read-only)
    private TextView itemCount;         // Item Count textfield (read-only)
    private Button saveButton;          // Save button
    private Button cancelButton;        // Cancel button

    // Copy of data to be received/sent
    private TripRecord record;

    private Toolbar toolbar;
    private Menu menu;
    private TripGisApplication application;

    // Simplified name to use in log messages and data messaging
    public static final String TAG = TripActivity.class.getSimpleName();

    /**
     * Initialization function for when app is first created. Used to set up
     * API calls and basic save_location. Actual interactions and data presentation
     * will be covered elsewhere. Refer to onResume and onPause.
     *
     * Overridden from Activity.
     *
     * @param savedInstanceState data passed from parent activity.
     */
    @Override
    public void onCreate( Bundle savedInstanceState )
    {
        Log.d( TAG, "onCreate( <Bundle> )" );
        super.onCreate(savedInstanceState);         // Call parent for default behavior
        setContentView(R.layout.trip_form);         // Access the associated layouts XML

        application = (TripGisApplication) getApplication();

        // Read the passed LocationRecord object
        record = this.getIntent().getParcelableExtra( TripRecord.TAG );
        Log.i( TAG, "Received " + record + " from " + MapsActivity.TAG );

        // NOTE: Keeping the parcelable syntax for future upgrades
        //   For now, just use what's stored in the application
        if( record != null )
            record = application.getTripRecord();

        Log.d( TAG, "Using " + record );

        // Initialize private variables for various "dynamic" content
        name = (EditText) findViewById( R.id.trip_name );
        startTime = (EditText) findViewById( R.id.trip_start_time );
        endTime = (EditText) findViewById( R.id.trip_end_time );
        itemCount = (EditText) findViewById( R.id.trip_item_count );
        Log.d( TAG, "itemCount[" + R.id.trip_item_count + "]: " + itemCount );

        name.setOnEditorActionListener( this );
        name.addTextChangedListener( this );
        saveButton = (Button) findViewById( R.id.save_button );
        cancelButton = (Button) findViewById( R.id.cancel_button );

        // Define handlers for user-interaction
        saveButton.setOnClickListener( this );
        cancelButton.setOnClickListener( this );

        //https://stackoverflow.com/questions/8991522/how-can-i-set-the-focus-and-display-the-keyboard-on-my-edittext-programmatical
        name.requestFocus();
        name.setSelectAllOnFocus( true );           // Highlight field at outset
        InputMethodManager imm = (InputMethodManager) getSystemService( Context.INPUT_METHOD_SERVICE );
        imm.showSoftInput( name, InputMethodManager.SHOW_IMPLICIT );

        /**********************************************************************
         * Prepare the Toolbar
         ******************************************************************** */
        // Toolbar is the main title on top of the activity
        toolbar = (Toolbar) findViewById( R.id.trip_toolbar );
        setSupportActionBar( toolbar );
        toolbar.showOverflowMenu();
        getSupportActionBar().setDisplayHomeAsUpEnabled( true );
    }

    /**
     * Called when the application is made active (after onCreate on startup).
     * Online documentation suggests that onCreate is not the appropriate
     * place to modify the default values for text. It's possible this is
     * no longer necessary, but better safe than sorry.
     *
     * References:
     *    https://stackoverflow.com/questions/4590957/how-to-set-text-in-an-edittext
     *    https://stackoverflow.com/questions/3399774/error-no-package-identifier-when-getting-value-for-resource-number
     *
     * Overridden from Activity.
     */
    @Override
    protected void onResume()
    {
        Log.d( TAG, "onResume()" );
        super.onResume();                   // Call parent for default behavior

        // Set the textfields in the form to the values received by the parent
        if( record != null )
        {
            Log.d( TAG, "Populating fields with " + record );
            name.setText( record.getName() );
            startTime.setText( record.getStartTime() );
            endTime.setText( record.getEndTime() );
            itemCount.setText( Integer.toString( record.getItemCount() ) );
        }
        else
        {
            Log.d( TAG, "Creating new trip..." );
            name.setText( "" );
        }
    }


    /**
     * Setup the menu. This is called once during starting of the application.
     *
     * Overridden from Activity.
     *
     * @param menu the menu and its map_items from the /res/menu/*.xml file
     * @return always true
     */
    @Override
    public boolean onCreateOptionsMenu( Menu menu )
    {
        // Define the menu and its map_items
        Log.d( TAG, "onCreateOptionsMenu( <Menu> )" );
        getMenuInflater().inflate( R.menu.trip_items, menu );

        // It is recommended to save an internal copy of the menu for editing if needed
        this.menu = menu;
        return true;
    }

    /**
     * Handler for user interaction with form.
     *
     * Overridden from View.OnClickListener
     *
     * @param view
     */
    @Override
    public void onClick( View view )
    {
        Log.d( TAG, "onClick( " + view + " )" );

        // Determine which button was pressed
        switch( view.getId() )
        {
            case R.id.save_button:                      // Save Button
                Log.d( TAG, "save_button" );
                if( saveTrip() )                        // Collect saved data
                    finish();                           // Close the Activity
                return;                                 // Always exit a case

            case R.id.cancel_button:                    // Cancel Button
                Log.i( TAG, "cancel_button" );
                setResult( Activity.RESULT_CANCELED );  // Indicate cancel
                finish();                               // Close the Activity
                return;                                 // Always exit a case

            default:
                // If we got here, the user's action was not recognized.
                // Invoke the superclass to handle it.
                Log.e( TAG, "Unrecognized action - " + view.getId() );

        }
    }

    public boolean saveTrip()
    {
        Log.d( TAG, "saveTrip()" );

        try
        {
            // Create a new intent object to pass back to the parent
            Intent resultIntent = new Intent();

            // Ensure a name is specified
            if( name.getText().toString().isEmpty() )
            {
                application.setError( TAG, "Must specify a name to save!" );
                return false;
            }

            // Update data from form fields
            record = new TripRecord( name.getText().toString() );
            application.setTripRecord( record );

            // Add the updated record data
            resultIntent.putExtra( LocationRecord.TAG, record );
            setResult( Activity.RESULT_OK, resultIntent );      // Indicate success
            return true;
        }
        catch( IOException e )
        {
            application.setError( TAG, "Failed to save trip - " + e );
            return false;
        }

    }

    /**
     * Handle user-selected action from the menu items (/res/menu/<file>.xml).
     *
     * Overridden from Activity.
     *
     * References:
     *    https://developer.android.com/training/implementing-navigation/ancestral.html
     *    https://developer.android.com/training/implementing-navigation/nav-drawer.html
     *    https://stackoverflow.com/questions/21805382/how-to-resume-activity-instead-of-restart-when-going-up-from-action-bar
     *
     * @param item the selected item
     * @return true if the action is processed
     */
    @Override
    public boolean onOptionsItemSelected( MenuItem item )
    {
        Log.d( TAG, "onOptionsItemSelected( " + item + " )" );
        switch (item.getItemId())
        {
            // Respond to the action bar's Up/Home button
            case android.R.id.home:
//                NavUtils.navigateUpFromSameTask( this );
                returnFromStack();
                return true;

            case R.id.action_save:
                Log.d( TAG, "action_save" );
                if( saveTrip() )                        // Collect saved data
                    finish();                           // Close the Activity
                return true;                            // Always exit a case

            case R.id.action_cancel:
                Log.d( TAG, "action_cancel" );
                setResult( Activity.RESULT_CANCELED );  // Indicate cancel
                finish();                               // Close the Activity
                return true;                            // Always exit a case

            default:
                // If we got here, the user's action was not recognized.
                // Invoke the superclass to handle it.
                Log.e( TAG, "Unrecognized action - " + item.getItemId() );
                return super.onOptionsItemSelected( item );
        }
    }

    @TargetApi(16)
    private void returnFromStack()
    {
        navigateUpTo( getIntent() );
    }

    /**
     * Overridden from TextWatcher.
     *
     * References:
     *    https://developer.android.com/reference/android/text/TextWatcher.html
     *
     * @param s
     * @param start
     * @param before
     * @param count
     */
    @Override
    public void onTextChanged( CharSequence s, int start, int before, int count )
    {
        // Determine if enter was pressed

    }

    /**
     * Overridden from TextWatcher.
     *
     * @param s
     * @param start
     * @param count
     * @param after
     */
    @Override
    public void beforeTextChanged( CharSequence s, int start, int count, int after )
    {
    }

    /**
     * Overridden from TextWatcher.
     *
     * @param s
     */
    @Override
    public void afterTextChanged( Editable s )
    {
    }

    /**
     * Overridden from OnEditorActionListener.
     *
     * @param textView
     * @param i
     * @param keyEvent
     * @return
     */
    @Override
    public boolean onEditorAction( TextView textView, int i, KeyEvent keyEvent )
    {
        Log.d( TAG, "onEditorAction( " + textView.getImeActionLabel() + ", " + i + ", " + keyEvent.getKeyCode() );

        if( keyEvent.getKeyCode() == KeyEvent.KEYCODE_ENTER )
        {
            saveTrip();                         // Collect saved data
        }

        return false;
    }
}
