package com.tripgis.mobile.data;

import android.os.Parcel;
import android.os.Parcelable;
import android.util.Log;

import com.tripgis.mobile.FileManager;
import com.google.android.gms.maps.model.LatLng;

import java.io.File;
import java.io.IOException;
import java.sql.Date;
import java.sql.Time;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.TimeZone;

/**
 * LocationRecord is the data of import being collected by the Android App.
 *
 * Created by pk on 7/9/17.
 */

public class LocationRecord implements Parcelable
{
    private Date date;          // Date/Time of entry
    private LatLng location;    // Lat/Long of position
    private String name;        // Title if provided
    private Date endTime;       // End time of activity

    //TODO: Add additional map_items as they become known

    // Name to use in log lines
    public static final String TAG = LocationRecord.class.getSimpleName();

    // Common date format used throughout tool
    public static final String DATE_FORMAT = "dd/MM/yyyy HH:mm:ss";

    // Only need one instance of the sdf object
    private static SimpleDateFormat _SDF_INSTANCE;

    // Number of fields expected in a given record
    public static final int REQUIRED_FIELDS = 5; // LatLng stored as two fields in text

    // Total number of fields in each record
    //   Used when writing records to pad each stored record with null fields
    public static final int TOTAL_NUM_FIELDS = REQUIRED_FIELDS + 0;

    public static final LocationRecord EMPTY = new LocationRecord();

    private LocationRecord() {}

    // Constructor
    public LocationRecord( Date date, LatLng location, String name, Date endTime )
    {
        Log.d( TAG, "LocationRecord( " + date + ", " + location + ", " + name + ", " + endTime + " ) " );
        this.date = date;
        this.location = location;
        this.name = name;
        this.endTime = endTime;
    }

    private static SimpleDateFormat getSdfInstance()
    {
        if( _SDF_INSTANCE == null )
        {
            _SDF_INSTANCE = new SimpleDateFormat( DATE_FORMAT );
            _SDF_INSTANCE.setTimeZone( TimeZone.getTimeZone( "UTC" ) );
        }

        return _SDF_INSTANCE;
    }

    public LocationRecord( int recordId, String[] data ) throws IOException
    {
        // Verify the string array passed in has the correct number of fields
        if( data.length != REQUIRED_FIELDS )
        {
            if( data.length == 4 )
                Log.w( TAG, "String[] only has 4 items, adding endtime" );
            else
            {
                for( int i = 0; i < data.length; i++ )
                    Log.e( TAG, "String[" + i + "] = " + data[i] );
                throw new IOException( "Invalid number of parameters (" + data.length + " vs. " + REQUIRED_FIELDS + ") [" + recordId + "]" );
            }
        }

        Log.d( TAG, "[" + recordId + "] Creating record from " + "'" + data[0] +
                "', '" + data[1] + "','" + data[2] + "','" + data[3] + "'..." );


        try
        {
            // Convert each string as needed to fit the specified object
            this.date = parseDate( data[0] );
            if( recordId < 5 )
                Log.d( TAG,"Date: " + this.date + ", Input: " + data[0] );

            this.location = new LatLng( Double.parseDouble( data[1] ), Double.parseDouble( data[2] ) );
            this.name = data[3];

            if( data.length > 4 )
                this.endTime = parseDate( data[4] );
        }
        catch( ParseException e )
        {
            throw new IOException( "Failed to parse data - " + e );
        }
    }

    private Date parseDate( String text ) throws ParseException
    {
        if( text.contains( "/" ) )
            return new Date( getSdfInstance().parse( text ).getTime() );
        else
            return new Date( Long.valueOf( text ) );
    }

    /**
     * Returns the current system time.
     *
     * References:
     *    https://stackoverflow.com/questions/6059207/how-can-i-set-a-calendar-with-utc-time
     *
     * @return a Date object or null if current time could not be obtained
     */
    public static Date getCurrentTime()
    {
        Calendar calendar = Calendar.getInstance( TimeZone.getTimeZone( "UTC" ) );
        return new Date( calendar.getTime().getTime() );

        //return new Date( System.currentTimeMillis() );
    }

    /**
     * Converts the date object to the standard date format (DATE_FORMAT).
     *
     * @return A string in DATE_FORMAT form.
     */
    public static String getDateString( Date date )
    {
        if( date != null )
            return getSdfInstance().format( date );
        return "";
    }

    /**
     * Converts the location's LatLng object to the standard format.
     *
     * @return A string representation of Lat/Long
     */
    public String getLocationString()
    {
        String lat = String.format( String.valueOf( location.latitude ), "%2.6f" );
        String lon = String.format( String.valueOf( location.longitude ), "%3.6f" );
        return lat + "," + lon;
    }

    // Getter methods for Lat/Lon as string individually
    public String getLatitudeString() { return getLocationString().split(",")[0]; }
    public String getLongitudeString() { return getLocationString().split(",")[1]; }

    // Getter methods
    public Date getDate()       { return date; }
    public LatLng getLocation() { return location; }
    public String getName()     { return name; }
    public Date getEndTime()    { return endTime; }

    // Setter methods
    public void setDate( Date d )       { date = d; }
    public void setLocation( LatLng l ) { location = l; }
    public void setName( String n )     { name = n; }
    public void setEndTime( Date d )    { endTime = d; }

    @Override
    public int describeContents()
    {
        return 0;
    }

    @Override
    public void writeToParcel(Parcel out, int flags)
    {
        Log.d( TAG, "Parcel Out: " + this );
        out.writeLong( date.getTime() );
        out.writeDouble( location.latitude );
        out.writeDouble( location.longitude );
        out.writeString( name );
        out.writeLong( ( endTime != null ? endTime.getTime() : 0 ) );
    }

    public static final Parcelable.Creator<LocationRecord> CREATOR
            = new Parcelable.Creator<LocationRecord>() {
        public LocationRecord createFromParcel(Parcel in) {
            return new LocationRecord(in);
        }

        public LocationRecord[] newArray(int size) {
            return new LocationRecord[size];
        }
    };

    private LocationRecord(Parcel in)
    {
        date = new Date( in.readLong() );
        location = new LatLng( in.readDouble(), in.readDouble() );
        name = in.readString();

        long et = in.readLong();
        if( et > 0 )
            endTime = new Date( et );

        Log.d( TAG, "Parcel In: " + this );
    }

    /**
     * Returns a CSV string representing the record
     *
     * @return CSV string
     */
    public String toDisplayString()
    {
        StringBuffer sb = new StringBuffer();
        sb.append( LocationRecord.getDateString( date ) );
        sb.append( FileManager.CSV_SEPARATOR );
        sb.append( getLocationString() );
        sb.append( FileManager.CSV_SEPARATOR );
        sb.append( FileManager.readString( getName() ) );
        if( getEndTime() != null )
        {
            sb.append( FileManager.CSV_SEPARATOR );
            sb.append( LocationRecord.getDateString( endTime ) );
        }
        return sb.toString();
    }

    public String toString()
    {
        StringBuffer sb = new StringBuffer();
        sb.append( date.getTime() );
        sb.append( FileManager.CSV_SEPARATOR );
        sb.append( getLocationString() );
        sb.append( FileManager.CSV_SEPARATOR );
        sb.append( FileManager.writeString( name ) );
        sb.append(FileManager.CSV_SEPARATOR);
        if( endTime != null )
            sb.append( endTime.getTime() );
        return sb.toString();
    }
}
