package com.tripgis.dmt;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;

import com.tripgis.dmt.file.EventFileManager;
import com.tripgis.dmt.file.FileManager;
import com.tripgis.dmt.file.LegacyFileManager;
import com.tripgis.dmt.file.TripFileManager;
import com.tripgis.dmt.types.EventRecord;
import com.tripgis.dmt.types.TripDbRecord;
import com.tripgis.dmt.util.DmtUtils;
import com.tripgis.dmt.util.LogTranslator;
import com.tripgis.shared.types.LocationRecord;

public class DirtFarmer 
{
    // Name to use in log lines
    public static final String TAG = DirtFarmer.class.getSimpleName();
    public static final LogTranslator Log = new LogTranslator( TAG );

	public static final String BMK_EXTENSION_REGEX = "." + DmtUtils.bmk + "$";

	private String source;
	private int tripId;
	private LegacyFileManager lfm;
	private TripFileManager tfm;
	private EventFileManager efm;
	
	public DirtFarmer( TripDbRecord tdr ) throws IOException
	{
        Log.d( TAG, "DirtFarmer( " + tdr + " )" );
		source = tdr.getSource();
		tripId = tdr.getTripId();
		if( source.equals( TripDbRecord.SOURCE_LEGACY ) )
			lfm = new LegacyFileManager( tdr.getFileName() );
		else if( source.endsWith( TripDbRecord.SOURCE_MOBILE ) )
			tfm = new TripFileManager( tdr.getFileName() );
		else
			efm = new EventFileManager( tdr.getFileName() );
		
		if( !getLastError().isEmpty() )
			throw new IOException( getLastError() );
	}
	
	public DirtFarmer( File file ) throws IOException
	{
        Log.d( TAG, "DirtFarmer( " + file + " )" );
		source = readFirstLine( file );
		if( source.equals( TripDbRecord.SOURCE_LEGACY ) )
			lfm = new LegacyFileManager( file, file.getName() );
		else if( source.equals( TripDbRecord.SOURCE_MOBILE ) )
			tfm = new TripFileManager( file, file.getName() );
		else
			efm = new EventFileManager( file, file.getName() );
	}
	
	public String getFileName()
	{
		if( source.equals( TripDbRecord.SOURCE_LEGACY ) )
			return lfm.getHeader().getFileName();
		else if( source.endsWith( TripDbRecord.SOURCE_MOBILE ) )
			return tfm.getHeader().getFileName();
		return efm.getHeader().getFileName();
	}
	
	public String getTripName()
	{
		if( source.equals( TripDbRecord.SOURCE_LEGACY ) )
			return lfm.getHeader().getTripName();
		else if( source.endsWith( TripDbRecord.SOURCE_MOBILE ) )
			return tfm.getHeader().getTripName();
		return efm.getHeader().getTripName();
	}
	
	public int getTripId()
	{
		if( tripId != 0 )
			return tripId;
		
		if( source.equals( TripDbRecord.SOURCE_LEGACY ) )
			return lfm.getHeader().getTripId();
		else if( source.endsWith( TripDbRecord.SOURCE_MOBILE ) )
			return 0; // tfm.getHeader().getTripId();

		return efm.getHeader().getTripId();
	}
	
	public ArrayList<EventRecord> getList()
	{
		if( source.equals( TripDbRecord.SOURCE_LEGACY ) )
			return lfm.getList();
		else if( source.endsWith( TripDbRecord.SOURCE_MOBILE ) )
			return tfm.getList();
		return efm.getList();
	}
	
	@SuppressWarnings("rawtypes")
	public FileManager getFileManager()
	{
		if( source.equals( TripDbRecord.SOURCE_LEGACY ) )
			return lfm;
		else if( source.endsWith( TripDbRecord.SOURCE_MOBILE ) )
			return tfm;
		return efm;
	}
	
    public boolean createBookmark() throws IOException
    {
    	File newFile = getWorkingFile();
    	String name = DmtUtils.removeExtension( newFile );
		name += "_" + DmtUtils.getTimestamp( LocationRecord.getCurrentTime() );
		name += "." + DmtUtils.bmk;
		return getFileManager().changeFile( name, true );
    }
    
    public boolean saveAs( File file, boolean overwrite ) throws IOException
    {
    	return getFileManager().changeFile( file.getName(), overwrite );
    }
    
    public File getWorkingFile() { return getFileManager().getWorkingFile(); }
    public String getLastError() { return getFileManager().getLastError(); }
    
    public boolean updateList( EventRecord er ) throws IOException
    {
    	Log.t( TAG, "updateList( " + er + " )" );
    	if( er.isNew() )
    		er.getMobileData().setName( er.getMobileData().getName().replaceFirst( DmtUtils.NEW_PREFIX, "" ) );
    	else if( er.isUpdated() )
    		er.getMobileData().setName( er.getMobileData().getName().replaceFirst( DmtUtils.UPDATE_PREFIX, "" ) );
    	
    	if( !getFileManager().replaceItem( er ) )
    		return getFileManager().addItem( er );
    	
    	return true;
    }
    
    public EventRecord createEventRecord( int tripId, String name ) throws IOException
    {
    	LocationRecord lr = LocationRecord.EMPTY.copy();
    	lr.setName( name );
    	EventRecord er = new EventRecord( lr );
    	er.setTripId( tripId );
    	if( updateList( er ) )
    			return er;
    	return EventRecord.EMPTY;
    }
    
    /**
     * References:
     *    https://stackoverflow.com/questions/8646984/how-to-list-files-in-an-android-directory
     *    https://stackoverflow.com/questions/924394/how-to-get-the-filename-without-the-extension-in-java
     *
     * @return
     * @throws IOException
     */
    public ArrayList<String> getBookmarks() throws IOException
    {
        Log.t( TAG, "getBookmarks()" );
        File path = FileManager.getAlbumStorageDir();
        Log.d( TAG, "Reading " + path + "..." );

        ArrayList<String> list = new ArrayList<>();
        File[] files = path.listFiles();
        if( files != null )
        {
            Log.d( TAG, "Number of files: " + files.length );
            for( int i = 0; i < files.length; i++ )
            {
                String name = files[i].getName();
                if( name.matches( BMK_EXTENSION_REGEX ) )
                {
                    Log.d( TAG, "Adding File '" + files[i].getName() + "'" );
                    list.add( name.replaceFirst( BMK_EXTENSION_REGEX, "" ) );
                }
            }
        }

        Log.d( TAG, "Number of Bookmarks for '" + getTripName() + "': " + list.size() );
        return list;
    }
    
	private String readFirstLine( File file )
	{
        Log.d( TAG, "readFirstLine( " + file + " )" );

        try
        {
        	if( file.length() == 0 )
        		return "File is empty";
        	
            // Prepare the file for reading
            FileReader fileReader = new FileReader( file );
            BufferedReader lineReader = new BufferedReader( fileReader );
            String line = lineReader.readLine();
            String source = TripDbRecord.SOURCE_EVENT;
            String[] data = line.split( FileManager.CSV_SEPARATOR, -1 );
            if( data.length <= LocationRecord.REQUIRED_FIELDS )
            {
            	Log.d( TAG, "First line contains a trip record: " + line );
            	// Could be event records or location records
            	line = lineReader.readLine();
                data = line.split( FileManager.CSV_SEPARATOR, -1 );
                if( data.length <= LocationRecord.REQUIRED_FIELDS )
                	source = TripDbRecord.SOURCE_MOBILE;
            }
            else //if( data.length > LocationRecord.REQUIRED_FIELDS )
            	source = TripDbRecord.SOURCE_LEGACY;

        	lineReader.close();
        	fileReader.close();
            
            Log.d( TAG, "Source: " + source );
            return source;
        }
        catch( FileNotFoundException e )
        {
            return "Read failed on '" + file + "': " + e;
        }
        catch( IOException e )
        {
            return "Could not read first line of '" + file + "': " + e;
        }
	}
	
	public String toString()
	{
		StringBuffer sb = new StringBuffer();
		sb.append( "Source=" + source + "," );
		sb.append( "TripName=" + getTripName() + "," );
		sb.append( "FileName=" + getFileName() + "," );
		sb.append( "ListSize=" + getList().size() );
		return sb.toString();
	}
}
