package com.tripgis.dmt;

import java.io.File;
import java.io.IOException;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;

import com.tripgis.dmt.db.DbModel;
import com.tripgis.dmt.file.FileManager;
import com.tripgis.dmt.geocode.Geocoder;
import com.tripgis.dmt.geocode.TripStatistics;
import com.tripgis.dmt.types.EventRecord;
import com.tripgis.dmt.types.EventSubType;
import com.tripgis.dmt.types.EventType;
import com.tripgis.dmt.types.GeocodeRecord;
import com.tripgis.dmt.types.MatchRecord;
import com.tripgis.dmt.types.TripDbRecord;
import com.tripgis.dmt.ui.AnalysisForm;
import com.tripgis.dmt.ui.EventForm;
import com.tripgis.dmt.ui.SecurePanel;
import com.tripgis.dmt.util.LatLng;
import com.tripgis.dmt.util.LogTranslator;
import com.tripgis.dmt.util.DmtUtils;

public class DmtController {
	
	private static DmtController obj;
	
	private EventForm gui;
	private DbModel db;
	private Geocoder geocoder;
	
//	private BookmarkFileManager bfm;
	private DirtFarmer dfm;
	
    // Name to use in log lines
    public static final String TAG = DmtController.class.getSimpleName();
    public static final LogTranslator Log = new LogTranslator( TAG );

	public static DmtController getInstance() 
	{
		if( obj == null )
		{
			Log.d( TAG, "Creating " + TAG + " object" );
			obj = new DmtController();
		}
		
		return obj;
	}
	
	private DmtController() 
	{
		Log.t( TAG, "DmtController()" );
		gui = new EventForm();
		try 
		{
			db = DbModel.getInstance();
			geocoder = Geocoder.getInstance();
		} 
		catch (SQLException e) 
		{
			setError( TAG, "Failed to connect to db - " + e );
		}
    }
	
	public void initialize()
	{
		Log.t( TAG, "initialize()" );
		
		gui.initialize();

		if( db.isConnected() )
		{
			gui.setMessage( TAG, "DB Connected - Populating GUI" );
			updateEventTypeList();
			updateEventSubTypeList( EventType.EMPTY );
			updateTripList();
		}

		if( db.isConnected() )
			gui.enableOnDbConnect();
		else
			gui.enableOnDbDisconnect();

		gui.setMessage( TAG, "GUI Initialized" );
	}
	
	public void updateTripList()
	{
		Log.t( TAG, "updateEventTypeList()" );
		gui.setTripList( db.getTripList() );
	}
	
	public void updateEventList( int tripId )
	{
		Log.t( TAG, "updateEventTypeList( " + tripId + " )" );
		gui.setEventList( db.getEventsByTripId( tripId ) );
	}
	
	public void updateEventTypeList()
	{
		Log.t( TAG, "updateEventTypeList()" );
		gui.setEventTypeList( db.getEventTypes() );
	}
	
	public void updateEventSubTypeList( EventType type )
	{
		Log.t( TAG, "updateEventSubTypeList( " + type + " )" );
		gui.setEventSubTypeList( db.getEventSubTypesByType( type ) );
	}
	
	public boolean openTrip( TripDbRecord tdr )
	{
		Log.t( TAG, "openTrip( " + tdr + " )" );
		try
		{
			int tripId = tdr.getTripId();
			if( tripId != 0 )
				tdr.setList( db.getEventsByTripId( tripId ) );
			//bfm = new BookmarkFileManager( tripId, tdr.getTripName() );
			dfm = new DirtFarmer( tdr );
			if( dfm.getList() == null )
			{
				setError( TAG, dfm.getFileManager().getLastError() );
				return false;
			}
			
			Log.d( TAG, "DIRTY: " + dfm );
			Log.d( TAG, "Original: " + dfm.getFileManager().getOriginalFile() );
			Log.d( TAG, "Working: " + dfm.getFileManager().getWorkingFile() );
			return( tdr.getItemCount() > 0 );
		}
		catch( IOException e )
		{
			setError( TAG, "Failed to open file - " + e );
			return false;
		}
	}
	
	public void display() 
	{ 
		gui.enableListeners( true );
		gui.display(); 
	}

	/**
	 * References:
	 *    http://www.wikihow.com/Close-a-Window-in-Java
	 */
	public void exitProgram()
	{
		Log.t( TAG, "exitProgram()" );

		if( gui.exitApplication() )
		{
			try
			{
//			if( db.isActiveTransaction() )
				db.disconnect();
			}
			catch( SQLException e )
			{
				setError( TAG, "Failed to disconnect from db - " + e );
			}
			
           System.exit(0);
		}
	}
	
	/******************************************************************
	 * FILE FUNCTIONS
	 *****************************************************************/

	public TripDbRecord openFile( File file )
	{
		Log.t( TAG, "openFile( " + file + " )" );
		
		try
		{
			//bfm = new BookmarkFileManager( 0, DmtUtils.removeExtension( file.getName() ), file );
			dfm = new DirtFarmer( file );

			Log.d( TAG, "DIRTY: " + dfm );
			Log.d( TAG, "Original: " + dfm.getFileManager().getOriginalFile() );
			Log.d( TAG, "Working: " + dfm.getFileManager().getWorkingFile() );
			/*	
			String name = bfm.getRecord().getTripName();
			TripFileManager tfm = bfm.getMobileFileManager();
			String mobileName = tfm.getRecord().getName();
			if( !name.equals( mobileName ) )
			{
				Log.w( TAG, "Mobile file name does not match stored: " + name + " vs. " + mobileName );
				name = mobileName;
				tfm.getRecord().setName( mobileName );
			}*/
			
			String name = dfm.getTripName();
			Log.d( TAG, "Opened file for trip '" + name + "' Locations: " + dfm.getList().size() );
			
			TripDbRecord trip = db.getTripByName( name );
			ArrayList<EventRecord> events = new ArrayList<>();
	
			if( trip != TripDbRecord.EMPTY )
			{
				Log.d( TAG, "Found trip '" + trip.getTripName() + "' in db, id: " + trip.getTripId() );
				events = db.getEventsByTripId( trip.getTripId() );
				
				if( mergeFileAndDbLists( events, dfm.getList() ) )
				{
					trip.setTripName( DmtUtils.setUpdPrefix( trip.getTripName() ) );
				}
			}
			else
			{
				Log.d( TAG, "Trip '" + name + "' is new!" );
				trip = new TripDbRecord( 0, DmtUtils.setNewPrefix( name ) );
				events = dfm.getList();
			}
			
			Log.d( TAG, "Total Events: " + events.size() );
			trip.setList( events );
			ArrayList<TripDbRecord> trips = db.getTripList();
			trips.add( trip );
			gui.setTripList( trips );
			return trip;
		}
		catch( IOException e )
		{
			setError( TAG, "Failed to open file - " + e );
			return null;
		}
	}
	
	private boolean mergeFileAndDbLists( ArrayList<EventRecord> db, ArrayList<EventRecord> file )
	{
		Log.t( TAG, "mergeFileAndDbLists( " + db.size() + ", " + file.size() + " )" );
		
		for( EventRecord record : file )
			record.getMobileData().setName( DmtUtils.setUpdPrefix( record.getMobileData().getName() ) );
		
		int size = db.size();
		TripStatistics.mergeLists( db, file );
		Log.d( TAG,  "Original Size: " + size + ", New: " + db.size() );
		return( size != db.size() );
	}
	
/*	public TripDbRecord openLegacyFile( File file )
	{
		Log.t( TAG, "openLegacyFile( " + file + " )" );
		
		try
		{
			LegacyFileManager lfm = new LegacyFileManager( file );
			file = lfm.convertFile();
			bfm = new BookmarkFileManager( 0, DmtUtils.removeExtension( file.getName() ), file );
			
			String name = bfm.getRecord().getTripName();
			
			Log.d( TAG, "Created new trip '" + name + "' file: " + bfm.getFile().toString() );
			Log.d( TAG, "   Locations: " + bfm.getList().size() );
			
			TripDbRecord trip = db.getTripByName( name );
			ArrayList<EventRecord> events = new ArrayList<>();
	
			boolean diffExist = false;
			if( trip != TripDbRecord.EMPTY )
			{
				Log.d( TAG, "Found trip '" + trip.getTripName() + "' in db, id: " + trip.getTripId() );
				events = db.getEventsByTripId( trip.getTripId() );
				Log.d( TAG, "   Stored Locations: " + events.size() );

				events = LegacyRecord.legacyToEventRecords( lfm.getList() );
//				if( mergeEventLists( events, lfm.getList() ) )
//				{
//					trip.setTripName( DmtUtils.setUpdPrefix( trip.getTripName() ) );
//					diffExist = true;
//				}
			}
			else
			{
				Log.d( TAG, "Trip '" + name + "' is new!" );
				trip = new TripDbRecord( 0, DmtUtils.setNewPrefix( name ) );
				events = LegacyRecord.legacyToEventRecords( lfm.getList() );
				diffExist = true;
			}
			
			Log.d( TAG, "Total Events: " + events.size() );
			trip.setList( events );
			ArrayList<TripDbRecord> trips = db.getTripList();
			trips.add( trip );
			gui.setTripList( trips );
			bfm.getRecord().setList( events );
			//gui.setChanged( diffExist );
//			gui.setEventList( events );
			return trip;
		}
		catch( IOException e )
		{
			setError( TAG, "Failed to open file - " + e );
			return null;
		}
	} */

/*	private boolean mergeEventLists( ArrayList<EventRecord> dbList, ArrayList<LegacyRecord> fileList )
	{
		Log.t( TAG, "mergeEventLists( " + dbList.size() + ", " + fileList.size() + " )" );
		boolean modified = false;
		for( LegacyRecord fRec : fileList )
		{
			boolean found = false;
			EventRecord newRec = (EventRecord) fRec;
			for( EventRecord record : dbList )
			{
				if( record.matches( newRec ) )
				{
					found = true;
					break;
				}
			}
			
			if( !found )
			{
				//newRec.setName( "NEW: " + newRec.getName() );
				dbList.add( newRec );
				modified = true;
			}
		}

		Log.d( TAG, "Unique records: " + dbList.size() );
		return modified;
	}
*/
	
/*	private boolean prepareEventList( ArrayList<EventRecord> dbList, ArrayList<LocationRecord> fileList )
	{
		Log.t( TAG, "prepareEventList( " + dbList.size() + ", " + fileList.size() + " )" );
		boolean modified = false;
		for( LocationRecord location : fileList )
		{
			EventRecord er = EventRecord.find( dbList, location );
			if( er == null )
			{
				location.setName( "NEW: " + location.getName() );
				dbList.add( new EventRecord( location ) );
				modified = true;
			}
			else if( er.isLocationModified( location ) )
			{
				location.setName( "UPD: " + location.getName() );
				er.setMobileData( location );
				modified = true;
			}
		}

		Log.d( TAG, "Unique records: " + dbList.size() );
		return modified;
	}
*/
	
	public ArrayList<TripDbRecord> closeFile()
	{
		Log.t( TAG, "closeFile()" );
/*		if( bfm != null )
		{
			String name = bfm.getRecord().getTripName();
			bfm.delete( true );
			bfm = null;
		} */

		// TODO - prompt to remove all bookmarks
		if( dfm != null )
			dfm.getFileManager().deleteFile( true );
		
		return db.getTripList();
	}

	/**
	 * @param bookmark
	 * @return
	 */
	public boolean saveFile( boolean bookmark )
	{
		Log.t( TAG, "saveFile( " + bookmark + " )" );
		try 
		{
			TripDbRecord tdr = gui.getSelectedTrip();

			// Update data from form
			EventRecord er = gui.getEventDataFromForm();
			Log.d( TAG, "Form Event Record: " + er );

			if( bookmark )
			{
				Log.d( TAG, "Creating new bookmark..." );
				dfm.createBookmark();
				gui.setMessage( TAG, "Trip [" + tdr.getTripId() + "] Saved to " + dfm.getWorkingFile() );
			}
			else
			{
				if( !dfm.updateList( er ) )
				{
					setError( TAG, "Failed to update file: " + dfm.getLastError() );
					return false;
				}

				//TODO - get trip record and add the new list to it...or something
//				db.getTrips().updateTrip( tdr );
				gui.setMessage( TAG, "Trip [" + tdr.getTripId() + "] " + tdr.getTripName() + " saved" );
			}

			gui.setChanged( false );
			return true;
		}
		catch (IOException e) 
		{
			DmtController.setError( TAG, "Failed to save data - " + e );
		}
		
		return false;
	}
	
	public boolean saveAsFile( File file )
	{
		Log.t( TAG, "saveAsFile( " + file + " )" );
		try 
		{
			if( FileManager.isFileCreated( file.getName() ) && 
					!yesNoDialog( "File exists. Overwrite?", "File Exists!" ) )
			{
				gui.setMessage( TAG, "Operation cancelled" );
				return false;
			}

			// Change the file data is being stored to
			dfm.saveAs( file, true );

			// Update data from form
			EventRecord er = gui.getEventDataFromForm();
			dfm.updateList( er );

			gui.setMessage( TAG, "Trip Saved as " + dfm.getWorkingFile() );
			gui.setChanged( false );
			return true;
		} 
		catch( IOException e )
		{
			DmtController.setError( TAG, "Failed to save as " + file + ": " + e );
			return false;
		}
	}

	/******************************************************************
	 * EVENT FUNCTIONS
	 *****************************************************************/
	
	public boolean removeEvent( EventRecord er )
	{
		Log.t( TAG, "removeEvent( " + er + " )" );
		try
		{
			return db.getEvents().removeEvent( er.getEventId() );
		}
		catch( SQLException e )
		{
			setError( TAG, "SQL: Failed to remove event '" + er.getMobileData().getName()+ "' - " + e );
			return false;
		}
	}

	public TripDbRecord createTrip( String name )
	{
		Log.t( TAG, "createTrip( " + name + " )" );
		try
		{
			return db.getTrips().insertTrip( name );
		}
		catch( SQLException e )
		{
			setError( TAG, "Failed to create trip '" + name + "' - " + e );
			return TripDbRecord.EMPTY;
		}
	}

	public EventRecord createEvent( String name )
	{
		Log.t( TAG, "createEvent( " + name + " )" );
		try
		{
			EventRecord er = db.getEvents().insertEvent( dfm.getTripId(), name );
			if( dfm.updateList( er ) )
			{
				setMessage( TAG, "Event " + er.getDescription() + " created!" );
				return er;
			}
			
			setError( TAG, "Failed to create event '" + name + "' - " + dfm.getLastError() );
		}
		catch( IOException e )
		{
			setError( TAG, "Failed to create event '" + name + "' - " + e );
		}
		catch( SQLException e )
		{
			setError( TAG, "SQL: Failed to create event '" + name + "' - " + e );
		}

		return EventRecord.EMPTY;
	}

	public EventType createEventType( String name )
	{
		Log.t( TAG, "createEventType( " + name + " )" );
		try
		{
			return db.getTypes().insertEventType( name );
		}
		catch( SQLException e )
		{
			setError( TAG, "Failed to create EventType '" + name + "' - " + e );
			return EventType.EMPTY;
		}
	}

	public EventSubType createEventSubType( EventType type, String name )
	{
		Log.t( TAG, "createEventSubType( " + name + " )" );
		try
		{
			return db.getTypes().insertEventSubType( type, name );
		}
		catch( SQLException e )
		{
			setError( TAG, "Failed to create EventSubType '" + name + "' (" + type + ") - " + e );
			return EventSubType.EMPTY;
		}
	}

	/******************************************************************
	 * DATABASE FUNCTIONS
	 *****************************************************************/

	/**
	 * Database (Model) Operation
	 */

/*	public static void authenticate() 
	{
		if( DmtController.promptSecure( "Please enter root password:", "Password" ) )
		{
			getInstance().db.setPassword( getUserResponse() );
			System.out.println( "Verified privileged user" );
		}
	} */

	public boolean connectToDb( boolean flag )
	{
		Log.t( TAG, "connectToDb()" );
		String tag = ( flag ? "Connect to" : "Disconnect from" );

		try
		{
			if( flag )
				db.connect();
			else
				db.disconnect();

			gui.setMessage( TAG, tag + " the database" );
			return true;
		}
		catch( SQLException e )
		{
			setError( TAG, "Failed to " + tag + " the database - " + e );
			return false;
		}
	}
	
	public ArrayList<EventRecord> saveEvent( EventRecord event )
	{
		Log.t( TAG, "saveEvent( " + event + " )" );
		try 
		{
			if( dfm.getTripId() == 0 )
			{
				Log.d( TAG, "Trip " + dfm.getTripName() + " not in DB, adding..." );
				commitChanges();
			}
			else
			{
				db.getEvents().updateEvent( event );
			}

			return db.getEvents().getEventsByTripId( event.getTripId() );
		} 
		catch( SQLException e ) 
		{
			setError( TAG, "SQL: Failed to commit changes - " + e );
		}
		
		return null;
	}
	
	public TripDbRecord commitChanges()
	{
		Log.t( TAG, "commitChanges()" );
		try 
		{
			EventRecord er = gui.getEventDataFromForm();
			TripDbRecord tdr = gui.getSelectedTrip();
			
			if( !dfm.updateList( er ) )
			{
				setError( TAG, "Failed to update event - " + dfm.getLastError() );
				return tdr;
			}
			
			tdr = db.getTrips().updateTrip( tdr );
			if( tdr == TripDbRecord.EMPTY )
				throw new SQLException( "Could not find '" + dfm.getTripName() );

			Log.d( TAG, "Committing changes for " + dfm.getList().size() + " events (TripId: " + tdr.getTripId() + ")" );
			for( EventRecord event : dfm.getList() )
			{
				event.setTripId( tdr.getTripId() );
				
				if( event.getDescription().isEmpty() )
					event.setDescription( event.getDisplayString() );
				
				if( event.getNotes().isEmpty() )
					event.setNotes( event.getMobileData().getName() );

				db.getEvents().updateEvent( event );
			}
			
			setMessage( TAG, "Changes committed to DB!" );
			// TODO - update drop-down to reflect saved
			// disable update, save, reset
			return tdr;
		} 
		catch( IOException e ) 
		{
			setError( TAG, "WRITE: Failed to commit changes - " + e );
		}
		catch( SQLException e ) 
		{
			setError( TAG, "SQL: Failed to commit changes - " + e );
		}

		return TripDbRecord.EMPTY;
	}
	
	public boolean exportDb( File file )
	{
		Log.t( TAG, "exportDb( " + file + " )" );
		String result = "Successful";
		
		//TODO - DmtController.getDb().isInitialized();
		if( file.exists() )
		{
			if( !yesNoDialog( "File " + file.getName() + " exists! Overwrite?", "Overwrite SQL" ) )
				result = "Cancelled";
		}
		else
		{
			DmtController.setMessage( TAG, "Exporting Database to " + file.getName() + "..." );
			if( !DmtController.getDb().exportDatabase( file ) )
				result = "Failed";
		}

		gui.setMessage( TAG, "Export Database " + result + "!" );
		return result.equals( "Success" );
	}
	
	
	public boolean importDb( File file )
	{
		Log.t( TAG, "importDb( " + file + " )" );
		String result = "Successful";

		//TODO boolean dbBuilt = DmtController.getDb().isInitialized();
		if( yesNoDialog( "Do you wish to continue building database from " +
				file.getName() + "?", "Rebuild Database" ) )
		{
			gui.setMessage( TAG, "Importing Database from " + file.getName() + "..." );
			if( !DmtController.getDb().importDatabase( file ) )
				result = "Failed";
		}
		else
			result = "Cancelled";

		gui.setMessage( TAG, "Import Database " + result + "!" );
		return result.equals( "Success" );
	}

	/******************************************************************
	 * GEOCODE FUNCTIONS
	 *****************************************************************/

	public EventRecord updateAssociation( EventRecord er )
	{
		// TODO - what's this for?
		Log.t( TAG, "updateAssociation()" );
		try 
		{
			if( !er.getGeoData().getPlaceId().isEmpty() )
			{
					er.setGeoData( geocoder.detail( er.getGeoData().getPlaceId() ) );
					Log.d( TAG, "Detail matched place id!" );
					return er;
			}

			ArrayList<MatchRecord> list = geocoder.reverseGeocode( er.getMobileData().getLocation() );
			if( list.size() == 0 )
			{
				Log.d( TAG, "Geocode returned 1 result, using it!" );
				er.setGeoData( list.get( 0 ).getRecord() );
				return er;
			}

			//TODO _ move gui.populateNearbyList( list );
			Log.d( TAG, "PlaceId empty and multiple places match Lat/Lng" );
			return null;
		}
		catch (IOException e) 
		{
			setError( TAG, "Failed to update association - " + e );
			return null;
		}
	}
	
	public EventRecord setAssociation( MatchRecord mr )
	{
		Log.t( TAG, "setAssociation( " + mr + " )" );
		if( mr.getRecord() == GeocodeRecord.EMPTY )
		{
			DmtController.setError( TAG, "No association selected!" );
			return null;
		}

		EventRecord er = gui.getEventDataFromForm();
		er.setGeoData( mr.getRecord() );
		return er;
	}
	
	public ArrayList<MatchRecord> searchNearby( LatLng location, String radius, String keyword )
	{
		Log.t( TAG, "searchNearby( " + radius + ", " + keyword + " )" );
		try 
		{
			HashMap<String,String> params = new HashMap<>();
			params.put( "radius", radius );
			params.put( "keyword", Geocoder.getParameterString( keyword ) );
			return geocoder.nearbySearch( location, params );
		} 
		catch( IOException e )
		{
			setError( TAG, "Failed to search nearby - " + e );
			return null;
		}
	}
	
	public ArrayList<MatchRecord> geocode( String address )
	{
		Log.t( TAG, "geocode( " + address + " )" );
		try 
		{
			return geocoder.geocode( address );
		} 
		catch( IOException e )
		{
			setError( TAG, "Failed to search nearby - " + e );
			return null;
		}
	}
	
	/******************************************************************
	 * ANALYSIS FUNCTIONS
	 *****************************************************************/
	
	public void updateDurations()
	{
		Log.t( TAG, "updateDurations()" );
		try 
		{
			int tripId = gui.getSelectedTrip().getTripId();
			int count = db.getEvents().updateDuration( tripId );
			setMessage( TAG, "Updated " + count + " durations!" );
		} 
		catch( SQLException e )
		{
			setError( TAG, e.toString() );
		}
	}
	
	public TripStatistics generateStatistics()
	{
		Log.t( TAG, "generateStatistics()" );
		TripStatistics ts = new TripStatistics( this );
		ts.generate();
		return ts;
	}

	public TripStatistics generateGapStatistics( EventRecord start, EventRecord end )
	{
		Log.t( TAG, "generateGapStatistics( " + start.getMobileData().getDate() + ", " + end.getMobileData().getDate() + " )" );
		TripStatistics ts = new TripStatistics( this );
		if( !ts.isGenerated() )
			ts.generate();
		ts.calculateGapStats( start, end );
		return ts;
	}
	
	public void closeForm( Object obj )
	{
		Log.t( TAG, "closeForm( " + obj + " )" );
		if( obj instanceof AnalysisForm )
		{
			Log.d( TAG, "Analysis Window Closed" );
			gui.enableOnAnalyzeClose();
		}
	}

	/******************************************************************
	 * GENERAL FUNCTIONS
	 *****************************************************************/

//	public BookmarkFileManager getBookmarkManager() { return bfm; }
/*	private static int count = 0;
	public static ArrayList<EventRecord> getEventsFromFileAndDbOld( TripDbRecord tdr )
	{
		Log.t( TAG, "getEventsFromFileAndDb( " + tdr.toString() + " )" );
		if( count++ > 3 )
			System.exit( 2 );
		
		TripFileManager currentTfm = getInstance().getTripFileManager();
		ArrayList<EventRecord> events = getDb().getEventsByTripId( tdr.getTripId() );
		if( currentTfm == null || currentTfm.getTrip().getName() != tdr.getTripRecord().getName() )
		{
			// TODO - getInstance().openFile( tdr.getTripRecord().getName() );
			currentTfm = getInstance().getTripFileManager();
		}

		events.add( 0, EventRecord.EMPTY );
		for( LocationRecord lr : currentTfm.getTrip().getSavedLocations() )
		{
			for( EventRecord er : events )
			{
				if( !er.matches( lr ) )
				{
					lr.setName( "NEW: " + lr.getName() );
					events.add( new EventRecord( lr ) );
				}
			}
		}
		
		return events;
	} */
	
	/**
	 * GUI (View) Operations
	 */
	public static EventForm getDisplay() { return getInstance().gui; }
	public static DbModel getDb()         { return getInstance().db; }
	public static Geocoder getGeocoder() { return getInstance().geocoder; }
//	public static TripDbRecord getStoredTrip() { return getInstance().storedTrip; } 
	
	public boolean notify( String text, String title ) { return getInstance().gui.getDialogs().notifyDialog( text,  title ); }
	public boolean yesNoDialog( String text, String title ) { return getInstance().gui.getDialogs().yesNoDialog( text,  title ); }
	public String promptOkCancel( String text, String title ) { return getInstance().gui.getDialogs().textDialog( text,  title ); }
	public String promptOkCancelSecure( String text, String title ) { return new SecurePanel( text,  title ).promptOkCancelSecure(); }
	
	public static void setMessage( String tag, String message )
	{
//		if( message.contains( "Unknown database" ) )
//		{
//			if( getInstance().gui.queryUser( "Database does not exist. Create?", "Create Database" ) )
//				getInstance().db.buildDatabase();
//		}
		getInstance().gui.setMessage( tag, message );
	}
	
	public static void setError( String tag, String message )
	{
		Log.e( TAG,  message );
		DmtController.getInstance().notify( message, "ERROR!" );
	}
	
	/**
	 * Main program for Travel program
	 * 
	 * References:
	 *    https://www.tutorialspoint.com/log4j/log4j_logging_levels.htm
	 *    https://stackoverflow.com/questions/5081316/where-is-the-correct-location-to-put-log4j-properties-in-an-eclipse-project
	 *    
	 * @param args 
	 */
    public static void main(String[] args) 
    {
        //Schedule a job for the event-dispatching thread:
        //creating and showing this application's GUI.
        javax.swing.SwingUtilities.invokeLater(
        		new Runnable() 
        		{
        			public void run() 
        			{
    					DmtController.getInstance().initialize();
    					DmtController.getInstance().display();
        			}
        });
    }
}
