package com.tripgis.dmt.db;

import java.io.File;
import java.sql.Connection;
import java.sql.Date;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Properties;

import com.tripgis.dmt.DmtController;
import com.tripgis.dmt.types.EventRecord;
import com.tripgis.dmt.types.EventSubType;
import com.tripgis.dmt.types.EventType;
import com.tripgis.dmt.types.TripDbRecord;
import com.tripgis.dmt.util.DmtProperties;
import com.tripgis.dmt.util.LogTranslator;
import com.tripgis.dmt.util.DmtUtils;

public class DbModel 
{
	public static DbModel singleton;

	// Name to use in log lines
    public static final String TAG = DbModel.class.getSimpleName();
    public static final LogTranslator Log = new LogTranslator( TAG );
	
	private static String host = DmtProperties.getProperty( "db.host" );
	private static String db = DmtProperties.getProperty( "db.name" );
	private static String port = DmtProperties.getProperty( "db.port" );
	private static String url = "jdbc:postgresql://" +
	   host + ":" + port + "/" + db;
	private static String username = DmtProperties.getProperty( "db.user" );
	private static String password = DmtProperties.getProperty( "db.pass" );

//	private static String psql = DmtProperties.getProperty( "db.script" );

	private Connection conn;
	private DbModelTrips dbTrips;
	private DbModelEvents dbEvents;
	private DbModelTypes dbTypes;
	
	/**
	 * References:
	 *    https://jdbc.postgresql.org/documentation/80/connect.html
	 *    https://www.liquidweb.com/kb/what-is-the-default-password-for-postgresql/
	 *    https://www.liquidweb.com/kb/change-postgresql-authentication-method-from-ident-to-md5
	 *    
	 * @throws SQLException
	 */
	private DbModel() throws SQLException
	{
		Log.t( TAG, "DbModel()" );
		dbTypes = new DbModelTypes( this );
		dbEvents = new DbModelEvents( this );
		dbTrips = new DbModelTrips( this );
		connect();
	}
	
	public static DbModel getInstance() throws SQLException
	{ 
		if( singleton == null )
			singleton = new DbModel();

		return singleton; 
	}
	
	public Connection getDbConnection() throws SQLException
	{
		checkConnection();
		return conn; 
	}
	
	public String getHost() { return host; }
	public String getUser() { return username; }
	public String getUrl() { return url; }

	public String getStatus() 
	{ 
		if( isConnected() ) return "Connected";
		else return "Disconnected";
	}
	
	/******************************************************************
	 * CONNECTION FUNCTIONS
	 *****************************************************************/

	public void connect() throws SQLException
	{
		Log.t( TAG, "connect()" );
		if( !isConnected() )
		{
			Log.d( TAG, "Connecting to " + url );
			Properties props = new Properties();
			props.setProperty( "user", username );
			props.setProperty( "password",  password );
			//props.setProperty( "ssl", true );
			conn = DriverManager.getConnection( url, username, password );
			
			//PreparedStatement pstmt = mysqlConnection.prepareStatement( "Show Tables" );
			//pstmt.execute();
			//pstmt.close();
			
			//validateDatabase();
		}
	}

	public void disconnect() throws SQLException
	{
		if( !conn.isClosed() )
			conn.close();

		conn = null;
	}

	public void reconnect() throws SQLException
	{
		//boolean tmp_pu = privilegedUser;
		//String tmp_password = password;		
		disconnect();
		//password = tmp_password;
		//privilegedUser = tmp_pu;
		connect();
	}
		
	public boolean isConnected()
	{
		try 
		{
			if( conn == null ) return false;
			return !conn.isClosed();
		}
		catch (SQLException e) 
		{
			DmtController.setError( TAG, "Failed to check connection - " + e );
			return false;
		}
	}
	
	private void checkConnection() throws SQLException
	{
		if( !isConnected() )
			throw new SQLException( "Not connected to database" );
	}
	
	/******************************************************************
	 * DATABASE FUNCTIONS
	 *****************************************************************/

	public int getNumberOfTables() throws SQLException
	{
		// TODO - \dt
		Statement stmt = conn.createStatement();
		ResultSet rs = stmt.executeQuery( "Show Tables" );
		
		int rowCount = 0;
		if( rs.last() ) 
		{
			rowCount = rs.getRow();
			rs.beforeFirst();
		}
		
		return rowCount;
	}
	
	public void validateDatabase() throws SQLException
	{
		int rowCount = getNumberOfTables();
		System.out.println( "Number of Tables: " + rowCount );
		if( rowCount == 0 )
		{
			System.out.println( "Database Empty!" );
			//System.out.println( "Building Database from scratch" );
			//buildDbTables();
		}
		else
		{
			Statement stmt = conn.createStatement();
			ResultSet rs = stmt.executeQuery( "Show Tables" );
			while( rs.next() )
			{
				String db = DmtUtils.getString( rs.getString( 1 ) );
				System.out.println( db );
			}
		}
	}
	
	public Statement executeStatement( String sql ) throws SQLException
	{
		Log.t( TAG, "executeStatement()" );
		Statement stmt = conn.createStatement();
		Log.d( TAG, "SQL: " + sql );
		stmt.execute( sql );
		return stmt;
	}
	
	public int getLastRecordId( String table, String idName )
	{
		Log.t( TAG, "getLastRecordId( " + table + ", " + idName + " )" );
		
		Statement st = null;
		ResultSet rs = null;
		try
		{
			st = conn.createStatement();
			rs = st.executeQuery( "SELECT MAX( " + idName + " ) FROM " + table );
			
			if( rs.next() )
			{
				int id = rs.getInt( 1 );
				closeQuery( rs, st );
				Log.d( TAG, "Last Id: " + id );
				return id;
			}
		}
		catch( SQLException e )
		{
			closeQuery( rs, st );
			Log.e( TAG, "Failed to get last " + table + " record - " + e );
		}
		return 0;
	}
	
	public void closeQuery( ResultSet rs, Statement st )
	{
		try {
			if( rs != null ) rs.close();
			if( st != null ) st.close();
		} catch( SQLException se ) {}
	}
	
	public String toSqlString( String text )
	{
		return "'" + text.replaceAll( "'",  "''" ) + "'";
	}

	public String getSqlString( String text )
	{
		if( text != null )
			return text.replaceAll( "''",  "'" );
		return "";
	}
	
	public String getRsString( ResultSet rs, int idx ) throws SQLException
	{
		return getSqlString( rs.getString( idx ) );
	}
	
	public Date getRsDate( ResultSet rs, int idx ) throws SQLException
	{
		if( rs.getTimestamp( idx ) != null )
			return new Date( rs.getTimestamp( idx ).getTime() );
		return null;
	}
	
	/******************************************************************
	 * TYPE FUNCTIONS
	 *****************************************************************/
	public DbModelTypes getTypes() { return dbTypes; }
	public ArrayList<EventType> getEventTypes() { return dbTypes.getEventTypes(); }
	public ArrayList<EventSubType> getEventSubTypesByType( EventType type ) { return dbTypes.getEventSubTypesByType( type ); }
	
	/******************************************************************
	 * TRIP FUNCTIONS
	 *****************************************************************/
	public DbModelTrips getTrips() { return dbTrips; }
	public ArrayList<TripDbRecord> getTripList() { return dbTrips.getTrips(); }
	public TripDbRecord getTripByName( String name ) { return dbTrips.getTripByName( name ); }

	/******************************************************************
	 * EVENT FUNCTIONS
	 *****************************************************************/
	public DbModelEvents getEvents() { return dbEvents; }
	public ArrayList<EventRecord> getEventsByTripId( int tripId ) { return dbEvents.getEventsByTripId( tripId ); }

	/******************************************************************
	 * IMPORT / EXPORT FUNCTIONS
	 *****************************************************************/

	public boolean isPrivilegedUser() 
	{ 
//		if( commandsAvailable && password.equals( DEFAULT_PASSWORD ) )
//			TravelController.authenticate();
		
//		System.out.println( "PU: " + privilegedUser );
//		return privilegedUser; 
		return true;
	}

//	private boolean checkCommandAvailability()
//	{
//		return DmtUtils.executeCommand( psql + " --help >/dev/null" );
//	}

	private void removeDatabase() throws SQLException
	{
		Log.t( TAG, "removeDatabase()" );
		if( !isPrivilegedUser() )
			throw new SQLException( "Error! Not superuser!" );
	
		// dropdb
		Log.d( TAG, "Removing current database..." );
		executeStatement( "Drop Database " + db );
		System.out.println( "Removed database " + db );
		Log.d( TAG, "Removing current database Complete!" );
	}
	
	public boolean buildDatabase()
	{
		Log.t( TAG, "buildDatabase()" );
		try
		{
			if( !isPrivilegedUser() )
				throw new SQLException( "Error! Not superuser!" );

			removeDatabase();

			// createdb
			executeStatement( "Create Database " + db );
			reconnect();
			// TODO importDatabase( "file" );
			return true;
		}
		catch( SQLException e )
		{
			DmtController.setMessage( TAG, "SQL Exception - " + e.getMessage() );
			return false;
		}
	}

	public boolean importDatabase( File sql )
	{
		Log.d( TAG, "importDatabase( " + sql + " )" );
		try
		{
			if( !isPrivilegedUser() )
				throw new SQLException( "Error! Not superuser!" );
			
			removeDatabase();
	
			// TODO
			// Setup expect file
			//boolean retVal = DmtUtils.executeCommand( mysql + " -u " + user + " -p" + pass + " < " + sql );
			//if( retVal )
			//{
			//	DmtController.setMessage( "Importing Database Completed!" );
			//	reconnect();
			//}
			//else
			//	DmtController.setMessage( "Failed to Import Database!" );
			return true;
		}
		catch( SQLException e )
		{
			DmtController.setMessage( TAG, "SQL Exception - " + e.getMessage() );
			return false;
		}
	}
	
	public boolean exportDatabase( File sql )
	{
		Log.t( TAG, "exportDatabase( " + sql + " )" );
		try
		{
			if( !isPrivilegedUser() )
				throw new SQLException( "Error! Not superuser!" );
	
			// TODO
	//		boolean retVal = DmtUtils.executeCommand( mysqldump + " -u " + user +
	//				" -p" + pass + " --database " + dbName + " > " + sql );
	//		if( retVal )
	//			DmtController.setMessage( "Export Database Completed!" );
	//		else
	//			DmtController.setMessage( "Failed to export database!" );
			return false;
		}
		catch( SQLException e )
		{
			DmtController.setMessage( TAG, "SQL Exception - " + e.getMessage() );
			return false;
		}
	}
}
