package com.tripgis.dmt.db;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;

import com.tripgis.dmt.file.FileManager;
import com.tripgis.dmt.types.EventRecord;
import com.tripgis.dmt.types.EventSubType;
import com.tripgis.dmt.types.EventType;
import com.tripgis.dmt.types.GeocodeRecord;
import com.tripgis.dmt.util.LatLng;
import com.tripgis.dmt.util.LogTranslator;
import com.tripgis.dmt.util.DmtUtils;
import com.tripgis.shared.types.LocationRecord;

public class DbModelEvents 
{
	// Name to use in log lines
    public static final String TAG = DbModelEvents.class.getSimpleName();
    public static final LogTranslator Log = new LogTranslator( TAG );

	private static final int IDX_DATE          = 1;
	private static final int IDX_LATLNG        = 2;
	private static final int IDX_TRIP_ID       = 3;
	private static final int IDX_DESCRIPTION   = 4;
	private static final int IDX_NOTES         = 5;
	private static final int IDX_PLACE_ID      = 6;
	private static final int IDX_EVENT_TYPE    = 7;
	private static final int IDX_EVENT_SUBTYPE = 8;
	private static final int IDX_STORE         = 9;
	private static final int IDX_ADDRESS       = 10;
	private static final int IDX_CITY          = 11;
	private static final int IDX_STATE         = 12;
	private static final int IDX_ZIP           = 13;
	private static final int IDX_COUNTRY       = 14;
	private static final int IDX_COUNTY        = 15;
	private static final int IDX_LOCATION_TYPE = 16;
	private static final int IDX_GEOMETRY      = 17;
	private static final int IDX_END_TIME      = 18;
	private static final int IDX_GEO_LATLNG    = 19;
	private static final int IDX_UTC_OFFSET    = 20;
	private static final int IDX_RATING        = 21;
	private static final int IDX_DURATION      = 22;
	private static final int IDX_EVENT_ID      = 23;

//	private static final int NUM_INSERT_FIELDS = 21;
//	private static final int NUM_SELECT_FIELDS = 23;
	
	public static final int DEFAULT_DURATION   = 1; // mins

	private static DbModel db;
    
	public DbModelEvents( DbModel db )
	{
		Log.t( TAG, "DbModelEvents()" );
		DbModelEvents.db = db;
	}

	/******************************************************************
	 * EVENT FUNCTIONS
	 *****************************************************************/

	public ArrayList<EventRecord> getEventsByTripId( int tripId )
	{
		Log.t( TAG, "getEventsByTripId( " + tripId + " )" );
		if( tripId > 0 )
			return getEventsWhere( "trip_id=" + tripId );
		return EventRecord.EMPTY_LIST;
	}
	
	private EventRecord buildEventRecord( ResultSet rs ) throws SQLException
	{
		LatLng ll = new LatLng( DmtUtils.getString( rs.getString( IDX_LATLNG ) ) );

		// MobilDataRecord : LocationRecord
		String description = db.getRsString( rs, IDX_DESCRIPTION );
		String notes = "";
		String title = description;
		if( rs.getString( IDX_NOTES ) != null )
		{
			notes = db.getRsString( rs, IDX_NOTES ); 
			title += FileManager.CSV_SEPARATOR + notes;
		}
		
		Log.d( TAG, "EventId: " + rs.getInt( IDX_EVENT_ID ) + ", Date: " + rs.getTimestamp( IDX_DATE ) );
		LocationRecord lr = new LocationRecord(
				db.getRsDate( rs, IDX_DATE ), ll, title, db.getRsDate( rs, IDX_END_TIME ) );
				
		GeocodeRecord gr = new GeocodeRecord(
				db.getRsString( rs, IDX_PLACE_ID ),
				db.getRsString( rs, IDX_LOCATION_TYPE ) );

		gr.setStore(       db.getRsString( rs, IDX_STORE      ) );
		gr.setAddress(     db.getRsString( rs, IDX_ADDRESS    ) );
		gr.setCity(        db.getRsString( rs, IDX_CITY       ) );
		gr.setState(       db.getRsString( rs, IDX_STATE      ) );
		gr.setZip(         db.getRsString( rs, IDX_ZIP        ) );
		gr.setCountry(     db.getRsString( rs, IDX_COUNTRY    ) );
		gr.setCounty(      db.getRsString( rs, IDX_COUNTY     ) );
		gr.setGeoLocation( db.getRsString( rs, IDX_GEO_LATLNG ) );
		gr.setUtcOffset(   db.getRsString( rs, IDX_UTC_OFFSET ) );
		gr.setRating(      db.getRsString( rs, IDX_RATING     ) );
		
		EventRecord er = new EventRecord( rs.getInt( IDX_EVENT_ID ), rs.getInt( IDX_TRIP_ID ) );
		er.setMobileData( lr );
		er.setGeoData( gr );
		er.setDescription( description );
		er.setNotes( notes );
		er.setEventSubType( db.getTypes().getEventSubType( 
				db.getRsString( rs, IDX_EVENT_TYPE ),
				db.getRsString( rs, IDX_EVENT_SUBTYPE ) ) );
		er.setGeometry( DmtUtils.getString( rs.getString( IDX_GEOMETRY ) ) );
		er.setDuration( rs.getDouble( IDX_DURATION ) );
		return er;
	}
	
	public EventRecord getEventByDate( int tripId, String date )
	{
		Log.t( TAG, "getEventByDate(" + tripId + ", " + date + " )" );
		ArrayList<EventRecord> list = getEventsWhere( "trip_id=" + tripId + " and date = '" + date + "'" );
		if( list.size() > 0 )
			list.get(0);
		return EventRecord.EMPTY;
	}

	public EventRecord getEventById( int id )
	{
		Log.t( TAG, "getEventById(" + id + " )" );
		if( id == 0 )
			return EventRecord.EMPTY;

		return getEventsWhere( "event_id=" + id ).get(0);
	}
	
	public ArrayList<EventRecord> getEventsByType( int tripId, EventType et )
	{
		Log.t( TAG, "getEventsByType(" + tripId + ", " + et + " )" );
		return getEventsWhere( "trip_id=" + tripId + " and type='" + et.getName() + "'" );
	}
	
	public ArrayList<EventRecord> getEventsBySubType( int tripId, EventSubType est )
	{
		Log.t( TAG, "getEventsBySubType(" + tripId + ", " + est + " )" );
		return getEventsWhere( "trip_id=" + tripId + " and subtype='" + est.getName() + "'" );
	}
	
	private ArrayList<EventRecord> getEventsWhere( String sql )
	{
		Log.t( TAG, "getEventsWhere( " + sql + " )" );
		
		Statement st = null;
		ResultSet rs = null;

		try
		{
			Connection conn = db.getDbConnection();
			st = conn.createStatement();
			rs = st.executeQuery( "SELECT " + getSelectFields() + " FROM events where " +
					sql + " order by date" );

			ArrayList<EventRecord> events = new ArrayList<>();
			while( rs.next() )
				events.add( buildEventRecord( rs ) );
			
			Log.d( TAG, "Number of Events = " + events.size() );
			return events;
		}
		catch( SQLException e )
		{
			db.closeQuery( rs, st );
			Log.e( TAG, "Failed to get event - " + e );
		}
		
		return EventRecord.EMPTY_LIST;
	}
		
	public EventRecord insertEvent( int tripId, String name )  throws SQLException
	{
		EventRecord er = new EventRecord( tripId, 0 );
		er.getMobileData().setName( name );
		return insertEvent( er ); 
	}
	
	private EventRecord insertEvent( EventRecord er ) throws SQLException
	{
		Log.t( TAG, "insertEvent( " + er + " )" );
		if( er.getEventId() != 0 )
			return updateEvent( er );

		else 
		{
			String date = LocationRecord.getSqlDateString( er.getMobileData().getDate() );
			EventRecord dbEvent = getEventByDate( er.getTripId(), date );
			if( dbEvent != EventRecord.EMPTY )
			{
				Log.d( TAG, "matched date '" + date + "'!" );
				er.setEventId( dbEvent.getEventId() );
				return updateEvent( er );
			}
		}
		
		String sql = "insert into events (" + getInsertFields() + ") " +
				"values ( " + getInsertText( er ) + " )";
		Statement stmt = db.executeStatement( sql );
		stmt.close();
		return getEventById( db.getLastRecordId( "events", "event_id" ) );
	}
	
	private String getSelectFields()
	{
		return getInsertFields() + ", event_id";
	}
	
	private String getInsertFields()
	{
		return "date, latlng, trip_id, description, notes, " +
	       "place_id, type, subtype, store, address, city, state, zip, " +
		   "country, county, location_type, geom, end_time, geolatlng, " +
	       "utc_offset, rating, duration";
	}
	
	
	private String getInsertText( EventRecord er )
	{
		StringBuffer sb = new StringBuffer();
		sb.append( "'" + LocationRecord.getSqlDateString( er.getMobileData().getDate() ) + "', " );
		sb.append( "'" + er.getMobileData().getLocation().getDisplayString() + "', " );
		sb.append( er.getTripId() + ", " );
		sb.append( db.toSqlString( er.getDescription() ) + ", " );
		sb.append( db.toSqlString( er.getNotes() ) + ", " );
		sb.append( db.toSqlString( er.getGeoData().getPlaceId() ) + ", " );
		sb.append( db.toSqlString( er.getEventSubType().getType().getName() ) + ", " );
		sb.append( db.toSqlString( er.getEventSubType().getName() ) + ", " );
		sb.append( db.toSqlString( er.getGeoData().getStore() ) + ", " );
		sb.append( db.toSqlString( er.getGeoData().getAddress() ) + ", " );
		sb.append( db.toSqlString( er.getGeoData().getCity() ) + ", " );
		sb.append( db.toSqlString( er.getGeoData().getState() ) + ", " );
		sb.append( db.toSqlString( er.getGeoData().getZip() ) + ", " );
		sb.append( db.toSqlString( er.getGeoData().getCountry() ) + ", " );
		sb.append( db.toSqlString( er.getGeoData().getCounty() ) + ", " );
		sb.append( db.toSqlString( er.getGeoData().getLocationType() ) + ", " );
		sb.append( "ST_SetSRID(ST_MakePoint(" + 
				er.getMobileData().getLocation().getLongitudeString() + 
				"," + er.getMobileData().getLocation().getLatitudeString() + "), 4326), " );
		if( er.getMobileData().getEndTime() != null )
			sb.append( "'" + LocationRecord.getSqlDateString( er.getMobileData().getEndTime() ) + "', " );
		else
			sb.append( "null, " );
		
		sb.append( "'" + er.getGeoData().getGeoLocation().getDisplayString() + "', " );
		sb.append( "'" + er.getGeoData().getUtcOffset() + "', " );
		sb.append( "'" + er.getGeoData().getRating() + "', " );
		sb.append( "'" + er.getDuration() + "'" );
		return sb.toString();
	}

	public EventRecord updateEvent( EventRecord er ) throws SQLException
	{
		Log.t( TAG, "updateEvent( " + er + " )" );
		if( er.getEventId() == 0 )
			return insertEvent( er );

		String sql = "update events set " + getUpdateText( er ) + 
				" where event_id = " + er.getEventId();
		Statement stmt = db.executeStatement( sql );
		stmt.close();
		return er;
	}

	private String getUpdateText( EventRecord er )
	{
		StringBuffer sb = new StringBuffer();
		sb.append( "date='" + LocationRecord.getSqlDateString( er.getMobileData().getDate() ) + "', " );
		sb.append( "latlng='" + er.getMobileData().getLocation().getDisplayString() + "', " );
		sb.append( "trip_id=" + er.getTripId() + ", " );
		sb.append( "description=" + db.toSqlString( er.getDescription() ) + ", " );
		sb.append( "notes=" + db.toSqlString( er.getNotes() ) + ", " );
		sb.append( "place_id=" + db.toSqlString( er.getGeoData().getPlaceId() ) + ", " );
		sb.append( "type=" + db.toSqlString( er.getEventSubType().getType().toString() ) + ", " );
		sb.append( "subtype=" + db.toSqlString( er.getEventSubType().getName() ) + ", " );
		sb.append( "store=" + db.toSqlString( er.getGeoData().getStore() ) + ", " );
		sb.append( "address=" + db.toSqlString( er.getGeoData().getAddress() ) + ", " );
		sb.append( "city=" + db.toSqlString( er.getGeoData().getCity() ) + ", " );
		sb.append( "state=" + db.toSqlString( er.getGeoData().getState() ) + ", " );
		sb.append( "zip=" + db.toSqlString( er.getGeoData().getZip() ) + ", " );
		sb.append( "country=" + db.toSqlString( er.getGeoData().getCountry() ) + ", " );
		sb.append( "county=" + db.toSqlString( er.getGeoData().getCounty() ) + ", " );
		sb.append( "location_type=" + db.toSqlString( er.getGeoData().getLocationType() ) + ", " );
		sb.append( "geom=ST_SetSRID(ST_MakePoint(" + 
				er.getMobileData().getLocation().getLongitudeString() + 
				"," + er.getMobileData().getLocation().getLatitudeString() + "), 4326), " );
		if( er.getMobileData().getEndTime() != null )
			sb.append( "end_time='" + LocationRecord.getSqlDateString( er.getMobileData().getEndTime() ) + "', " );

		sb.append( "geolatlng='" + er.getGeoData().getGeoLocation().getDisplayString() + "', " );
		sb.append( "utc_offset='" + er.getGeoData().getUtcOffset() + "', " );
		sb.append( "rating='" + er.getGeoData().getRating() + "', " );
		sb.append( "duration=" + er.getDuration() );
		return sb.toString();
	}

	public boolean removeEvent( int eventId ) throws SQLException
	{
		Log.t( TAG, "removeEvent( " + eventId + " )" );
		if( eventId <= 0 )
			throw new SQLException( "Event " + eventId + " does not exist!" );
		
		String sql = "delete from events where event_id = " + eventId;
		Statement stmt = db.executeStatement( sql );
		stmt.close();
		return true;
	}	

	
	private int updateEvents( ArrayList<EventRecord> list ) throws SQLException
	{
		Log.t( TAG, "updateEvent( " + list.size() + " )" );

		int count = 0;
		for( EventRecord event : list )
		{
			double duration = DmtUtils.getDuration( event.getMobileData().getDate(), event.getMobileData().getEndTime() );
			duration /= 60;
			String sql = "update events set duration=" + duration + " where event_id=" + event.getEventId();
			try
			{
				Statement stmt = db.executeStatement( sql );
				stmt.close();
				count++;
			}
			catch( SQLException e )
			{
				throw new SQLException( (count+1) + ": Failed to update event[" + event.getEventId() + "] - " + e );
			}
		}
		
		Log.d( TAG, "Items affected: " + count );
		return count;
	}
	
	
	/**
	 * update events set duration=(select (extract(epoch from end_time)-extract(epoch from date))/60 from events where event_id=377) where event_id=377;
	 * Set end time to 5 sec if not specified (arbitrary)
	 * @throws SQLException 
	 */
	public int updateDuration( int tripId ) throws SQLException
	{
		Log.t( TAG, "updateDuration( " + tripId + " )" );
		ArrayList<EventRecord> list = getEventsByTripId( tripId );
		return updateEvents( list );
	}
	
	public static final double MI_TO_KM = 1609.34;
	public static final double HALF_MILE = 1609.34 * 0.5;
	
	/**
	 * SELECT event_id, description, date FROM events WHERE GeometryType(ST_Centroid(geom)) = 'POINT' AND ST_Distance_Sphere( ST_Point(ST_X(ST_Centroid(geom)), ST_Y(ST_Centroid(geom))), (ST_MakePoint(-111.9428, 33.424945))) <= 1000 and trip_id=5;
	 * 
	 * References:
	 *    https://gis.stackexchange.com/questions/77072/return-all-results-within-a-30km-radius-of-a-specific-lat-long-point
	 *    
	 * @param point - position to center around
	 * @param radius - in meters
	 * @return
	 * @throws SQLException
	 */
	public ArrayList<EventRecord> getEventsWithinRange( int tripId, LatLng point, double radius ) throws SQLException
	{
		Log.t( TAG, "getEventsWithinRange(" + tripId + ", " + point + ", " + radius + " )" );
		String whereClause = "trip_id = " + tripId + " and GeometryType(ST_Centroid(geom)) = 'POINT' AND " +
				"ST_Distance_Sphere( ST_Point(ST_X(ST_Centroid(geom)), ST_Y(ST_Centroid(geom))), " +
				"(ST_MakePoint(" + point.longitude + "," + point.latitude + "))) <=" + radius;
		return getEventsWhere( whereClause );
	}

	public ArrayList<EventRecord> getEventsWithinRange( EventRecord event, double radius ) throws SQLException
	{
		return getEventsWithinRange( event.getTripId(), event.getMobileData().getLocation(), radius );
	}
}
