package com.tripgis.dmt.db;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;

import com.tripgis.dmt.DmtController;
import com.tripgis.dmt.types.TripDbRecord;
import com.tripgis.dmt.util.LogTranslator;

public class DbModelTrips 
{
	// Name to use in log lines
    public static final String TAG = DbModelTrips.class.getSimpleName();
    public static final LogTranslator Log = new LogTranslator( TAG );
    
    private DbModel db;
    
    private boolean tripsChanged;

	private ArrayList<TripDbRecord> trips;
    
	public DbModelTrips( DbModel db )
	{
		Log.t( TAG, "DbModelTrips()" );
		this.db = db;
	}

	/******************************************************************
	 * TRIP FUNCTIONS
	 *****************************************************************/

	public boolean isTripsValid()
	{
		Log.d( TAG, "tripsChange=" + tripsChanged + ", size=" + trips.size() );
		if( !tripsChanged && trips.size() > 0 )
			return true;
		
		return false;
	}
	
	public ArrayList<TripDbRecord> getTrips()
	{
		Log.t( TAG, "getTrips()" );
		trips = new ArrayList<>();
		try
		{
			Connection conn = db.getDbConnection();
			if( isTripsValid() )
				return trips;
			
			Statement st = conn.createStatement();
			ResultSet rs = st.executeQuery( "SELECT trip_id, name, source, file FROM trips order by name" );
			
			while( rs.next() )
			{
				TripDbRecord trip = new TripDbRecord( rs.getInt(1), db.getRsString( rs, 2 ) );
				trip.setSource( rs.getString(3) );
				trip.setFileName( db.getRsString( rs, 4 ) );
				trips.add( trip );
			}
	
			db.closeQuery( rs, st );
			Log.d( TAG, "Number of Trips = " + trips.size() );
		}
		catch( SQLException e )
		{
			DmtController.setError( TAG, "Failed to get trip list - " + e );
		}
		
		return trips;
	}
	
	public TripDbRecord getTripById( int id )
	{
		Log.t( TAG, "getTripById( " + id + " )" );
		getTrips();
		for( TripDbRecord trip : trips )
		{
			if( trip.getTripId() == id )
				return trip;
		}

		Log.d( TAG, "Trip " + id + " not found!" );
		return TripDbRecord.EMPTY;
	}
	
	public TripDbRecord getTripByName( String name )
	{
		Log.t( TAG, "getTripByName( " + name + " )" );
		getTrips();
		for( TripDbRecord trip : trips )
		{
			Log.d( TAG, "comparing " + trip.getTripName() + " to " + name );
			if( trip.getTripName().equals( name ) )
				return trip;
		}
		Log.d( TAG, "Trip " + name + " not found!" );
		return TripDbRecord.EMPTY;
	}
		
	public TripDbRecord insertTrip( String name )  throws SQLException
	{
		TripDbRecord tr = TripDbRecord.EMPTY.copy();
		tr.setTripName( name );
		return insertTrip( tr ); 
	}
	
	private TripDbRecord insertTrip( TripDbRecord tr ) throws SQLException
	{
		Log.t( TAG, "insertTrip( " + tr + " )" );
		if( tr.getTripId() != 0 )
			return updateTrip( tr );

		TripDbRecord dbTrip = getTripByName( tr.getTripName() );
		if( dbTrip != TripDbRecord.EMPTY )
		{
			Log.d( TAG, "matched name '" + dbTrip.getTripName() + "'!" );
			tr.setTripId( dbTrip.getTripId() );
			return updateTrip( tr );
		}

		String sql = "insert into trips ( " + getFields() + 
				" ) values( " + getInsertText( tr ) + " )";
		Statement stmt = db.executeStatement( sql );
		stmt.close();
		tripsChanged = true;
		return getTripById( db.getLastRecordId( "trips", "trip_id" ) );
	}
	
	public String getFields()
	{
		return "name, source, file";
	}
		
	public String getInsertText( TripDbRecord tr )
	{
		StringBuffer sb = new StringBuffer();
		sb.append( db.toSqlString( tr.getTripName() ) + ", " );
		sb.append( db.toSqlString( tr.getSource() ) + ", " );
		sb.append( db.toSqlString( tr.getFileName() ) );
		return sb.toString();
	}

	public TripDbRecord updateTrip( TripDbRecord tr ) throws SQLException
	{
		Log.t( TAG, "updateTrip( " + tr + " )" );
		if( tr.getTripId() == 0 )
			return insertTrip( tr );

		String sql = "update trips set " + getUpdateText( tr ) + 
				" where trip_id = " + tr.getTripId();
		Statement stmt = db.executeStatement( sql );
		stmt.close();
		tripsChanged = true;
		return tr;
	}

	public String getUpdateText( TripDbRecord tr )
	{
		StringBuffer sb = new StringBuffer();
		sb.append( "name=" + db.toSqlString( tr.getTripName() ) + ", " );
		sb.append( "source=" + db.toSqlString( tr.getSource() ) + ", " );
		sb.append( "file=" + db.toSqlString( tr.getFileName() ) );
		return sb.toString();
	}
}
