package com.tripgis.dmt.db;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;

import com.tripgis.dmt.DmtController;
import com.tripgis.dmt.types.EventSubType;
import com.tripgis.dmt.types.EventType;
import com.tripgis.dmt.util.LogTranslator;
import com.tripgis.dmt.util.DmtUtils;

public class DbModelTypes 
{
	// Name to use in log lines
    public static final String TAG = DbModelTypes.class.getSimpleName();
    public static final LogTranslator Log = new LogTranslator( TAG );

    private DbModel db;
    
    private boolean eventTypesChanged;
    private boolean eventSubtypesChanged;

	private ArrayList<EventType> eventTypes;
	private ArrayList<EventSubType> eventSubTypes;
    
	public DbModelTypes( DbModel db )
	{
		Log.t( TAG, "DbModelTypes()" );
		this.db = db;
	}

	/******************************************************************
	 * EVENT_TYPE FUNCTIONS
	 *****************************************************************/

	public ArrayList<EventType> getEventTypes()
	{
		Log.t( TAG, "getEventTypes()" );
		try
		{
			Connection conn = db.getDbConnection();
			if( isEventTypesValid() )
				return eventTypes;
			
			Log.d( TAG, "Rebuilding Event Types..." );
			Statement st = conn.createStatement();
			ResultSet rs = st.executeQuery( "SELECT name FROM event_types order by name" );
			
			eventTypes = new ArrayList<>();
			while( rs.next() )
				eventTypes.add( new EventType( DmtUtils.getString( rs.getString( 1 ) ) ) );
	
			rs.close();
			st.close();
			
			Log.d( TAG, "Total Number of Types = " + eventTypes.size() );
			eventTypesChanged = false;
			return eventTypes;
		}
		catch( SQLException e )
		{
			DmtController.setError( TAG, "Failed to get event types - " + e );
		}

		return EventType.EMPTY_LIST;
	}
	
	public EventType getEventType( String name )
	{
		Log.t( TAG, "getEventType( " + name + " )" );
		if( name.isEmpty() )
			return EventType.EMPTY;
		
		getEventTypes();
		for( EventType type : eventTypes )
		{
			if( type.getName().equals( name ) )
				return type;
		}
		
		return EventType.EMPTY;
	}
	
	public boolean isEventTypesValid()
	{
		if( eventTypesChanged )
			return false;
		else if( eventTypes == null )
			return false;
		
		return true;
	}
	
	public boolean isEventSubTypesValid()
	{
		if( eventSubtypesChanged )
			return false;
		else if( eventSubTypes == null )
			return false;
		
		return true;
	}
	
	public ArrayList<EventSubType> getEventSubTypesByType( EventType type )
	{
		Log.t( TAG, "getEventSubTypeByType( " + type.getName() + " )" );
		if( type == EventType.EMPTY )
			return EventSubType.EMPTY_LIST;
		
		getEventSubTypes();
		
		ArrayList<EventSubType> list = new ArrayList<>();
		for( EventSubType subtype : eventSubTypes )
		{
			if( subtype.getType() == type )
				list.add( subtype );
		}
		
		Log.d( TAG, "Type: " + type + ", Number of Subtypes: " + list.size() );
		return list;
	}
	
	public ArrayList<EventSubType> getEventSubTypes( String name )
	{
		return getEventSubTypesByType( getEventType( name ) );
	}
	
	public EventSubType getEventSubType( String type, String name ) throws SQLException
	{
		Log.t( TAG, "getEventSubType( " + type + ", " + name + " )" );

		if( type.isEmpty() && name.isEmpty() )
			return EventSubType.EMPTY;
		
		ArrayList<EventSubType> list = getEventSubTypes( type );
		for( EventSubType subtype : list )
		{
			//Log.d( TAG, "matching " + subtype.getName() + " to " + name );
			if( subtype.getName().equals( name ) )
			{
				//Log.d( TAG, "matched " + name );
				return subtype;
			}
		}
		
		return EventSubType.EMPTY;
	}
	
	public ArrayList<EventSubType> getEventSubTypes()
	{
		Log.t( TAG, "getEventSubTypes()" );
		try
		{
			if( isEventSubTypesValid() )
				return eventSubTypes;
			
			Log.d( TAG, "Rebuilding Event Subtypes..." );
			// This will force updating types if they are out of date
			// This also calls checkConnection()
			getEventTypes();

			Connection conn = db.getDbConnection();
			Statement st = conn.createStatement();
			ResultSet rs = st.executeQuery( "SELECT event_type, name FROM event_subtypes order by event_type, name" );
			
			eventSubTypes = new ArrayList<>();
			while( rs.next() )
			{
				EventType type = getEventType( DmtUtils.getString( rs.getString( 1) ) );
				eventSubTypes.add( new EventSubType( type, DmtUtils.getString( rs.getString( 2) ) ) );
			}
	
			rs.close();
			st.close();
			
			Log.d( TAG, "Total Number of Subtypes: " + eventSubTypes.size() );
			eventSubtypesChanged = false;
			return eventSubTypes;
		}
		catch( SQLException e )
		{
			DmtController.setError( TAG, "Failed to get event subtype list - " + e );
		}
		
		return EventSubType.EMPTY_LIST;
	}
	
	public EventType insertEventType( String name )  throws SQLException
	{ return insertEventType( new EventType( name ) ); }
	
	public EventType insertEventType( EventType type ) throws SQLException
	{
		Log.t( TAG, "insertEventType( " + type + " )" );
		String sql = "insert into event_types values( '" + type.getName() + "' )";
		Statement stmt = db.executeStatement( sql );
		stmt.close();
		eventTypesChanged = true;
		return type;
	}
	
	public EventSubType insertEventSubType( EventType type, String name ) throws SQLException
	{ return insertEventSubType( new EventSubType( type, name ) ); }
	
	public EventSubType insertEventSubType( EventSubType type ) throws SQLException
	{
		Log.t( TAG, "insertEventSubType( " + type + " )" );
		String sql = "insert into event_subtypes ( name, event_type ) values " +
				"( '" + type.getName() + "', '" + type.getType() + "' )";
		Statement stmt = db.executeStatement( sql );
		stmt.close();
		eventSubtypesChanged = true;
		return type;
	}
}
