package com.tripgis.dmt.file;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.sql.Date;
import java.util.ArrayList;

import com.tripgis.dmt.types.EventRecord;
import com.tripgis.dmt.util.ConversionUtilities;
import com.tripgis.dmt.util.DmtProperties;
import com.tripgis.dmt.util.LogTranslator;
import com.tripgis.dmt.util.DmtUtils;
import com.tripgis.shared.types.LocationRecord;

/**
 * References:
 *    https://developer.android.com/guide/topics/data/data-storage.html#filesInternal
 *    https://docs.oracle.com/javase/7/docs/api/java/io/File.html
 *    https://www.mkyong.com/java/how-to-export-data-to-csv-file-java/
 *    https://docs.oracle.com/javase/tutorial/java/generics/types.html
 *    https://stackoverflow.com/questions/18783916/check-if-a-generic-t-implements-an-interface
 *    http://docs.oracle.com/javase/tutorial/extra/generics/methods.html
 *
 * Created by pk on 7/3/17.
 */

public abstract class FileManager<H,T>
{
    protected File originalFile;
    protected File workingFile;
    protected H header;
    protected ArrayList<EventRecord> list;
    protected String lastError;
    
    protected boolean overWriteOriginal = false;

    // Name to use in log lines
    public static final String TAG = FileManager.class.getSimpleName();
    public static final LogTranslator Log = new LogTranslator( TAG );

    // Directory to store all files used by application
    protected static final String BASE_DIR = DmtProperties.getProperty( "file.basedir" );

    // Line separator to use for each stored record
    public static final String CSV_SEPARATOR = DmtProperties.getProperty( "file.separator" );;

    // Text to replace line separators
    public static final String CSV_SUBSTITUTION = DmtProperties.getProperty( "file.substitution" );;

    /**
     * Three parameters may be sent:
     *    <original> - originalFile to read from (always read-only)
     *    <name> - string to use as the base for workingFile
     *    		override any or all of the following functions
     *    		getPathname() + stripPostfix( <name> ) + 
     *    				getFilenamePostfix() + getExtension()
     *			DEFAULT: <original>.getName() [java.io.File]
     *    <overwrite> - boolean to indicate workingFile should be overwritten
     *    		DEFAULT: true
     */
	public FileManager( String name )
    { this( null, name ); }

	public FileManager( File original )
    { this( original, original.getName() ); }

	public FileManager( File original, String name )
    { this( original, name, false ); }

    public FileManager( File original, String name, boolean forceRead )
    {
        Log.d( TAG, "FileManager( " + original + ", " + name + ", " + forceRead + " )" );
        lastError = null;
        
        if( original == null ) 
        	originalFile = getOriginalFilename( name );
        else
        	originalFile = original;

        workingFile = getWorkingFilename( name );
        Log.d( TAG, "Original File: " + originalFile );
        Log.d( TAG, " Working File: " + workingFile );
    	if( !read( forceRead ) )
    		return; // header and list will be null

        try 
        {
			prepareOriginalFile();

			// createNewFile() only creates a file if it does not exist
	        //    and we don't care if the file existed already at this point.
			if( workingFile.createNewFile() )
				Log.i( TAG, "Created " + workingFile );

			// Always write data (forceRead allows using working data)
			write( true );
        }
        catch (IOException e) 
        {
        	lastError = "Failed to initialize - " + e.getMessage();
        	return; // header and list will be null if write failed
		}

        Log.i( TAG, "Current storage file: " + workingFile + " [" + DmtUtils.prettySize( workingFile.length() ) + "]" );
        lastError = "";
    }
    
    private File getOriginalFilename( String name )
    {
    	name = stripPostfix( name );
        name += getExtension();
        return new File( getPathname(), name );
    }
    
    protected File getWorkingFilename( String name )
    {
    	name = stripPostfix( name );
        name += getFilenamePostfix() + getExtension();
        return new File( getPathname(), name );
    }
    
    private void prepareOriginalFile() throws IOException
    {
    	Log.t( TAG, "prepareOriginalFile()" );

    	// The only time we should ever overwrite the original file is if 
		// it does not exist in the BASE_DIR - we are making a copy
        if( !originalFile.getPath().contains( FileManager.BASE_DIR + "$" ) )
        {
        	Log.i( TAG, "originalFile not in base directory, copying..." );
        	originalFile = new File( getPathname(), originalFile.getName() );
        	if( originalFile.exists() )
        		Log.w( TAG, "File already exists in base dir, skipping" );

        	else
        	{
	        	originalFile.createNewFile();
	        	overWriteOriginal = true;
	        	write( originalFile );
	        	overWriteOriginal = false;
        	}
        }
    }
    
    /************************************************************
     * FILENAME MODIFIER FUNCTIONS (Override as needed)
     ***********************************************************/

    protected File getPathname() 
    { 
    	try
    	{
	        return getAlbumStorageDir();
    	}
    	catch( IOException e )
    	{
    		lastError = "Failed to build path - " + e;
    		e.printStackTrace();
    		// this is catastrophic...as it should be
    		return null;
    	}
    }

    protected String getFilenamePostfix()
    {
        return "_" + DmtUtils.getTimestamp( LocationRecord.getCurrentTime() );
    }

    protected String stripPostfix( String name )
    {
    	return stripBasePostfix( name );
    }
    
    protected static String stripBasePostfix( String name )
    {
        int idx = name.indexOf( "_" );
        if( idx < 0 )
        {
        	Log.w( TAG, "No postfix found, ignoring request." );
        	return DmtUtils.removeExtension( name );
        }
        
        String base = name.substring( 0, idx );
        if( base.matches( "\\d+" ) )
        {
        	Log.d( TAG, "base '" + base + "' is a timestamp, taking next index..." );
        	idx = name.indexOf( "_", idx + 1 );
            if( idx < 0 )
            {
            	Log.w( TAG, "No postfix found, ignoring request." );
            	return DmtUtils.removeExtension( name );
            }
            
            Log.d( TAG, "Better? " + name.substring( 0, idx ) );
        }
        
        return name.substring( 0, idx );
    }

    protected String getExtension() { return "." + DmtUtils.csv; }
    
    /************************************************************
     * REQUIRED OVERRIDE FUNCTIONS
     ***********************************************************/

    public File getOriginalFile() { return originalFile; }
    public File getWorkingFile() { return workingFile; }
    public ArrayList<EventRecord> getList() { return list; }
    public H getHeader() { return header; }
    public String getLastError() { return lastError; }
    
    public boolean matches( EventRecord a, EventRecord b )
    {
    	if( a.getEventId() != 0 )
    		return a.getEventId() == b.getEventId();
    	
    	return a.getMobileData().getDate() == b.getMobileData().getDate();
    }

    // Conversion tools
    protected abstract EventRecord itemToEvent( T t );
    protected abstract T eventToItem( EventRecord e );
    
    // Read from a file (CSV assumed, not required)
    // use parseLine( lineReader ) to get String[] for header
    protected abstract H readHeader( BufferedReader lineReader ) throws IOException;
    protected abstract EventRecord readListItem( String[] data ) throws IOException;
    
    // Write to the same file
    protected abstract boolean writeHeader( FileWriter writer, H h ) throws IOException;
    protected abstract String getListItemString( EventRecord e );
    
    /************************************************************
     * GENERIC GET/SET FUNCTIONS
     ***********************************************************/

    public static boolean isFileCreated( String name )
    {
    	try 
    	{
			File file = new File( getAlbumStorageDir().getAbsolutePath(), name );
			if( file.exists() )
				return true;
    	}
    	catch (IOException e) 
    	{
    		Log.e( TAG, "Failed to create filename for " + name );
		}

    	return false;
    }
    
    /**
     * Attempts to open the DIRECTORY_DOCUMENTS public storage area. Creates
     * a directory under it called BASE_DIR.
     *
     * @return a File object containing the absolute path of the directory.
     */
    public static File getAlbumStorageDir() throws IOException
    {
        Log.d( TAG, "getAlbumStorageDir()" );
        File path = new File( ConversionUtilities.getDocumentsDirectory(), BASE_DIR );

        if( !path.exists() )        // No need to create the path if it already exists
        {
            if (!path.mkdirs())     // Create all necessary directories (equiv. to mkdir -p)
                throw new IOException("Directory '" + path + "' not created");
        }

        // Return the desired path of the stored file
        Log.i( TAG, "Device media stoage state: " + ConversionUtilities.getStorageState() );
        Log.d( TAG, "Total: " + DmtUtils.prettySize( path.getTotalSpace() ) +
                ", Free: " + DmtUtils.prettySize( path.getFreeSpace() ) +
                ", Usable: " + DmtUtils.prettySize( path.getUsableSpace() ) );
        return path;
    }

    /************************************************************
     * READ/WRITE FUNCTIONS
     ***********************************************************/

    /**
     * The first time this is called (at initialization), list will be
     * null. A byproduct of the read action is an empty list being 
     * produced if the file happens to be new.
     * 
     * @param forceRead - force reading of the working file
     * @return - true if read succeeded
     */
    protected boolean read( boolean forceRead )
    {
    	Log.t( TAG, "read( " + forceRead + " )" );
    	
    	Log.d( TAG, "List: " + list + ", Original Exists: " + originalFile.exists() );
    	// Only read original if list is not defined
    	if( list == null && originalFile.exists() )
    	{
        	Log.d( TAG, "Working Exists: " + workingFile.exists() );
    		// Do not read original if working exists and forceRead not set
    		if( !workingFile.exists() || !forceRead )
    			return read( originalFile );
    	}

    	return read( workingFile );
    }
    
    protected boolean write( boolean overwrite )
    {
    	if( workingFile.exists() && !overwrite )
    	{
    		lastError = "File " + workingFile + " exists, use overwrite flag to force";
    		return false;
    	}
    	
    	return write( workingFile );
    }

    /**
     * Reads the stored file data, parses it, and returns a list of Record found.
     *
     * References:
     *    https://www.mkyong.com/java/how-to-read-and-parse-csv-file-in-java/
     *    https://stackoverflow.com/questions/12395862/tokenising-a-string-containing-empty-tokens
     *
     * @return - true if (partial) success, false if complete failure
     */
    protected boolean read( File file )
    {
        Log.d( TAG, "read( " + file + " )" );
        int count = 0;

        try
        {
            // Prepare the file for reading
            FileReader fileReader = new FileReader( file );
            BufferedReader lineReader = new BufferedReader( fileReader );

            list = new ArrayList<>();
            header = readHeader( lineReader );
            
        	count++;
        	String[] data = parseLine( lineReader );
        	while( data != null )
        	{
        		EventRecord e = readListItem( data );
        		if( e != null && e != EventRecord.EMPTY )
        			list.add( e );
        		
            	count++;
        		data = parseLine( lineReader );
        	}
        	
        	lineReader.close();
        	fileReader.close();

            Log.d( TAG, "Header: " + header );
            Log.d( TAG, "List size: " + list.size() );
            return true;
        }
        catch( FileNotFoundException e )
        {
            lastError = "Read failed on '" + file + "': " + e;
            return false;
        }
        catch( IOException e )
        {
            lastError = "Item " + count + ": Read failed on '" + file + "': " + e;
            return false;
        }
    }
    
    protected String[] parseLine( BufferedReader lineReader ) throws IOException
    {
        String line = lineReader.readLine();
        if( line == null )
        {
        	lastError = "Empty line";
        	return null;
        }
        
        return line.split( CSV_SEPARATOR, -1 );
    }

    /**
     * Opens the stored file for writing and adds a new line for each record.
     * Data can be appended or the file can be overwritten each time.
     *
     * @return true if complete success
     */
    protected boolean write( File file )
    {
        Log.t( TAG, "write( " + file + " )" );
        Log.d( TAG, "File: " + file.toString() + ", Header: " + header + ", Items: " + list.size() );

        int lineCount = 0;      // Number of lines written
        try 
        {
	        for( EventRecord record : list )
	        {
	            String text = getListItemString( record );
	            if( !text.matches( "\\A\\p{ASCII}*\\z" ) )
	                throw new IOException( "Line " + lineCount + ": Record contains non-ASCII - '" + text + "'" );
	        }
        
            FileWriter writer = new FileWriter( file );   // Writer object
	        if( writeHeader( writer, header ) )
	        	lineCount++;
	
	        // Iterate through each record in the passed list
	        for( EventRecord record : list )
	        {
	            lineCount++;                                   // Increment line count (for logging)
	            if( lineCount < 5 )
	                Log.d( TAG, "Writing " + record );
	            writer.append( getListItemString( record ) + "\n" );  // Write out the data
	        }
	
	        writer.flush();
			writer.close();
	
	        // Return the number of lines written
	        Log.i( TAG, "Lines written: " + lineCount );
	        return lineCount != 0;
		} 
        catch( IOException e )
        {
        	lastError = "Write failed, line " + lineCount + " - " + e;
        	return false;
		}
    }
    
    public boolean addItem( EventRecord e )
    {
    	Log.t( TAG, "addItem( " + e + " )" );
    	if( getList().add( e ) )
    		return write( true );
    	return true;
    }
    
    public boolean replaceItem( EventRecord e )
    {
    	Log.t( TAG, "replaceItem( " + e + " )" );

    	for( int i = 0; i < list.size(); i++ )
    	{
    		EventRecord oldItem = list.get(i);
    		if( matches( oldItem, e ) )
    		{
    			Log.d( TAG, "Found match for '" + e + "'" );
    			list.remove( i );
    			list.add( i, e );
        		return write( true );
    		}
    	}

    	return false;
    }

    /**
     * Because the date is used as a secondary primary key (to event id),
     * it is necessary to change it after the match.
     * 
     * @param d - the new date
     * @param e - the record being replaced
     * @return - true if item found/replaced
     * @throws IOException
     */
    public boolean replaceItemDate( Date d, EventRecord e )
    {
    	Log.t( TAG, "replaceItem( " + d + " , " + e + " )" );

    	for( int i = 0; i < list.size(); i++ )
    	{
    		EventRecord oldItem = list.get(i);
    		if( matches( oldItem, e ) )
    		{
    			Log.d( TAG, "Found match for '" + e + "'" );
    			list.remove( i );
    			e.getMobileData().setDate( d );
    			list.add( i, e );
        		return write( true );
    		}
    	}

    	return false;
    }

    public static String prepareString( String text )
    {
    	if( text == null ) return "";
        return text.replaceAll( CSV_SEPARATOR, CSV_SUBSTITUTION );
    }

    public static String readString( String text )
    {
        return text.replaceAll( CSV_SUBSTITUTION, CSV_SEPARATOR );
    }

    public boolean changeFile( String name, boolean overwrite ) throws IOException
    {
    	Log.t( TAG, "changeFile( " + name + ", " + overwrite + " )" );
    	lastError = "Function not implemented!";
    	return false;
    }
    
    /**
     * Deletes the file from the filesystem. 
     *
     * @return the results of the delete action (true is good)
     */
    public boolean deleteFile( boolean cleanAll )
    {
        // Attempt to delete the file
        boolean rc = workingFile.delete();
        
        if( cleanAll )
        	deleteAllFiles();

        // No need to continue processing if delete failed
        if( rc )
            Log.i( TAG, "Successfully removed '" + workingFile + "'" );
        else
            lastError = "Failed to delete file '" + workingFile + "'";

        // Return the results of the delete operation
        return rc;
    }
    
    /**
     * References:
     *    https://stackoverflow.com/questions/8646984/how-to-list-files-in-an-android-directory
     *    https://stackoverflow.com/questions/924394/how-to-get-the-filename-without-the-extension-in-java
     *
     * @return
     * @throws IOException
     */
    public void deleteAllFiles()
    {
        Log.d( TAG, "deleteAllFiles()" );

		try 
		{
			String match = stripPostfix( workingFile.getName() ) + "_.*";
	        File path = FileManager.getAlbumStorageDir();
	        Log.d( TAG, "Searching for '" + match + "' in " + path + "..." );
	
	        File[] files = path.listFiles();
	        if( files != null )
	        {
	            Log.d( TAG, "Number of files: " + files.length );
	            for( int i = 0; i < files.length; i++ )
	            {
	            	if( files[i].getName().matches( match ) )
	            	{
	            		Log.i( TAG, "Deleting File '" + files[i].getName() + "'" );
	            		files[i].delete();
	            	}
	            }
	        }
		}
		catch( IOException e ) 
		{
			lastError = "Failed to delete all files - " + e;
		}
    }
    
    public String toString()
    {
    	StringBuffer sb = new StringBuffer();
    	sb.append( "File=" + originalFile + "," );
    	sb.append( "File=" + workingFile + "," );
    	sb.append( "Header=[" + header + "]," );
    	sb.append( "ListSize=" + list.size() + "," );
    	sb.append( "LastError=" + lastError + "," );
    	return sb.toString();
    }
}
