package com.tripgis.dmt.file;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Map;

import com.tripgis.dmt.types.EventRecord;
import com.tripgis.dmt.types.LegacyRecord;
import com.tripgis.dmt.types.TripDbRecord;
import com.tripgis.dmt.util.DmtUtils;
import com.tripgis.dmt.util.LogTranslator;

/**
 * References:
 *    https://developer.android.com/guide/topics/data/data-storage.html#filesInternal
 *    https://docs.oracle.com/javase/7/docs/api/java/io/File.html
 *    https://www.mkyong.com/java/how-to-export-data-to-csv-file-java/
 *
 * Created by pk on 7/3/17.
 */

public class LegacyFileManager extends FileManager<TripDbRecord, LegacyRecord>
{
    // Name to use in log lines
    public static final String TAG = LegacyFileManager.class.getSimpleName();
    public static final LogTranslator Log = new LogTranslator( TAG );
    
    private HashMap<String,Integer> mapIndices;

	public LegacyFileManager( String name )
    { super( null, stripBasePostfix( name ) ); }
    public LegacyFileManager( File original ) { super( original ); }
    public LegacyFileManager( File original, String name )
    { super( original, name, false ); }
    public LegacyFileManager( File original, String name, boolean forceRead )
    { super( original, name, forceRead ); }

	@Override
	protected EventRecord itemToEvent( LegacyRecord t) 
	{
		return (EventRecord) t;
	}

	@Override
	protected LegacyRecord eventToItem( EventRecord e )
	{
		return new LegacyRecord( e );
	}

	@Override
	protected TripDbRecord readHeader( BufferedReader lineReader ) throws IOException 
	{
		Log.t( TAG, "readHeader( " + lineReader + " )" );
		mapIndices = LegacyRecord.buildHeaderFromCsvArray( super.parseLine( lineReader ) ); 
		TripDbRecord tdr = new TripDbRecord( 0, DmtUtils.removeExtension( originalFile.getName() ) );
		tdr.setFileName( DmtUtils.removeExtension( originalFile.getName() ) );
		tdr.setSource( TripDbRecord.SOURCE_LEGACY );
		return tdr;
	}

	@Override
	protected EventRecord readListItem( String[] data ) throws IOException 
	{
		Log.t( TAG, "readListItem( " + data.length + " )" );
		return LegacyRecord.buildFromCsvArray( mapIndices, data ); 
	}

	/**
	 * References:
	 *    https://stackoverflow.com/questions/21054415/how-to-sort-a-hashmap-by-the-integer-value
	 */
	@SuppressWarnings("unchecked")
	@Override
	protected boolean writeHeader( FileWriter writer, TripDbRecord h ) throws IOException 
	{
		Log.t( TAG, "writeHeader( " + writer + ", " + h + " )" );
		
		if( overWriteOriginal )
		{
			Log.t( TAG, "writing headers back out (copying original)..." );
			Object[] a = mapIndices.entrySet().toArray();
			Arrays.sort(a, new Comparator<Object>() 
			{
			    public int compare( Object o1, Object o2 ) 
			    {
			        return ((Map.Entry<String, Integer>) o1).getValue()
			                   .compareTo( ((Map.Entry<String, Integer>) o2).getValue() );
			    }
			});
			
			for( int i = 0; i < a.length - 1; i++ )
			{
				Map.Entry<String,Integer> entry = (Map.Entry<String, Integer>) a[i];
				writer.append( entry.getKey() + FileManager.CSV_SEPARATOR );
			}
	
			Map.Entry<String,Integer> entry = (Map.Entry<String, Integer>) a[a.length-1];
			writer.append( entry.getKey() + "\n" );
		}
		else
			writer.append( header.generateCsvString() );

		return true;
	}

	@Override
	protected String getListItemString(EventRecord e) 
	{
		if( overWriteOriginal )
			return ((LegacyRecord) e).generateCsvString( mapIndices );
		
		return e.generateCsvString();
	}
}
