package com.tripgis.dmt.geocode;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.HttpURLConnection;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;

import org.json.simple.JSONArray;
import org.json.simple.JSONObject;
import org.json.simple.parser.JSONParser;

import com.tripgis.dmt.types.DistanceRecord;
import com.tripgis.dmt.types.EventRecord;
import com.tripgis.dmt.types.GeocodeRecord;
import com.tripgis.dmt.types.MatchRecord;
import com.tripgis.dmt.util.DmtProperties;
import com.tripgis.dmt.util.LatLng;
import com.tripgis.dmt.util.LogTranslator;
import com.tripgis.dmt.util.DmtUtils;

/**
 * To set up an API Key:
 *    https://console.developers.google.com/apis/dashboard?project=pro-bruin-172600&duration=PT1H
 *    
 * @author pk
 *
 */
public class Geocoder 
{
	// Name to use in log lines
    public static final String TAG = Geocoder.class.getSimpleName();
    public static final LogTranslator Log = new LogTranslator( TAG );

    private static final String googleGeocodeUrl = "https://maps.googleapis.com/maps/api/geocode";
	private static final String googlePlaceUrl = "https://maps.googleapis.com/maps/api/place";
	private static final String googleDistanceUrl = "https://maps.googleapis.com/maps/api/distancematrix";

	private static final String detailSearch  = "details";
	private static final String nearbySearch  = "nearbysearch";

//	private static final String XML = "xml";
	private static final String JSON = "json";
	
	private static final String API_KEY = "key=" + DmtProperties.getProperty( "geocoder.api.key" );


	// Types: https://developers.google.com/places/web-service/supported_types
	// https://console.developers.google.com/apis/credentials/key/216?project=pro-bruin-172600
	// https://code.google.com/archive/p/json-simple/

	private static Geocoder instance;
	private JSONParser parser;

	public static final String STATUS_OK                    = "OK";
	public static final String STATUS_ZERO_RESULTS          = "ZERO_RESULTS";
	public static final String STATUS_INVALID_REQUEST       = "INVALID_REQUEST";	
	public static final String STATUS_MAX_ELEMENTS_EXCEEDED = "MAX_ELEMENTS_EXEECED";	
	public static final String STATUS_OVER_QUERY_LIMIT      = "OVER_QUERY_LIMIT";	
	public static final String STATUS_REQUEST_DENIED        = "REQUEST_DENIED";	
	public static final String STATUS_UNKNOWN_ERROR         = "UNKNOWN_ERROR";

	public static final String LT_ROOFTOP            = "ROOFTOP";
	public static final String LT_RANGE_INTERPOLATED = "RANGE_INTERPOLATED";
	public static final String LT_GEOMETRIC_CENTER   = "GEOMETRIC_CENTER";
	public static final String LT_APPROXIMATE        = "APPROXIMATE";

	private String status;
	private String prevToken = null;
	private String nextToken = null;

	private Geocoder()
	{
		parser = new JSONParser();
	}

	public static Geocoder getInstance()
	{
		if( instance == null )
			instance = new Geocoder();
		return instance;
	}
	
	private void resetTokens()
	{
		prevToken = null;
		nextToken = null;
	}
	
	public String getLastRequestStatus() { return status; }
	public String getPrevToken() { return prevToken; }
	public String getNextToken() { return nextToken; }

	public static String getParameterString( String text )
	{
		return text.replaceAll( " ", "+" );
	}
	
	private String getRequest( String url ) throws IOException //throws Exception
	{
		Log.t( TAG, "getRequest( " + url + " )" );
		final URL obj = new URL( url );
		final HttpURLConnection conn = (HttpURLConnection) obj.openConnection();
		conn.setRequestMethod( "GET" );

		if( conn.getResponseCode() != 200 )
		{
			throw new IOException( "URL: " + url + ", Response: " + conn.getResponseCode() );
		}

		InputStreamReader isr = new InputStreamReader( conn.getInputStream() );
		BufferedReader reader = new BufferedReader( isr );
		String line;
		StringBuffer buffer   = new StringBuffer();

		line = reader.readLine();
		while( line != null )
		{
			buffer.append( line );
			line = reader.readLine();
		}

		reader.close();
		Log.d( TAG, "Returning " + buffer.length() + " characters." );
		return buffer.toString();
	}


	// Used to find address of a given position
	// latlng should be in format:  <lat>,<lng>
	// https://developers.google.com/maps/documentation/geocoding/intro
	// Example: http://maps.googleapis.com/maps/api/geocode/json?latlng=32.99417,-96.6773&sensor=true
	private String buildReverseGeocodeSearch( String format, LatLng latlng )
	{
		String url = googleGeocodeUrl + "/" + format + "?latlng=" + latlng.getDisplayString() + "&" + API_KEY;
		Log.d( TAG, "url=" + url );
		return url;
	}

	// Obtain a lat/long for the given address
	//https://maps.googleapis.com/maps/api/geocode/json?address=2504+Buttercup+Dr,+Richardson,+TX+75082,+USA&key=AIzaSyB_KambThXCgU9RTaioveQ6AzDy3ujTdSM
	private String buildGeocodeSearch( String format, String address )
	{
		String url = googleGeocodeUrl + "/" + format + "?address=" + 
				DmtUtils.toUrlString( address ) + "&" + API_KEY;
		Log.d( TAG, "url=" + url );
		return url;
	}

	// latlng should be in format:  <lat>,<lng>
	// https://maps.googleapis.com/maps/api/place/nearbysearch/json?location=32.994174,-96.677350&radius=50&key=AIzaSyB_KambThXCgU9RTaioveQ6AzDy3ujTdSM
	private String buildNearbySearch( String format, HashMap<String, String> params )
	{
		String url = googlePlaceUrl + "/" + nearbySearch + "/" + 
			format + "?";
		for( String name : params.keySet() )
			url += name + "=" + params.get( name ) + "&";
		url += API_KEY;
 		Log.d( TAG, "url=" + url );
		return url;
	}

	// https://maps.googleapis.com/maps/api/place/details/json?placeid=ChIJ0fWdTUwZTIYR1IgFRx4lQLQ&key=AIzaSyB_KambThXCgU9RTaioveQ6AzDy3ujTdSM
	private String buildDetailSearch( String format, String placeId )
	{
		String url = googlePlaceUrl + "/" + detailSearch + "/" +
			format + "?placeid=" + placeId + "&" + API_KEY;
		Log.d( TAG, "url=" + url );
		return url;
	}

	/**
	 * Other Options: 
		units (metric, imperial)
		avoid (tolls, highways, ferries, indoor)
		arrival_time, departure_time (long)
		traffic_model (best_guess, pessimistic, optimistic)
		transit_mode (bus, subway, train, tram, rail)		// mode == transit only
		transit_routing_preferences (less_walking, fewer_transfers)
		mode (driving, walking, bicycling, transit)
	 */
	// https://developers.google.com/maps/documentation/distance-matrix/intro
	// https://maps.googleapis.com/maps/api/distancematrix/json?units=imperial&origins=place_id:ChIJ0fWdTUwZTIYR1IgFRx4lQLQ&destinations=70+Inge+Dr,+Richarson,+TX|4505+Plaza+Dr,+Irving,+TX&key=AIzaSyB_KambThXCgU9RTaioveQ6AzDy3ujTdSM
	private String buildDistanceSearch( String format, HashMap<String, String> params )
	{
		String url = googleDistanceUrl + "/" + format + "?";
		for( String name : params.keySet() )
			url += name + "=" + params.get( name ) + "&";
		url += API_KEY;
		Log.d( TAG, "url=" + url );
		return url;
	}
	
	// Parse Options
	// ZERO_RESULTS: https://maps.googleapis.com/maps/api/place/nearbysearch/json?location=32.994174,-96.677350&keyword=dogfish&radius=50&key=AIzaSyB_KambThXCgU9RTaioveQ6AzDy3ujTdSM
	// INVALID_REQUEST: https://maps.googleapis.com/maps/api/place/nearbysearch/json?location=92.994174,-96.677350&keyword=dogfish&radius=50&key=AIzaSyB_KambThXCgU9RTai


	// http://maps.googleapis.com/maps/api/geocode/json?latlng=32.99417,-96.6773&sensor=true
	// https://www.mkyong.com/java/json-simple-example-read-and-write-json/
	// https://code.google.com/archive/p/json-simple/
	@SuppressWarnings("unchecked")
	private ArrayList<GeocodeRecord> parseGeocodeSearchResponse( String result ) throws IOException
	{
		Log.t( TAG, "parseGeocodeSearchResponse( " + result + " )" );
		try
		{
			ArrayList<GeocodeRecord> list = new ArrayList<>();
	
	        Object obj = parser.parse( result );
	        Log.d( TAG, "obj: " + obj );
	
	        JSONObject jObj = (JSONObject) obj;
			status = (String) jObj.get( "status" );
			Log.d( TAG, "Status: " + status );
			if( !status.equals( STATUS_OK ) )
				return null;
			
			JSONArray results = (JSONArray) jObj.get( "results" );
			if( results.size() == 0 )
				throw new IOException( "Result set is empty" );
	
			for( Object iObj : results )
			{
				GeocodeRecord gr = GeocodeRecord.EMPTY.copy();
				JSONObject resultItem = (JSONObject) iObj;
				JSONArray address = (JSONArray) resultItem.get( "address_components" );
				Iterator<JSONObject> itr = address.iterator();
				while( itr.hasNext() )
					setAddressComponent( gr, itr.next() );
				gr.setPlaceId( (String) resultItem.get( "place_id" ) );
	
				//ignore formatted_address
				gr.setRating( getDoubleValue( resultItem, "rating" ) );
	
				JSONArray types = (JSONArray) resultItem.get( "types" );
				gr.setLocationType( (String) types.get( 0 ) );
	
				JSONObject geometry = (JSONObject) resultItem.get( "geometry" );
				// ignore gemoetry's bounds and viewport
				// geometry's location_type may be more useful for locationType? (ROOFTOP)
				JSONObject location = (JSONObject) geometry.get( "location" );
				double lat = ((Number) location.get( "lat" )).doubleValue();
				double lng = ((Number) location.get( "lng" )).doubleValue();
				gr.setGeoLocation( new LatLng( lat, lng ) );
	
				Log.d( TAG, "Adding " + gr );
				list.add( gr );
			}
	
			Log.d( TAG, "Number of Results: " + list.size() );
			return list;
		}
		catch( org.json.simple.parser.ParseException e) 
		{
			throw new IOException( "Failed to parse URL - " + e );
		}
	}

	/**
	 * References:
	 *    https://stackoverflow.com/questions/21690413/decoding-floating-point-number-in-simple-json-java
	 *    
	 * Example:
	 *    https://maps.googleapis.com/maps/api/place/nearbysearch/json?location=32.994174,-96.677350&radius=50&key=AIzaSyB_KambThXCgU9RTaioveQ6AzDy3ujTdSM
	 *    
	 * @param result
	 * @return
	 * @throws IOException
	 */
	@SuppressWarnings("unchecked")
	private ArrayList<GeocodeRecord> parseNearbySearchResponse( String result ) throws IOException
	{
		Log.t( TAG, "parseNearbySearchResponse( " + result + " )" );
		try
		{
			ArrayList<GeocodeRecord> list = new ArrayList<>();
	
	        Object obj = parser.parse( result );
	        Log.d( TAG, "obj: " + obj );
	
	        JSONObject jObj = (JSONObject) obj;
			status = (String) jObj.get( "status" );
			Log.d( TAG, "Status: " + status );
			if( !status.equals( STATUS_OK ) )
				return null;
			
			// Manage pages
			if( nextToken != null )
				prevToken = nextToken;
			nextToken = (String) jObj.get( "next_page_token" );
			
			JSONArray results = (JSONArray) jObj.get( "results" );
			Log.d( TAG, "Results: " + results.size() );
			if( results.size() == 0 )
				throw new IOException( "result set is empty" );
	
			for( Object iObj : results )
			{
				GeocodeRecord gr = GeocodeRecord.EMPTY.copy();
				JSONObject resultItem = (JSONObject) iObj;
				Log.d( TAG, "Item: " + resultItem );
				JSONArray address = (JSONArray) resultItem.get( "address_components" );

				if( address != null )
				{
					Log.d( TAG, "Address: " + address );
					Iterator<JSONObject> itr = address.iterator();
					while( itr.hasNext() )
						setAddressComponent( gr, itr.next() );
				}
				else
				{
					gr.setAddress( (String) resultItem.get( "vicinity" ) );
				}
				
				gr.setPlaceId( (String) resultItem.get( "place_id" ) );
				gr.setStore( (String) resultItem.get( "name" ) );
	
				//ignore icon, id, reference, scope, vicinity, and
				// photos [ height, width, photo_reference, html_attributions [] ]
				
				gr.setRating( getDoubleValue( resultItem, "rating" ) );
	
				JSONArray types = (JSONArray) resultItem.get( "types" );
				gr.setLocationType( (String) types.get( 0 ) );
	
				JSONObject geometry = (JSONObject) resultItem.get( "geometry" );
	
				// ignore gemoetry's viewport
				JSONObject location = (JSONObject) geometry.get( "location" );
				double lat = ((Number) location.get( "lat" )).doubleValue();
				double lng = ((Number) location.get( "lng" )).doubleValue();
				gr.setGeoLocation( new LatLng( lat, lng ) );
	
				Log.d( TAG, "Adding " + gr );
				list.add( gr );
			}
	
			Log.d( TAG, "Number of Results: " + list.size() );
			return list;
		}
		catch( org.json.simple.parser.ParseException e) 
		{
			throw new IOException( "Failed to parse URL - " + e );
		}
	}
	
	private String getDoubleValue( JSONObject obj, String key )
	{
		Object rObj = obj.get( key );
		if( rObj != null )
		{
			double val = ((Number) rObj).doubleValue();
			return Double.toString( val );
		}
		return "";
	}

	// https://maps.googleapis.com/maps/api/place/details/json?placeid=ChIJ0fWdTUwZTIYR1IgFRx4lQLQ&key=AIzaSyB_KambThXCgU9RTaioveQ6AzDy3ujTdSM
	@SuppressWarnings("unchecked")
	private GeocodeRecord parseDetailSearchResponse( String result ) throws IOException
	{
		Log.t( TAG, "parseNearbySearchResponse( " + result + " )" );
		try 
		{
			Object obj = parser.parse( result );
	        Log.d( TAG, "obj: " + obj );
	
	        JSONObject jObj = (JSONObject) obj;
			status = (String) jObj.get( "status" );
			Log.d( TAG, "Status: " + status );
			if( !status.equals( STATUS_OK ) )
				return null;
			
			GeocodeRecord gr = GeocodeRecord.EMPTY.copy();
			JSONObject resultItem = (JSONObject) jObj.get( "results" );
			gr.setPlaceId( (String) resultItem.get( "place_id" ) );
	
			//ignore formatted_address
			// Ignore adr_address, icon, id, name, reference, scope, vicinity, and
			// url, photos [ height, width, photo_reference, html_attributions [] ]
	
			JSONArray address = (JSONArray) resultItem.get( "address_components" );
			Iterator<JSONObject> itr = address.iterator();
			while( itr.hasNext() )
				setAddressComponent( gr, itr.next() );
	
			gr.setRating( getDoubleValue( resultItem, "rating" ) );
			
			JSONArray types = (JSONArray) resultItem.get( "types" );
			gr.setLocationType( (String) types.get( 0 ) );	// premise
	
			JSONObject geometry = (JSONObject) resultItem.get( "geometry" );
	
			// ignore gemoetry's viewport
			JSONObject location = (JSONObject) geometry.get( "location" );
			double lat = ((Number) location.get( "lat" )).doubleValue();
			double lng = ((Number) location.get( "lng" )).doubleValue();
			gr.setGeoLocation( new LatLng( lat, lng ) );
	
			gr.setUtcOffset( (String) resultItem.get( "utc_offset" ) ); // -300
	
			Log.d( TAG, "Returning " + gr );
			return gr;
		}
		catch( org.json.simple.parser.ParseException e) 
		{
			throw new IOException( "Failed to parse URL - " + e );
		}
	}

	// https://maps.googleapis.com/maps/api/distancematrix/json?units=imperial&origins=place_id:ChIJ0fWdTUwZTIYR1IgFRx4lQLQ&destinations=70+Inge+Dr,+Richarson,+TX|4505+Plaza+Dr,+Irving,+TX&key=AIzaSyB_KambThXCgU9RTaioveQ6AzDy3ujTdSM
	public DistanceRecord parseDistanceMatrixResponse( String result ) throws IOException
	{
		Log.t( TAG, "parseDistanceMatrixResponse( " + result + " )" );
		try
		{
	        Object obj = parser.parse( result );
	        Log.d( TAG, "obj: " + obj );
	    	
	        JSONObject jObj = (JSONObject) obj;
			status = (String) jObj.get( "status" );
			Log.d( TAG, "Status: " + status );
			if( !status.equals( STATUS_OK ) )
				return DistanceRecord.EMPTY;
	
			DistanceRecord dr = DistanceRecord.EMPTY.copy();
			
			JSONArray dstA = (JSONArray) jObj.get( "destination_addresses" );
			String dstAddress = (String) dstA.get( 0 );
			dr.setDestination( GeocodeRecord.parseAddress( dstAddress ) );
			
			JSONArray orgA = (JSONArray) jObj.get( "origin_addresses" );
			String orgAddress = (String) orgA.get( 0 );
			dr.setOrigin( GeocodeRecord.parseAddress( orgAddress ) );
			
			JSONArray  rows = (JSONArray) jObj.get( "rows" );
			JSONObject rowObj = (JSONObject) rows.get( 0 );
			JSONArray  elements = (JSONArray) rowObj.get( "elements" );
			JSONObject elObj = (JSONObject) elements.get( 0 );
			String elementStatus = (String) elObj.get( "status" );
			if( !elementStatus.equals( STATUS_OK ) )
			{
				status = "element:" + elementStatus;
				return DistanceRecord.EMPTY;
			}
	
			JSONObject distance = (JSONObject) elObj.get( "distance" );
			dr.setTextDistance( (String) distance.get( "text" ) );
			dr.setDistance( ((Number) distance.get( "value" )).longValue() );
	
			JSONObject duration = (JSONObject) elObj.get( "duration" );
			dr.setDuration( ((Number) duration.get( "value" )).longValue() );
	
			// ignore duration_in_traffic (for now)
			// ignore fare (currency, value, text )
	
			Log.d( TAG, "Returning: " + dr );
			return dr;
		}
		catch( org.json.simple.parser.ParseException e) 
		{
			throw new IOException( "Failed to parse URL - " + e );
		}
	}
	
	private void setAddressComponent( GeocodeRecord gr, JSONObject resultObj ) throws IOException
	{
		Log.t( TAG, "setAddressComponent( " + gr + ", " + resultObj + " )" );
		String shortName = (String) resultObj.get( "short_name" ); // ignore long_name
		JSONArray types  = (JSONArray) resultObj.get( "types" );
		String type = (String) types.get( 0 );

		Log.d( TAG, "type: " + type + ", value: " + shortName );
		switch( type )
		{
			case "street_number":
				gr.setAddress( shortName );
				break;
			case "route":
				gr.setAddress( gr.getAddress() + " " + shortName );
				break;
			case "locality":
				gr.setCity( shortName );
				break;
			case "administrative_area_level_2":
				gr.setCounty( shortName );
				break;
			case "administrative_area_level_1":
				gr.setState( shortName );
				break;
			case "country":
				gr.setCountry( shortName );
				break;
			case "postal_code":
				gr.setZip( shortName );
				break;
			case "postal_code_suffix":
				// do nothing
				break;
			default:
				Log.d( TAG, "Adding " + type + " to store (" + shortName + ")" );
				gr.setStore( shortName );
				break;
		};
	}

	private ArrayList<MatchRecord> buildMatchRecordList( ArrayList<GeocodeRecord> list )
	{
		ArrayList<MatchRecord> newList = new ArrayList<>();
		for( GeocodeRecord gr : list )
			newList.add( new MatchRecord( gr ) );
		return newList;
	}

	public DistanceRecord getDistance( EventRecord origin, EventRecord dest ) throws IOException
	{
		Log.t( TAG, "getDistance( " + origin + ", " + dest + " )" );
		HashMap<String,String> params = new HashMap<>();
		return getDistance( origin, dest, params );
	}

	public DistanceRecord getDistance( EventRecord origin, EventRecord dest, HashMap<String,String> params ) throws IOException
	{
		Log.t( TAG, "getDistance( " + origin + ", " + dest + ", " + params.size() + " )" );
		if( !origin.getGeoData().getPlaceId().isEmpty() )
			params.put( "origins", "place_id:" + origin.getGeoData().getPlaceId() );
		else if( origin.getMobileData().getLocation().isValid() )
			params.put( "origins", origin.getMobileData().getLocation().getDisplayString() );
		else
			params.put( "origins", DmtUtils.toUrlString( origin.getGeoData().getFormattedAddress() ) );

		if( !dest.getGeoData().getPlaceId().isEmpty() )
			params.put( "destinations", "place_id:" + dest.getGeoData().getPlaceId() );
		else if( dest.getMobileData().getLocation().isValid() )
			params.put( "destinations", dest.getMobileData().getLocation().getDisplayString() );
		else
			params.put( "destinations", DmtUtils.toUrlString( dest.getGeoData().getFormattedAddress() ) );

		String url = buildDistanceSearch( JSON, params );
		String data = getRequest( url );
		return parseDistanceMatrixResponse( data );
	}

	public ArrayList<MatchRecord> reverseGeocode( LatLng latlng ) throws IOException
	{
		Log.t( TAG, "reverseGeocode( " + latlng + " )" );
		// To obtain a list of addresses that are proximily close to specified
		String url = Geocoder.getInstance().buildReverseGeocodeSearch( JSON, latlng );
		String data = Geocoder.getInstance().getRequest( url );
		ArrayList<GeocodeRecord> list = Geocoder.getInstance().parseGeocodeSearchResponse( data );
		String status = Geocoder.getInstance().getLastRequestStatus();
		if( list == null )
			throw new IOException( "Last Request did not work, RC=" + status );

		return buildMatchRecordList( list );
	}

	public ArrayList<MatchRecord> geocode( String address ) throws IOException
	{
		Log.t( TAG, "geocode( " + address + " )" );
		// To obtain a lat/long
		String url = Geocoder.getInstance().buildGeocodeSearch( JSON, address );
		String data = Geocoder.getInstance().getRequest( url );
		ArrayList<GeocodeRecord> list = Geocoder.getInstance().parseGeocodeSearchResponse( data );
		String status = Geocoder.getInstance().getLastRequestStatus();
		if( list == null )
			throw new IOException( "Last Request did not work, RC=" + status );

		return buildMatchRecordList( list );
	}

	public ArrayList<MatchRecord> nearbySearch( LatLng latlng, HashMap<String, String> params ) throws IOException
	{
		Log.t( TAG, "nearbySearch( " + latlng + ", " + params.size() + " )" );
		// To obtain a list of addresses that are proximily close to specified
		params.put( "location", latlng.getDisplayString() );
		String url = Geocoder.getInstance().buildNearbySearch( JSON, params );
		String data = Geocoder.getInstance().getRequest( url );
		if( params.get( "token" ) == null )
			resetTokens();
		ArrayList<GeocodeRecord> list = Geocoder.getInstance().parseNearbySearchResponse( data );
		String status = Geocoder.getInstance().getLastRequestStatus();
		if( list == null )
			throw new IOException( "Last Request did not work, RC=" + status );

		return buildMatchRecordList( list );
	}

	public GeocodeRecord detail( String placeId ) throws IOException
	{
		Log.t( TAG, "detail( " + placeId + " )" );
		// To obtain a list of addresses that are proximily close to specified
		String url = Geocoder.getInstance().buildDetailSearch( JSON, placeId );
		String data = Geocoder.getInstance().getRequest( url );
		GeocodeRecord gr = Geocoder.getInstance().parseDetailSearchResponse( data );
		String status = Geocoder.getInstance().getLastRequestStatus();
		if( gr == null )
			throw new IOException( "Last Request did not work, RC=" + status );

		return gr;
	}
}
