package com.tripgis.dmt.geocode;

import java.sql.Date;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;

import com.tripgis.dmt.DmtController;
import com.tripgis.dmt.db.DbModel;
import com.tripgis.dmt.db.DbModelEvents;
import com.tripgis.dmt.types.EventRecord;
import com.tripgis.dmt.types.EventSubType;
import com.tripgis.dmt.types.EventType;
import com.tripgis.dmt.types.TripDbRecord;
import com.tripgis.dmt.ui.EventForm;
import com.tripgis.dmt.util.LogTranslator;
import com.tripgis.dmt.util.DmtUtils;
import com.tripgis.shared.types.LocationRecord;

public class TripStatistics 
{
	
    // Name to use in log lines
    public static final String TAG = TripStatistics.class.getSimpleName();
    public static final LogTranslator Log = new LogTranslator( TAG );

	private DbModel db;
	private EventForm gui;
	
	private BaseStats tripData;
	private HashMap<String, EventStats> mapData;
	private double totalTime;
	
	public class BaseStats
	{
		public int id;
		public Date startTime;
		public Date endTime;
		public double duration;
		public int numEvents;
	}
	
	public class EventStats extends BaseStats
	{
		public double pctOfTotalTime;
		public double pctOfEventTime;
		public double pctOfActivity;
	}
	
	public TripStatistics( DmtController c )
	{
		Log.t( TAG, "TripStatistics( " + c + " )" );
		db = DmtController.getDb();
		gui = DmtController.getDisplay();
		tripData = null;
		mapData = new HashMap<>();
	}

	public void generate()
	{
		Log.t( TAG, "generate()" );
		
		calculateTripData( gui.getSelectedTrip() );
		EventType airport = new EventType( "airport" );
		EventType hotel = new EventType( "hotel" );
		EventType lodging = new EventType( "lodging" );
		EventType entertainment = new EventType( "entertainment" );
		EventSubType estAirport = new EventSubType( airport, "" ); //"security" );
		EventSubType estHotel = new EventSubType( hotel, "" ); //"checkn" );
		EventSubType estAirbnb = new EventSubType( lodging, "" ); //"airbnb" );
		EventSubType estCurling = new EventSubType( entertainment, "curling" );
		
		EventSubType[] types = new EventSubType[]{ estAirport, estHotel, estCurling, estAirbnb };
		
		HashMap<String,ArrayList<EventRecord>> mapLists = populateLists( tripData.id, types );
		mergeLists( mapLists.get( lodging.getName() ), mapLists.get( hotel.getName() ) );
		
		for( EventSubType est : new EventSubType[]{ estAirport, estAirbnb, estCurling } )
		{
			String name = getEstKey( est );
			ArrayList<EventRecord> list = getPoints( DbModelEvents.HALF_MILE, mapLists.get( name ) );
			if( list != null )
				mapData.put( name, this.calculateRollupData( 0, list ) );
		}
		
//		HashMap<String,ArrayList<EventRecord>> mapLists = generateEventLists( tripData.id, types );

		Log.i( TAG, "Total Time from Start to finish: " + DmtUtils.getPrettyDuration( totalTime ) );
		printStats( "Trip", tripData );
		for( String name : mapData.keySet() )
		{
			EventStats stats = mapData.get( name );
			printStats( name, stats );
			Log.i( TAG, "% Total Time: " + String.format( "%.2f", stats.pctOfTotalTime * 100 ) );
			Log.i( TAG, "% Event Time: " + String.format( "%.2f", stats.pctOfEventTime * 100 ) );
			Log.i( TAG, " % of Events: " + String.format( "%.2f", stats.pctOfActivity * 100 ) );
		}
	}
	
	public void printStats( String title, BaseStats stats )
	{
		Log.i( TAG, title + " Statistics: " );
		Log.i( TAG, "      Start: " + LocationRecord.getSqlDateString( stats.startTime ) );
		Log.i( TAG, "        End: " + LocationRecord.getSqlDateString( stats.endTime ) );
		Log.i( TAG, "   Duration: " + DmtUtils.getPrettyDuration( stats.duration ) );
		Log.i( TAG, "  NumEvents: " + stats.numEvents );
	}
	
	public EventStats calculateRollupData( int id, ArrayList<EventRecord> list )
	{
		if( list == null || list.size() == 0 )
			return new EventStats();
		
		Log.t( TAG, "calculateRollupData( " + id + ", " + list.size() + " )" );
		EventStats stats = new EventStats();
		stats.id = id;
		stats.startTime = list.get(0).getMobileData().getDate();
		LocationRecord lr = list.get(list.size()-1).getMobileData(); 
		if( lr.getEndTime() != null )
			stats.endTime   = lr.getEndTime();
		else
			stats.endTime   = lr.getDate();
		stats.numEvents = list.size();
		
		for( EventRecord er : list )
		{
			stats.duration += DmtUtils.getDuration( er.getMobileData().getDate(), er.getMobileData().getEndTime() );
			//Log.d( TAG, "EventId: " + er.getEventId() + ", Duration: " + stats.duration );
		}

		if( tripData != null )
		{
			stats.pctOfActivity = stats.numEvents * 1.0 / tripData.numEvents;
			stats.pctOfTotalTime     = stats.duration * 1.0 / totalTime;
			stats.pctOfEventTime     = stats.duration * 1.0 / tripData.duration;
			//Log.d( TAG, "CALC: " + stats.duration + " * 1.0 / " + tripData.duration + " = " + stats.pctOfTime );
			//Log.d( TAG, "CALC: " + stats.numEvents + " * 1.0 / " + tripData.numEvents + " = " + stats.pctOfActivity );
		}
		else
		{
			totalTime = DmtUtils.getDuration( stats.startTime, stats.endTime );
		}
		return stats;
	}
	
	public BaseStats calculateTripData( TripDbRecord tdr )
	{
		Log.t( TAG, "calculateTripData( " + tdr + " )" );
		tripData       = calculateRollupData( tdr.getTripId(), tdr.getList() );
		return tripData;
	}
	
	public void printStatistics()
	{
		Log.t( TAG, "printStatistcs()" );
		
	}
	
	private String getEstKey( EventSubType est )
	{
		String name = est.getType().getName();
		if( !est.getName().isEmpty() )
			name += "::" + est.getName();
		return name;
	}
	
	private HashMap<String,ArrayList<EventRecord>> populateLists( int tripId, EventSubType[] types )
	{
		Log.t( TAG, "populateLists( " + tripId + ", " + types.length + " )" );
		HashMap<String,ArrayList<EventRecord>> mapLists = new HashMap<>();
		
		for( EventSubType type : types )
		{
			String name = getEstKey( type );
			if( type.getName().isEmpty() )
				mapLists.put( name, db.getEvents().getEventsByType( tripId, type.getType() ) );
			else
				mapLists.put( name, db.getEvents().getEventsBySubType( tripId, type ) );

			int eventCount = ( mapLists.get( name ) != null ? mapLists.get( name ).size() : 0 );
			Log.d( TAG, "Item: " + name + ", Events: " + eventCount );
		}
		
		return mapLists;
	}
	
	public static ArrayList<EventRecord> mergeLists( ArrayList<EventRecord> keep, ArrayList<EventRecord> add )
	{
		if( keep == null )
		{
			if( add != null )
			{
				keep = add;
				return keep;
			}
			
			return null;
		}
		

		if( add == null )
			return keep;
		
		Log.t( TAG, "mergeLists( " + keep.size() + ", " + add.size() + " )" );
		for( EventRecord event : add )
		{
			boolean found = false;
			//Log.t( TAG, "Search: " + event );
			for( EventRecord check : keep )
			{
				//Log.t( TAG, "      : " + check );
				if( event.matches( check ) )
				{
					found = true;
					break;
				}
			}
			
			if( !found )
			{
				Log.d( TAG, "   Adding " + event.getEventId() + ": " + event );
				keep.add( event );
			}
		}
		
		return keep;
	}
	
	private ArrayList<EventRecord> getPoints( double distance, ArrayList<EventRecord> list )
	{
		if( list == null )
			return null;
		
		Log.t( TAG, "generateEventLists( " + distance + ", " + list.size() + " )" );
		ArrayList<EventRecord> out = new ArrayList<>();
		for( EventRecord event : list )
		{
			ArrayList<EventRecord> tmp = getPoints( distance, event );
			if( tmp != null )
				mergeLists( out, tmp );
		}
		
		return out;
	}
	
	private ArrayList<EventRecord> getPoints( double distance, EventRecord event )
	{
		Log.t( TAG, "generateEventLists( " + distance + ", " + event + " )" );
		try 
		{
			return db.getEvents().getEventsWithinRange( event, distance );
		}
		catch( SQLException e ) 
		{
			Log.e( TAG, "Failed to get points - " + e );
			return null;
		}
	}
	
/*	private HashMap<String,ArrayList<EventRecord>> generateEventLists( int tripId, EventSubType[] types )
	{
		Log.t( TAG, "generateEventLists( " + tripId + ", " + types.length + " )" );
		HashMap<String,ArrayList<EventRecord>> mapLists = new HashMap<>();
		
		try 
		{
			for( int i = 0; i < types.length; i++ )
			{
				ArrayList<EventRecord> list = db.getEvents().getEventsBySubType( tripId, types[i] );
				String name = types[i].getType().getName();
				for( EventRecord record : list )
				{
					if( mapLists.get( name ) == null )
						mapLists.put( name, db.getEvents().getEventsWithinRange( tripId, record.getMobileData().getLocation(), DbModelEvents.HALF_MILE ) );
					else
					{
						for( EventRecord er : list )
							if( EventRecord.find( mapLists.get( name ), er.getMobileData() ) == EventRecord.EMPTY )
								mapLists.get( name ).add( er );
					}
				}

				if( mapLists.get( name ) == null )
				{
					Log.w( TAG, "No events found for " + name );
				}
				else
				{
					int eventCount = ( mapLists.get( name ) != null ? mapLists.get( name ).size() : 0 );
					Log.d( TAG, "Item: " + name + ", Points: " + list.size() + ", Events: " + eventCount );
					mapData.put( name, calculateRollupData( i, mapLists.get( name ) ) );
				}
			}
		}
		catch (SQLException e) 
		{
			DmtController.setError( TAG, "Failed to obtain stats - " + e );
		}
		
		return mapLists;
	} */

	public boolean isGenerated() { return tripData != null; }
	
	public EventStats calculateGapStats( EventRecord start, EventRecord end )
	{
		Log.t( TAG, "calculateGapStats( " + start.getMobileData().getDate() + ", " + end.getMobileData().getDate() + " )" );
		ArrayList<EventRecord> list = new ArrayList<>();
		
		boolean foundStart = false;
		long matchTime = start.getMobileData().getDate().getTime();
		for( EventRecord er : gui.getSelectedTrip().getList() )
		{
			if( !foundStart )
			{
				if( er.getMobileData().getDate().getTime() <= matchTime )
				{
					//Log.d( TAG, "adding leading item" );
					list.add( er );
				}
				else
				{
					foundStart = true;
					matchTime = end.getMobileData().getDate().getTime();
				}
			}
			else if( er.getMobileData().getDate().getTime() >= matchTime )
			{
				//Log.d( TAG, "adding tailing item" );
				list.add( er );
			}
		}

		//Log.d( TAG, "Number of items: " + list.size() );
		EventStats stats = new EventStats();
		stats.startTime = start.getMobileData().getDate();
		stats.endTime   = end.getMobileData().getDate();
		stats.numEvents = list.size();
		stats.duration  = DmtUtils.getDuration( 
				list.get(0).getMobileData().getDate(), 
				( start.getMobileData().getEndTime() == null ?
						start.getMobileData().getDate() :
						start.getMobileData().getEndTime() )
		);
		stats.duration += DmtUtils.getDuration( 
				end.getMobileData().getDate(), 
				( list.get(list.size()-1).getMobileData().getEndTime() == null ?
						list.get(list.size()-1).getMobileData().getDate() :
							list.get(list.size()-1).getMobileData().getEndTime() )
		);
		
		if( tripData != null )
		{
			stats.pctOfActivity = stats.numEvents * 1.0 / tripData.numEvents;
			stats.pctOfTotalTime     = stats.duration * 1.0 / totalTime;
			stats.pctOfEventTime     = stats.duration * 1.0 / tripData.duration;
		}

		printStats( "GAP", stats );
		Log.i( TAG, "% Total Time: " + String.format( "%.2f", stats.pctOfTotalTime * 100 ) );
		Log.i( TAG, "% Event Time: " + String.format( "%.2f", stats.pctOfEventTime * 100 ) );
		Log.i( TAG, " % of Events: " + String.format( "%.2f", stats.pctOfActivity * 100 ) );
		mapData.put( "GAP", stats );
		return stats;
	}
}
