package com.tripgis.dmt.types;

import java.io.IOException;
import java.io.Serializable;
import java.util.ArrayList;

import com.tripgis.dmt.DmtDataType;
import com.tripgis.dmt.file.FileManager;
import com.tripgis.dmt.util.LatLng;
import com.tripgis.dmt.util.LogTranslator;
import com.tripgis.dmt.util.DmtUtils;
import com.tripgis.shared.types.LocationRecord;


/**
 * References:
 *    https://stackoverflow.com/questions/5950557/good-examples-using-java-util-logging
 *    
 * @author pk
 *
 */
public class EventRecord implements Serializable, DmtDataType<EventRecord>
{
	/**
	 * 
	 */
	private static final long serialVersionUID = 4015415712051504266L;
	private static final String TAG = EventRecord.class.getSimpleName();
	private static final LogTranslator Log = new LogTranslator( TAG );
	
	public static final EventRecord EMPTY = new EventRecord( "" );
	public static final EventRecord NEW = new EventRecord( "NEW" );
	public static final ArrayList<EventRecord> EMPTY_LIST = new ArrayList<>();

	public static final int IDX_EVENT_ID      = 0;
	public static final int IDX_TRIP_ID       = 1;
	public static final int IDX_DATE          = 2;
	public static final int IDX_LATITUDE      = 3;
	public static final int IDX_LONGITUDE     = 4;
	public static final int IDX_LOC_NAME      = 5;
	public static final int IDX_END_TIME      = 6;
	public static final int IDX_PLACE_ID      = 7;
	public static final int IDX_STORE         = 8;
	public static final int IDX_ADDRESS       = 9;
	public static final int IDX_CITY          = 10;
	public static final int IDX_STATE         = 11;
	public static final int IDX_ZIP           = 12;
	public static final int IDX_COUNTRY       = 13;
	public static final int IDX_COUNTY        = 14;
	public static final int IDX_OFFSET        = 15;
	public static final int IDX_LOCATION_TYPE = 16;
	public static final int IDX_GEO_LATITUDE  = 17;
	public static final int IDX_GEO_LONGITUDE = 18;
	public static final int IDX_RATING        = 19;
	public static final int IDX_DESCRIPTION   = 20;
	public static final int IDX_NOTE          = 21;
	public static final int IDX_EVENT_TYPE    = 22;
	public static final int IDX_EVENT_SUBTYPE = 23;
	public static final int IDX_GEOMETRY      = 24;
	public static final int IDX_DURATION      = 25;
	
    // Number of fields expected in a given record
    public static final int REQUIRED_FIELDS = 6 +
    		LocationRecord.REQUIRED_FIELDS + GeocodeRecord.REQUIRED_FIELDS + EventSubType.REQUIRED_FIELDS;

    public static final int MAX_DISPLAY_LEN = 20;

    protected int eventId;
	protected int tripId;
	protected LocationRecord mobileData;
	protected String description;
	protected String notes;
	protected GeocodeRecord geoData;
	protected EventSubType eventSubType;
	protected String geom;
	protected double duration;
	
	protected EventRecord( String text ) 
	{
		mobileData    = LocationRecord.EMPTY.copy();
		mobileData.setName( text );
		geoData      = GeocodeRecord.EMPTY.copy();
		eventSubType = EventSubType.EMPTY;
		description  = "";
		notes        = "";
		geom         = "";
	}

	public EventRecord( int eId, int tId )
	{
		this( "" );
		eventId      = eId;
		tripId       = tId;
	}
	
	public EventRecord( LocationRecord lr )
	{
		this( lr.getName() );
		mobileData = lr.copy();
	}
	
	public EventRecord copy()
	{
		EventRecord er  = new EventRecord( mobileData.getName() );
		er.eventId      = eventId;
		er.tripId		= tripId;
		er.mobileData 	= mobileData.copy();
		er.geoData      = geoData.copy();
		er.description  = description;
		er.notes        = notes;
		er.eventSubType = eventSubType.copy();
		er.geom         = geom;
		er.duration     = duration;
		return er;
	}
	
	public boolean isLocationModified( LocationRecord lr )
	{
		if( !mobileData.getName().equals( lr.getName() ) ||
			!mobileData.getEndTime().equals( lr.getEndTime() ) )
			return true;
		// FIXME: How to compare position (if needed)?
		return false;
	}
	
	public static EventRecord find( ArrayList<EventRecord> list, LocationRecord lr )
	{
		Log.t( TAG, "find( " + list.size() + ", " + lr + " )" );
		for( EventRecord er : list )
			if( er.matches( lr ) )
				return er;
		
		Log.d( TAG, "Location " + lr + " not found in list!" );
		return EMPTY;
	}
	
	public static ArrayList<EventRecord> createEventRecords( ArrayList<LocationRecord> list )
	{
		Log.t( TAG, "createEventRecords( " + list.size() + " )" );
		ArrayList<EventRecord> newList = new ArrayList<>();
		for( LocationRecord location : list )
			newList.add( new EventRecord( location ) );

		Log.d( TAG, "Returning " + list.size() + " event records." );
		return newList;
	}
	
	public int getEventId()               { return eventId;      }
	public int getTripId()                { return tripId;       }
	public LocationRecord getMobileData() { return mobileData;   }
	public String getDescription()        { return description;  }
	public String getNotes()              { return notes;        }
	public GeocodeRecord getGeoData()     { return geoData;      }
	public EventSubType getEventSubType() { return eventSubType; }
	public String getGeometry()           { return geom;         }
	public double getDuration()           { return duration;     }

	public void setEventId(      int eventId               ) { this.eventId = eventId;           }
	public void setTripId(       int tripId                ) { this.tripId = tripId;             }
	public void setMobileData(   LocationRecord mobilData  ) { this.mobileData = mobilData;      }
	public void setDescription(  String description        ) { this.description = description;   }
	public void setNotes(        String notes              ) { this.notes = notes;               }
	public void setGeoData(      GeocodeRecord geoData     ) { this.geoData = geoData;           }
	public void setEventSubType( EventSubType eventSubType ) { this.eventSubType = eventSubType; }
	public void setGeometry(     String geom               ) { this.geom = geom;                 }
	public void setDuration(     double duration           ) { this.duration = duration;         }

	public static EventRecord buildFromCsvArray( String[] data ) throws IOException
	{
		Log.t( TAG, "buildFromCsvArray( " + data.length + " )" );

        // Verify the string array passed in has the correct number of fields
        if( data.length != REQUIRED_FIELDS )
            throw new IOException( "Invalid number of parameters (" + data.length + " vs. " + REQUIRED_FIELDS + ")" );

        Log.d( TAG, "Creating record from " + "'" + data[0] + "', (" + data.length + " items )..." );
        
        int lrIdxBase = 2;
        int grIdxBase = lrIdxBase + LocationRecord.REQUIRED_FIELDS;
        int dIdxBase = grIdxBase + GeocodeRecord.REQUIRED_FIELDS;
        int estIdxBase = dIdxBase + 2;
        int sIdxBase = estIdxBase + EventSubType.REQUIRED_FIELDS;

        String[] lrArray = new String[LocationRecord.REQUIRED_FIELDS];
        for( int i = 0; i < lrArray.length; i++ )
        	lrArray[i] = data[lrIdxBase+i];
        
        String[] grArray = new String[GeocodeRecord.REQUIRED_FIELDS];
        for( int i = 0; i < grArray.length; i++ )
        	grArray[i] = data[grIdxBase+i];

        String[] estArray = new String[EventSubType.REQUIRED_FIELDS];
        for( int i = 0; i < estArray.length; i++ )
        	estArray[i] = data[estIdxBase+i];

        LocationRecord lr = LocationRecord.buildFromCsvArray( lrArray );
        EventRecord er = new EventRecord( 
        		Integer.parseInt( data[0] ), 
        		Integer.parseInt( data[1] ) );
        er.setMobileData( lr );
        er.setGeoData( GeocodeRecord.buildFromCsvArray( grArray ) );
        er.setDescription( data[dIdxBase] );
        er.setNotes( data[dIdxBase+1] );
        er.setEventSubType( EventSubType.buildFromCsvArray( estArray ) );
        er.setGeometry( data[sIdxBase] );
        
        if( !data[sIdxBase+1].isEmpty() )
        	er.setDuration( Double.parseDouble( data[sIdxBase+1] ) * 60 );
        
        Log.d( TAG, "EventRecord: " + er );
        return er;
	}

	public String getDisplayString()
	{
		String name = FileManager.readString( description );
		if( name.isEmpty() )
			name = FileManager.readString( mobileData.getName() );

		int end = name.length() > MAX_DISPLAY_LEN ? MAX_DISPLAY_LEN : name.length();
		return name.substring( 0, end );
	}

	public boolean matches( LocationRecord lr )
	{
		if( getMobileData().getDate() == lr.getDate() )
			return true;
		// FIXME: is this good enough?
		return false;
	}
	
	@Override
	public boolean matches( EventRecord t ) 
	{
		if( eventId != 0 )
			return eventId == t.eventId;
		else if( mobileData.getLocation() != LatLng.EMPTY )
			return mobileData.getLocation() == mobileData.getLocation();
		
		return mobileData.getDate() == t.getMobileData().getDate();
	}
	
	public boolean isNew()
	{
		return mobileData.getName().matches( DmtUtils.NEW_PREFIX );
	}
	
	public boolean isUpdated()
	{
		return mobileData.getName().matches( DmtUtils.UPDATE_PREFIX );
	}

	@Override
	public String generateCsvString()
	{
//		Log.t( TAG, "generateCsvString()" );
		StringBuffer sb = new StringBuffer();
		sb.append( eventId + FileManager.CSV_SEPARATOR );
		sb.append( tripId + FileManager.CSV_SEPARATOR );
		sb.append( mobileData.generateCsvString() + FileManager.CSV_SEPARATOR );
		sb.append( geoData.generateCsvString() + FileManager.CSV_SEPARATOR );
		sb.append( FileManager.prepareString( description ) + FileManager.CSV_SEPARATOR );
		sb.append( FileManager.prepareString( notes ) + FileManager.CSV_SEPARATOR );
		sb.append( eventSubType.generateCsvString() + FileManager.CSV_SEPARATOR );
		sb.append( FileManager.prepareString( geom ) );
		sb.append( duration );
		return sb.toString();
	}

	public String toString()
	{
		StringBuffer sb = new StringBuffer();
		sb.append( "EventId=" + eventId + ',' );
		sb.append( "TripId=" + tripId + ',' );
		sb.append( "MobilData=[" + mobileData.generateCsvString() + "]" + ',' );
		sb.append( "GeocodeData=[" + geoData.generateCsvString() + "]"  + ',' );
		sb.append( "Description=" + FileManager.prepareString( description ) + ',' );
		sb.append( "Notes=" + FileManager.prepareString( notes ) + ',' );
		sb.append( "EventSubType=[" + eventSubType.generateCsvString() + "]"  + ',' );
		sb.append( "Geometry=" + FileManager.prepareString( geom ) );
		sb.append( "Duration=" + duration );
		return sb.toString();
	}
}
