package com.tripgis.dmt.types;

import java.io.IOException;
import java.io.Serializable;

import com.tripgis.dmt.DmtDataType;
import com.tripgis.dmt.file.FileManager;
import com.tripgis.dmt.util.LatLng;
import com.tripgis.dmt.util.LogTranslator;
import com.tripgis.dmt.util.DmtUtils;

public class GeocodeRecord implements Serializable, DmtDataType<GeocodeRecord>
{
	/**
	 * 
	 */
	private static final long serialVersionUID = 3130336617205342472L;

    // Name to use in log lines
    public static final String TAG = GeocodeRecord.class.getSimpleName();
    public static final LogTranslator Log = new LogTranslator( TAG );

	// Number of fields expected in a given record
    public static final int MINIMUM_FIELDS = 8;
    public static final int REQUIRED_FIELDS = 13; // Actually 12

	public static GeocodeRecord EMPTY = new GeocodeRecord();

	private String placeId;
	private String store;
	private String address;
	private String city;
	private String state;
	private String zip;
	private String country;
	private String county;
	private String utcOffset;
	private String locationType;
	private LatLng geoLocation;
	private String rating;
	
	private GeocodeRecord()
	{
		Log.t( TAG, "GeocodeRecord()" );
		placeId      = "";
		store        = "";
		address      = "";
		city         = "";
		state        = "";
		zip          = "";
		country      = "";
		county       = "";
		utcOffset    = "";
		locationType = "";
		geoLocation  = LatLng.EMPTY;
		rating       = "";
	}
	
	public GeocodeRecord( String pid, String l )
	{
		this();
		Log.t( TAG, "GeocodeRecord( " + pid + ", " + l + " )" );
		placeId      = pid;
		locationType = l;
		rating       = "";
	}
	
	public GeocodeRecord copy()
	{
		Log.t( TAG, "copy() = " + this );
		GeocodeRecord gr = new GeocodeRecord();
		gr.placeId      = placeId;
		gr.store        = store;
		gr.address      = address;
		gr.city         = city;
		gr.state        = state;
		gr.zip          = zip;
		gr.country      = country;
		gr.county       = county;
		gr.utcOffset	= utcOffset;
		gr.locationType = locationType;
		gr.geoLocation  = geoLocation;
		gr.rating       = rating;
		return gr;
	}

	public String getPlaceId()      { return placeId;      }
	public String getStore()        { return store;        }
	public String getAddress()      { return address;      }
	public String getCity()         { return city;         }
	public String getState()        { return state;        }
	public String getZip()          { return zip;          }
	public String getCountry()      { return country;      }
	public String getCounty()       { return county;       }
	public String getUtcOffset()	{ return utcOffset;    }
	public String getLocationType() { return locationType; }
	public LatLng getGeoLocation()  { return geoLocation;  }
	public String getRating()       { return rating;       }

	public void setPlaceId( String placeId )           { this.placeId      = DmtUtils.storeString( placeId, 32 );      }
	public void setStore( String store )               { this.store        = DmtUtils.storeString( store, 32 );        }
	public void setAddress( String address )           { this.address      = DmtUtils.storeString( address, 32 );      }
	public void setCity( String city )                 { this.city         = DmtUtils.storeString( city, 32 );         }
	public void setState( String state )               { this.state        = DmtUtils.storeString( state, 32 );        }
	public void setZip( String zip )                   { this.zip          = DmtUtils.storeString( zip, 8 );           }
	public void setCountry( String country )           { this.country      = DmtUtils.storeString( country, 32 );      }
	public void setCounty( String county )             { this.county 	   = DmtUtils.storeString( county, 32 );       }
	public void setUtcOffset( String utc )             { this.utcOffset    = DmtUtils.storeString( utc, 8 );           }
	public void setLocationType( String locationType ) { this.locationType = DmtUtils.storeString( locationType, 32 ); }
	public void setGeoLocation( LatLng ll )  		   { this.geoLocation  = ll;		                                }
	public void setRating( String rating )             { this.rating       = DmtUtils.storeString( rating, 4 );        }

	public void setGeoLocation( String geom )		   
	{ 
		if( !geom.isEmpty() )
			this.geoLocation  = new LatLng( geom );
		else this.geoLocation = LatLng.EMPTY;
	}
	
	// https://stackoverflow.com/questions/3654446/java-regex-help-splitting-string-on-spaces-and-commas
	public static GeocodeRecord parseAddress( String text )
	{
		Log.t( TAG, "parseAddress( " + text + " )" );
		String[] parts = text.split( "\\s*,\\s*" );
		int idx = 0;
		GeocodeRecord gr = new GeocodeRecord();
		if( parts.length > 4 )	// SOMETHING MORE, 4505 Plaza Dr, Irving, TX 75063, USA
		{
			String store = parts[idx];
			while( idx++ < parts.length - 3 )
				store += parts[idx];
			Log.d( TAG, "Store=" + store );
			gr.setStore( store );
		}

		if( parts.length - idx > 0 ) // 4505 Plaza Dr, Irving, TX 75063, USA
		{
			gr.setAddress( parts[idx++] );
			Log.d( TAG, "Address=" + gr.getAddress() );
		}

		if( parts.length - idx > 0 ) // Irving, TX 75063, USA
		{
			gr.setCity( parts[idx++] );
			Log.d( TAG, "City=" + gr.getCity() );
		}

		// TX 75063, USA -- OR -- TX 75063
		if( parts.length - idx > 0 )
		{
			String[] subparts = parts[idx].split( "\\s+" );
			Log.d( TAG, "State=" + subparts[0] + ", Zip=" + subparts[1] );
			gr.setState( subparts[0] );
			gr.setZip( subparts[1] );
			idx++;
		}

		if( parts.length - idx > 0 ) // USA
		{
			Log.d( TAG, "Country=" + parts[idx] );
			gr.setCountry( parts[idx] );
		}

		Log.d( TAG, "Returning " + gr );
		return gr;
	}

	public String getFormattedAddress()
	{
		StringBuffer sb = new StringBuffer();
		if( !store.isEmpty() )
			sb.append( store + ", " );
		if( !address.isEmpty() )
			sb.append( address + ", " );
		if( !city.isEmpty() )
			sb.append( city + ", " );
		if( !state.isEmpty() )
			sb.append( state + ", " );
		if( !zip.isEmpty() )
			sb.append( zip + ", " );
		String str = sb.toString();
		str.replace( ",\\s$", "" );
		return str;
	}

	public static GeocodeRecord buildFromCsvArray( String[] data ) throws IOException
	{
		Log.t( TAG, "buildFromCsvArray( " + data.length + " )" );
		// TODO - read in rating/geom?

        // Verify the string array passed in has the correct number of fields
        if( data.length != REQUIRED_FIELDS && data.length != MINIMUM_FIELDS )
            throw new IOException( "Invalid number of parameters (" + data.length + " vs. " + REQUIRED_FIELDS + ")" );

        Log.d( TAG, "Creating record from " + "'" + data[0] + "', (" + data.length + " items )..." );
    	GeocodeRecord gr = new GeocodeRecord( data[0], data[9] );
    	
    	gr.setStore( data[1] );
    	gr.setAddress( data[2] );
    	gr.setCity( data[3] );
    	gr.setState( data[4] );
    	gr.setZip( data[5] );
    	gr.setCountry( data[6] );
    	gr.setCounty( data[7] );
    	gr.setUtcOffset( data[8] );
    	
    	if( !( data[10].isEmpty() || data[11].isEmpty() ) )
    	{
	    	gr.setGeoLocation( new LatLng( 
	    			Double.parseDouble( data[10] ), 
	    			Double.parseDouble( data[11] ) )
	    	);
    	}
    	
    	gr.setRating( data[12] );
    	return gr;
	}

	@Override
	public boolean matches( GeocodeRecord t )
	{
		return t.placeId.equals( placeId );
	}

	@Override
	public String generateCsvString() 
	{
		StringBuffer sb = new StringBuffer();
		sb.append( FileManager.prepareString( store ) + FileManager.CSV_SEPARATOR );
		sb.append( FileManager.prepareString( placeId ) + FileManager.CSV_SEPARATOR );
		sb.append( FileManager.prepareString( address ) + FileManager.CSV_SEPARATOR );
		sb.append( FileManager.prepareString( city ) + FileManager.CSV_SEPARATOR );
		sb.append( FileManager.prepareString( state ) + FileManager.CSV_SEPARATOR );
		sb.append( FileManager.prepareString( zip ) + FileManager.CSV_SEPARATOR );
		sb.append( FileManager.prepareString( country ) + FileManager.CSV_SEPARATOR );
		sb.append( FileManager.prepareString( county ) + FileManager.CSV_SEPARATOR );
		sb.append( utcOffset + FileManager.CSV_SEPARATOR );
		sb.append( FileManager.prepareString( locationType ) + FileManager.CSV_SEPARATOR );
		sb.append( geoLocation.getDisplayString() + FileManager.CSV_SEPARATOR );
		sb.append( rating );
		return sb.toString();
	}
	
	public String getDisplayString() 
	{
		StringBuffer sb = new StringBuffer();
		if( !store.isEmpty() )
			sb.append( store + ", " );
		sb.append( address + ", " );
		
		if( !city.isEmpty() )
			sb.append( city + ", " + state + ", " + zip + ", " + country + ", " );

		if( !country.isEmpty() )
			sb.append( "[" + county + ", " + locationType + "]" );
		else
			sb.append( "[" + locationType + "]" );
		
		return sb.toString();
	}

	public String toString()
	{
		StringBuffer sb = new StringBuffer();
		sb.append( "PlaceId=" + placeId + "," );
		sb.append( "Store=" + store + "," );
		sb.append( "Address=" + address + "," );
		sb.append( "City=" + city + "," );
		sb.append( "State=" + state + "," );
		sb.append( "Zip=" + zip + "," );
		sb.append( "Country=" + country + "," );
		sb.append( "County=" + county + "," );
		sb.append( "UtcOffset=" + utcOffset + "," );
		sb.append( "Type=" + locationType + "," );
		sb.append( "GeoLocation=[" + geoLocation + "]," );
		sb.append( "Rating=" + rating );
		return sb.toString();
	}
	
	//TODO aspects -> [ rating, type:overall ]
	// photos -> [ height, html_attributions, photo_reference, width ]
	//		look into place photos documentation [place details]
	// icon
	// price_level [0 free]
	// opening_hours -> [open_now:boolean]
	// rankby=distance param might be good
	// other restrictions: opennow, type, min/maxprice
}
