package com.tripgis.dmt.types;

import java.io.IOException;
import java.io.Serializable;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.HashMap;

import com.tripgis.dmt.DmtDataType;
import com.tripgis.dmt.file.FileManager;
import com.tripgis.dmt.util.DmtUtils;
import com.tripgis.dmt.util.LogTranslator;
import com.tripgis.shared.types.LocationRecord;

/**
 * References:
 *    https://stackoverflow.com/questions/5950557/good-examples-using-java-util-logging
 *    
 * @author pk
 *
 */
public class LegacyRecord extends EventRecord implements Serializable, DmtDataType<EventRecord>
{
	/**
	 * 
	 */
	private static final long serialVersionUID = 4015415712051504266L;
	private static final String TAG = LegacyRecord.class.getSimpleName();
	private static final LogTranslator Log = new LogTranslator( TAG );
	
	private LegacyRecord( String text ) 
	{
		super( text );
	}

	public LegacyRecord( int eId, int tId )
	{
		super( eId, tId );
	}
	
	public LegacyRecord( EventRecord er )
	{
		super( "" );
		Log.t( TAG, "LegacyRecord( " + er + " )" );
		eventId      = er.eventId;
		tripId		 = er.tripId;
		mobileData 	 = er.mobileData.copy();
		geoData      = er.geoData.copy();
		description  = er.description;
		notes        = er.notes;
		eventSubType = er.eventSubType.copy();
		geom         = er.geom;
		duration     = er.duration;
	}
	
	public static ArrayList<EventRecord> legacyToEventRecords( ArrayList<LegacyRecord> list )
	{
		Log.t( TAG, "createEventRecords( " + list.size() + " )" );
		ArrayList<EventRecord> newList = new ArrayList<>();
		for( LegacyRecord record : list )
			newList.add( (EventRecord) record );

		Log.d( TAG, "Returning " + list.size() + " event records." );
		return newList;
	}
	
	public static HashMap<String,Integer> buildHeaderFromCsvArray( String[] data ) throws IOException
	{
		if( data == null ) throw new IOException( "Header line empty" );
		Log.t( TAG, "buildFromCsvArray( " + data.length + " )" );
		HashMap<String,Integer> map = new HashMap<>();
		for( int i = 0; i < data.length; i++ )
		{
			String field = data[i].toLowerCase();
			Log.d( TAG, "Adding Field " + field + ": " + i );
			map.put( field.toLowerCase(), i );
		}
		
		Log.d( TAG, "Number of fields: " + map.size() );
		return map;
	}

	public static LegacyRecord buildFromCsvArray( HashMap<String,Integer> map, String[] data ) throws IOException
	{
		Log.t( TAG, "buildFromCsvArray( " + data.length + " )" );
		
		String[] eventData = new String[REQUIRED_FIELDS];
		for( int i = 0; i < eventData.length; i++ )
			eventData[i] = "";
		
		eventData[IDX_EVENT_ID]  = "0";
		eventData[IDX_TRIP_ID]   = "0";
		eventData[IDX_LATITUDE]  = "0";
		eventData[IDX_LONGITUDE] = "0";

		String[] requiredFields = new String[] { "date", "start" };
		String missingFields = "";
		for( String key : requiredFields )
		{
			if( map.get( key ) == null )
				missingFields += key + ", ";
		}
		
		if( !missingFields.isEmpty() )
		{
			missingFields.replace( ",\\s$", "" );
			throw new IOException( "Input file missing required field(s) - " + missingFields );
		}
		
		int date_idx = map.get( "date" );
		if( date_idx != 0 )
			throw new IOException( "First column must be date field" );
		
		int start_idx = map.get( "start" );
		int end_idx = map.get( "end" );
		int zone_idx = map.get( "zone" );
		int location_idx = map.get( "location" );
		int store_idx = map.get( "store" );
		int address_idx = map.get( "address" );
		int city_idx     = map.get( "city" );
		int state_idx    = map.get( "state" );
		int event_idx    = map.get( "event" );
		int subtype_idx  = map.get( "subtype" );
		int note_idx     = map.get( "notes" );
		
		if( start_idx != 0 && zone_idx != 0 )
			eventData[IDX_DATE] = data[date_idx] + " " + data[start_idx] + " " + data[zone_idx];

		else if( start_idx != 0 )
			eventData[IDX_DATE] = data[date_idx] + " " + data[start_idx];

		try 
		{
			eventData[IDX_DATE] = LocationRecord.getDateString(
					LocationRecord.parseDate( eventData[IDX_DATE] ) );
		}
		catch (ParseException e) 
		{
			throw new IOException( "Failed to parse date '" + eventData[IDX_DATE] + "'" );
		}
		
		if( location_idx != 0 && note_idx != 0 )
			eventData[IDX_LOC_NAME] = data[location_idx] + " " + data[note_idx];
		else if( location_idx != 0 )
			eventData[IDX_LOC_NAME] = data[location_idx];
		else if( note_idx != 0 )
			eventData[IDX_LOC_NAME] = data[note_idx];

		if( !data[end_idx].isEmpty() )
		{
			if( end_idx != 0 && zone_idx != 0 )
				eventData[IDX_END_TIME] = data[date_idx] + " " + data[end_idx] + " " + data[zone_idx];
			else if( end_idx != 0 )
				eventData[IDX_END_TIME] = data[date_idx] + " " + data[end_idx];
		}

		if( location_idx != 0 && store_idx != 0 )
			eventData[IDX_STORE] = data[location_idx] + " " + data[store_idx];
		else if( location_idx != 0 )
			eventData[IDX_STORE] = data[location_idx];
		else if( store_idx != 0 )
			eventData[IDX_STORE] = data[store_idx];

		if( address_idx != 0 )
			eventData[IDX_ADDRESS] = data[address_idx];

		if( city_idx != 0 )
			eventData[IDX_CITY] = data[city_idx];

		if( state_idx != 0 )
		{
			if( data[state_idx].length() > 2 )
				eventData[IDX_COUNTRY] = data[state_idx];
			else
				eventData[IDX_STATE] = data[state_idx];
		}
		
		if( event_idx != 0 )
			eventData[IDX_EVENT_TYPE] = data[event_idx];

		if( subtype_idx != 0 )
			eventData[IDX_EVENT_SUBTYPE] = data[subtype_idx];
		
		if( note_idx != 0 )
			eventData[IDX_NOTE] = data[note_idx];
        
		// TODO - add mileage, trip, driver

		return new LegacyRecord( EventRecord.buildFromCsvArray( eventData ) );
	}

	public String generateCsvString( HashMap<String,Integer> indices )
	{
//		Log.t( TAG, "generateCsvString()" );
		StringBuffer sb = new StringBuffer();
		
		for( String key : indices.keySet() )
		{
			switch( key )
			{
				case "date":
					sb.append( DmtUtils.getDateString( getMobileData().getDate() ) + FileManager.CSV_SEPARATOR );
					break;
				case "start":
					sb.append( DmtUtils.getTimeString( getMobileData().getDate() ) + FileManager.CSV_SEPARATOR );
					break;
				case "end":
					sb.append( DmtUtils.getTimeString( getMobileData().getEndTime() ) + FileManager.CSV_SEPARATOR );
					break;
				case "zone":
					sb.append( DmtUtils.getTimeZone( getMobileData().getDate() ) + FileManager.CSV_SEPARATOR );
					break;
				case "location":
					sb.append( FileManager.prepareString( getGeoData().getStore() ) + FileManager.CSV_SEPARATOR );
					break;
				case "store":
					sb.append( FileManager.CSV_SEPARATOR );
					break;
				case "address":
					sb.append( FileManager.prepareString( getGeoData().getAddress() ) + FileManager.CSV_SEPARATOR );
					break;
				case "city":
					sb.append( FileManager.prepareString( getGeoData().getCity() ) + FileManager.CSV_SEPARATOR );
					break;
				case "state":
					sb.append( FileManager.prepareString( getGeoData().getState() ) + FileManager.CSV_SEPARATOR );
					break;
				case "event": 
					sb.append( FileManager.prepareString( eventSubType.getType().getName() ) + FileManager.CSV_SEPARATOR );
					break;
				case "subtype":
					sb.append( FileManager.prepareString( eventSubType.getName() ) + FileManager.CSV_SEPARATOR );
					break;
				case "notes":
					sb.append( FileManager.prepareString( "[" + description + "]" + notes ) + FileManager.CSV_SEPARATOR );
					break;
			};
		}
		
		sb.replace( sb.lastIndexOf( FileManager.CSV_SEPARATOR ), 1, "" );
		return sb.toString();
	}
}
