package com.tripgis.dmt.types;

import java.io.IOException;
import java.io.Serializable;
import java.util.ArrayList;

import com.tripgis.dmt.DmtDataType;
import com.tripgis.dmt.DmtListType;
import com.tripgis.dmt.file.FileManager;
import com.tripgis.dmt.util.LogTranslator;
import com.tripgis.dmt.util.DmtUtils;

public class TripDbRecord implements Serializable, DmtDataType<TripDbRecord>, DmtListType<EventRecord>
{
	/**
	 * 
	 */
	private static final long serialVersionUID = 3577342098538860767L;

    // Name to use in log lines
    public static final String TAG = TripDbRecord.class.getSimpleName();
    public static final LogTranslator Log = new LogTranslator( TAG );

	public static final TripDbRecord EMPTY = new TripDbRecord( "" );
	public static final TripDbRecord NEW = new TripDbRecord( "NEW" );

    // Number of fields expected in a given record
    public static final int REQUIRED_FIELDS = 4;

    public static final String SOURCE_EVENT = "event";
    public static final String SOURCE_MOBILE = "mobile";
    public static final String SOURCE_LEGACY = "legacy";

    private int tripId;
	private String tripName;
	private String source;
	private String fileName;
	private ArrayList<EventRecord> eventList;
	
	private TripDbRecord( String n ) 
	{ 
		Log.t( TAG, "TripDbRecord( " + n + " )" );
		tripId            = 0; 
		tripName          = n;
		source            = SOURCE_EVENT;
		fileName          = n;
		eventList         = new ArrayList<>();
	}
	
	public TripDbRecord( int id, String n ) 
	{ 
		Log.t( TAG, "TripDbRecord( " + id + ", " + n + " )" );
		tripId    = id; 
		tripName  = n;
		source    = SOURCE_EVENT;
		fileName  = n;
		eventList = new ArrayList<>();
	}
	
	/**
	 * References:
	 *    http://beginnersbook.com/2013/12/how-to-clone-an-arraylist-to-another-arraylist/
	 */
	@SuppressWarnings("unchecked")
	public TripDbRecord copy()
	{
		Log.t( TAG, "copy()" );
		TripDbRecord tdr      = new TripDbRecord( tripName );
		tdr.tripId            = tripId;
		tdr.source            = source;
		tdr.fileName          = fileName;
		tdr.eventList         = (ArrayList<EventRecord>) eventList.clone();
		return tdr;
	}

	public void setTripId( int id ) { tripId = id; }
	public void setTripName( String n ) { tripName = n; }
	public void setSource( String s ) { source = s; }
	public void setFileName( String s ) { fileName = s; }
	public void setList( ArrayList<EventRecord> list ) { eventList = list; }

	public int getTripId() { return tripId; }
	public String getTripName() { return tripName; }
	public String getSource() { return source; }
	public String getFileName() { return fileName; }
	public ArrayList<EventRecord> getEventList() { return eventList; }

	public boolean isDbDataPresent() { return tripId == 0; }
	
	public static TripDbRecord buildFromCsvArray( String[] data ) throws IOException
	{
		Log.t( TAG, "buildFromCsvArray( " + data.length + " )" );

        // Verify the string array passed in has the correct number of fields
        if( data.length != REQUIRED_FIELDS )
            throw new IOException( "Invalid number of parameters (" + data.length + " vs. " + REQUIRED_FIELDS + ")" );

        Log.d( TAG, "Creating record from " + "'" + data[0] + "', '" + data[1] + "'..." );
        TripDbRecord tdr = new TripDbRecord( Integer.parseInt( data[0] ), data[1] );
        tdr.setSource( data[2] );
        tdr.setFileName( data[3] );
        return tdr;
	}

	@Override
	public String generateCsvString()
	{
		Log.t( TAG, "generateCsvString()" );
		StringBuffer sb = new StringBuffer();
		sb.append( tripId );
		sb.append( FileManager.CSV_SEPARATOR );
		sb.append( FileManager.prepareString( tripName ) );
		sb.append( FileManager.CSV_SEPARATOR );
		sb.append( FileManager.prepareString( source ) );
		sb.append( FileManager.CSV_SEPARATOR );
		sb.append( FileManager.prepareString( fileName ) );
		return sb.toString();
	}

	@Override
	public boolean matches( TripDbRecord t )
	{
		if( tripId != 0 )
			return t.tripId == tripId;
		return t.tripName == tripName;
	}
	
	public boolean isChanged() { return isNew() || isUpdated(); }

	public boolean isNew()
	{
		return tripName.matches( DmtUtils.NEW_PREFIX );
	}
	
	public boolean isUpdated()
	{
		return tripName.matches( DmtUtils.UPDATE_PREFIX );
	}
	
	public String getDisplayString()
	{
		return tripName;
	}
	
	public String toString() 
	{
		StringBuffer sb = new StringBuffer();
		sb.append( "Id=" + tripId + "," );
		sb.append( "Name=" + tripName + "," );
		sb.append( "Source=" + source + "," );
		sb.append( "FileName=" + fileName + "," );
		sb.append( "NumEvents=" + eventList.size() );
		return sb.toString(); 
	}

	@Override
	public ArrayList<EventRecord> getList() { return eventList; }

	@Override
	public int getItemCount() { return eventList.size(); }

	@Override
	public EventRecord getFirstRecord() { return eventList.get( 0 ); }

	@Override
	public EventRecord getLastRecord() { return eventList.get( eventList.size() - 1 ); }
}
