package com.tripgis.dmt.util;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.TimeZone;

import com.tripgis.dmt.db.DbModelEvents;

public class DmtUtils 
{
	private static StringBuffer lastCommandOutput;
	
	public static final String NEW_PREFIX = "NEW: ";
	public static final String UPDATE_PREFIX = "UPD: ";

	public final static String jpeg = "jpeg";
	public final static String jpg  = "jpg";
	public final static String gif  = "gif";
	public final static String tiff = "tiff";
	public final static String tif  = "tif";
	public final static String png  = "png";

	public final static String sql  = "sql";
	public final static String xlsx = "xlsx";
	public final static String xls  = "xls";
	public final static String ods  = "ods";
	public final static String csv  = "csv";
	public final static String bmk  = "bmk";
	
	public static final String EST = "UTC-5:00";
	public static final String EDT = "UTC-4:00";
	public static final String CST = "UTC-6:00";
	public static final String CDT = "UTC-5:00";
	public static final String MST = "UTC-7:00";
	public static final String MDT = "UTC-6:00";
	public static final String PST = "UTC-8:00";
	public static final String PDT = "UTC-7:00";
	public static final String HST = "UTC-10:00";
	
	public static final int DAYS_BETWEEN_EPOCHS = 25567; // Days between 1/1/1900 and 1/1/1970
	
    // Name to use in log lines
    public static final String TAG = DmtUtils.class.getSimpleName();
    public static final LogTranslator Log = new LogTranslator( TAG );
    
	public static boolean executeCommand( String command )
	{
		lastCommandOutput = new StringBuffer();
		boolean retVal = false;
		
		//https://www.mkyong.com/java/how-to-execute-shell-command-from-java/
		try 
		{
			String cmdArray[] = { "/bin/sh", "-c", command };
			Process p = Runtime.getRuntime().exec( cmdArray );
			p.waitFor();
			BufferedReader reader = new BufferedReader(
					new InputStreamReader( p.getInputStream() ) );
			
			String line;
			while(( line = reader.readLine() ) != null )
				lastCommandOutput.append( line + "\n" );
			
			retVal = p.exitValue() == 0;
			if( !retVal )
				Log.w( TAG, "Command returned non-zero value - " + p.exitValue() );

		} 
		catch (IOException e) 
		{
			Log.e( TAG, "IO Exception - " + e.getMessage() );
			e.printStackTrace();
		} 
		catch (InterruptedException e) 
		{
			Log.e( TAG, "Interrupted Exception - " + e.getMessage() );
			e.printStackTrace();
		}

		if( !retVal )
			Log.d( TAG, "Command: " + command );

		return retVal;
	}
	
	public static String storeString( String text, int max )
	{
		if( text.length() > max )
			return text.substring( 0, max - 1 );
		return text;
	}
	
	public static StringBuffer getLastCommandOutput() { return lastCommandOutput; }
	
	public static String getExtension( File f )
	{
		String ext = null;
		String s = f.getName();
		int i = s.lastIndexOf( '.' );
		
		if( i > 0 && i < s.length() - 1 )
			ext = s.substring( i + 1 ).toLowerCase();
		
		return ext;
	}
	
	public static String removeExtension( File f )
	{
		int idx = f.getName().lastIndexOf( '.' );
		//ALT: name.replaceFirst("[.][^.]+$", "" );
		String name = f.getName().substring( 0, idx );
		return name;
	}
	
	public static String removeExtension( String s )
	{
		int idx = s.lastIndexOf( '.' );
		
		if( idx < 0 )	// Don't do a substring if there is no '.'
			return s;

		String name = s.substring( 0, idx );
		return name;
	}
	
	public static String setNewPrefix( String name ) { return setPrefix( NEW_PREFIX, name ); }
	public static String setUpdPrefix( String name ) { return setPrefix( UPDATE_PREFIX, name ); }
	public static String setPrefix( String pfx, String name ) 
	{ 
		String title = removeExtension( name );
		title.replaceFirst( "(" + NEW_PREFIX + "|" + UPDATE_PREFIX + ")", "" );
		return pfx + title;
	}

	public static boolean isSqlFile( File f )        { return getExtension( f ).equals( sql ); }
	public static boolean isExcelFile( File f )      { return getExtension( f ).equals( xls ); }
	public static boolean isExcel10File( File f )    { return getExtension( f ).equals( xlsx ); }
	public static boolean isOpenOfficeFile( File f ) { return getExtension( f ).equals( ods ); }
	
	public static Date getDate( long timeInSecs )
	{
		TimeZone.setDefault(TimeZone.getTimeZone("GMT-1"));
		Calendar c = new GregorianCalendar();
		c.setTimeInMillis( timeInSecs * 1000 );
		return c.getTime();
	}

	public static String getTimeString( double d )
	{
		//DateFormat timeFormat = new SimpleDateFormat( "HH:mm" );
		int hours = (int) (d / 3600.0);
		int mins = ( (int) (d) % 3600 / 3600 );
		if( hours == 0 || mins == 0 ) return "";
		
		return String.format( "%02d:%02d", hours, mins ); 
	}
	
	public static String getDateString( java.sql.Date date )
	{
		DateFormat dateFormat = new SimpleDateFormat( "yyyy-MM-dd");
		return dateFormat.format( date );
	}
	
	public static String getTimeString( java.sql.Date date )
	{
		DateFormat dateFormat = new SimpleDateFormat( "hh:mm:ss" );
		return dateFormat.format( date );
	}

	public static String getTimestamp( java.sql.Date date )
	{
		DateFormat dateFormat = new SimpleDateFormat( "yyyy-MM-dd_hhmmss");
		return dateFormat.format( date );
	}
	
	public static String getTimeZone( java.sql.Date date )
	{
		DateFormat dateFormat = new SimpleDateFormat( "yyyy-MM-dd_hhmmss");
		return dateFormat.format( date );
	}
	
	public static double getDuration( java.sql.Date start, java.sql.Date end )
	{
		if( end == null )
			return DbModelEvents.DEFAULT_DURATION;
		
		double duration = end.getTime() - start.getTime();
		return duration /= 1000;
	}
	
	public static String getPrettyDuration( double duration )
	{
		long conv = (long) duration;
		return String.format( "%02d:%02d:%02d", conv / 3600, 
				( conv % 3600 ) / 60, conv % 60 );
	}
	
	public static String getDoubleString( Double d )
	{
		if( d == null || d == 0 ) return "";
		return Double.toString( d );
	}
	
	public static String getString( String text )
	{
		if( text == null ) return "";
		return text;
	}

	public static String toUrlString( String text )
	{
		return text.replaceAll( " ", "+" );
	}

    public static String prettySize( long bytes )
    {
        Log.d( TAG, "prettySize( " + bytes + " )" );
        if( bytes > Math.pow( 1024, 4 ) )
            return String.format( java.util.Locale.US, "%.2f Tb", bytes / Math.pow( 1024, 4 ) );
        else if( bytes > Math.pow( 1024, 3 ) )
            return String.format( java.util.Locale.US, "%.2f Gb", bytes / Math.pow( 1024, 3 ) );
        else if( bytes > Math.pow( 1024, 2 ) )
            return String.format( java.util.Locale.US, "%.2f Mb", bytes / Math.pow( 1024, 2 ) );
        else if( bytes > Math.pow( 1024, 1 ) )
            return String.format( java.util.Locale.US, "%.2f Kb", bytes / Math.pow( 1024, 1 ) );
        else
            return String.format( java.util.Locale.US, "%d bytes", bytes );
    }

	/**
	 * References:
	 *    https://stackoverflow.com/questions/367626/how-do-i-fix-the-expression-of-type-list-needs-unchecked-conversion
	 *    
	 * @param clazz
	 * @param c
	 * @return
	 */
	public static <T> ArrayList<T> castList( Class<? extends T> clazz, Collection<?> c ) 
	{
	    ArrayList<T> r = new ArrayList<T>(c.size());
	    for( Object o: c )
	      r.add( clazz.cast(o) );
	    return r;
	}
}
