package com.tripgis.shared.types;

import java.io.IOException;
import java.io.Serializable;
import java.sql.Date;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.TimeZone;

import com.tripgis.dmt.DmtDataType;
import com.tripgis.dmt.file.FileManager;
import com.tripgis.dmt.util.LatLng;
import com.tripgis.dmt.util.LogTranslator;

/**
 * LocationRecord is the data of import being collected by the Android App.
 *
 * Created by pk on 7/9/17.
 */

public class LocationRecord implements Serializable, DmtDataType<LocationRecord>
{
    /**
	 * 
	 */
	private static final long serialVersionUID = 4594231092988059651L;

	private Date date;          // Date/Time of entry
    private LatLng location;    // Lat/Long of position
    private String name;        // Title if provided
    private Date endTime;       // End time of activity

    // Name to use in log lines
    public static final String TAG = LocationRecord.class.getSimpleName();
    public static final LogTranslator Log = new LogTranslator( TAG );

    public static final Date EMPTY_DATE = new Date( 0 );
    public static final LocationRecord EMPTY = 
			new LocationRecord( EMPTY_DATE, LatLng.EMPTY, "", null );
    
    // Common date format used throughout tool
    public static final String DATE_FORMAT = "dd/MM/yyyy HH:mm:ss";
    public static final String DATE_FORMAT_2 = "yyyy-MM-dd HH:mm:ss z";
    public static final String DATE_FORMAT_3 = "MM/dd/yy HH:mm";

    // Only need one instance of the sdf object
    private static SimpleDateFormat _SDF_INSTANCE;
    private static SimpleDateFormat _SDF_INSTANCE_2;
    private static SimpleDateFormat _SDF_INSTANCE_3;

    // Number of fields expected in a given record
    public static final int REQUIRED_FIELDS = 5; // LatLng stored as two fields in text

    // Total number of fields in each record
    //   Used when writing records to pad each stored record with null fields
    public static final int TOTAL_NUM_FIELDS = REQUIRED_FIELDS + 0;

    //TODO: Add additional map_items as they become known
    
    private LocationRecord() {};

    // Constructor
    public LocationRecord( Date date, LatLng location, String name, Date endTime )
    {
        Log.d( TAG, "LocationRecord( " + date + ", " + location + ", " + name + ", " + endTime + " ) " );
        this.date     = date;
        this.location = location;
        this.name     = name;
        this.endTime  = endTime;
    }
    
    public LocationRecord copy()
    {
    	Log.t( TAG, "copy() = " + this );
    	LocationRecord lr = new LocationRecord();
    	lr.date     = date;
    	lr.location = location;
    	lr.name     = name;
    	lr.endTime  = endTime;
    	return lr;
    }

    private static SimpleDateFormat getSdfInstance()
    {
        if( _SDF_INSTANCE == null )
        {
            _SDF_INSTANCE = new SimpleDateFormat( DATE_FORMAT );
            _SDF_INSTANCE.setTimeZone( TimeZone.getTimeZone( "UTC" ) );
            _SDF_INSTANCE_2 = new SimpleDateFormat( DATE_FORMAT_2 );
            _SDF_INSTANCE_2.setTimeZone( TimeZone.getTimeZone( "UTC" ) );
            _SDF_INSTANCE_3 = new SimpleDateFormat( DATE_FORMAT_3 );
            _SDF_INSTANCE_3.setTimeZone( TimeZone.getTimeZone( "UTC" ) );
        }

        return _SDF_INSTANCE;
    }

    /**
     * References:
     *    https://stackoverflow.com/questions/2149680/regex-date-format-validation-on-java
     *    https://stackoverflow.com/questions/25767825/reading-system-time-in-cst-time-zone-using-java
     *    
     * @param text
     * @return
     * @throws ParseException
     */
    public static Date parseDate( String text ) throws ParseException
    {
    	try
    	{
	    	getSdfInstance();
	    	if( text.matches( "\\d{2}/\\d{2}/\\d{4}.+" ) )
	            return parseDate( _SDF_INSTANCE, text );
	
	    	else if( text.matches( "\\d{4}-\\d{2}-\\d{2}.+" ) )
	            return parseDate( _SDF_INSTANCE_2, text );
	
	    	else if( text.matches( "\\d{2}/\\d{2}/\\d{2}.+" ) )
	            return parseDate( _SDF_INSTANCE_3, text );
	        else
	            return new Date( Long.valueOf( text ) );
    	}
    	catch( ParseException e )
    	{
    		e.printStackTrace();
    		throw e;
    	}
    }
    
    private static Date parseDate( SimpleDateFormat sdf, String text ) throws ParseException
    {
    	String[] parts = text.split( " " );
    	TimeZone tz = sdf.getTimeZone();
    	if( parts.length == 3 )
    	{
    		sdf.setTimeZone( TimeZone.getTimeZone( parts[2] ) );
    		//text = parts[0] + " " + parts[1];
    	}
    	
    	Date d = new Date( sdf.parse( text ).getTime() );
    	sdf.setTimeZone( tz );
    	return d;
    }

    /**
     * Returns the current system time.
     *
     * References:
     *    https://stackoverflow.com/questions/6059207/how-can-i-set-a-calendar-with-utc-time
     *
     * @return a Date object or null if current time could not be obtained
     */
    public static Date getCurrentTime()
    {
        Calendar calendar = Calendar.getInstance( TimeZone.getTimeZone( "UTC" ) );
        return new Date( calendar.getTime().getTime() );

        //return new Date( System.currentTimeMillis() );
    }

    /**
     * Converts the date object to the standard date format (DATE_FORMAT).
     *
     * @return A string in DATE_FORMAT form.
     */
    public static String getDateString( Date date )
    {
    	getSdfInstance();
        if( date != null )
            return _SDF_INSTANCE.format( date );
        return "";
    }
    
    public static String getSqlDateString( Date date )
    {
    	getSdfInstance();
        if( date != null && date != EMPTY_DATE )
            return _SDF_INSTANCE_2.format( date );
        return "";
    }

    // Getter methods
    public Date getDate()       { return date;     }
    public LatLng getLocation() { return location; }
    public String getName()     { return name;     }
    public Date getEndTime()    { return endTime;  }

    // Setter methods
    public void setDate(     Date d   ) { date = d;     }
    public void setLocation( String s ) { location = new LatLng( s ); }
//    public void setLocation( LatLng l ) { location = l; }
    public void setName(     String n ) { name = n;     }
    public void setEndTime(  Date d   ) { endTime = d;  }

    public static LocationRecord buildFromCsvArray( String[] data ) throws IOException
    {
    	Log.t( TAG, "buildFromCsvArray( " + data.length + " )" );

    	// Verify the string array passed in has the correct number of fields
        if( data.length != REQUIRED_FIELDS )
        {
            if( data.length == 4 )
                Log.w( TAG, "String[] only has 4 items, adding endtime" );
            else
            {
                for( int i = 0; i < data.length; i++ )
                    Log.e( TAG, "String[" + i + "] = " + data[i] );
                throw new IOException( "Invalid number of parameters (" + data.length + " vs. " + REQUIRED_FIELDS + ")" );
            }
        }

        Log.d( TAG, "Creating record from " + "'" + data[0] +
                "', '" + data[1] + "','" + data[2] + "','" + data[3] + "'..." );

        try
        {
        	LocationRecord lr = new LocationRecord();
            // Convert each string as needed to fit the specified object
            lr.date = parseDate( data[0] );
            lr.location = new LatLng( Double.parseDouble( data[1] ), Double.parseDouble( data[2] ) );
            lr.name = data[3];

            if( data.length > 4 && !data[4].equals( "" ) )
                lr.endTime = parseDate( data[4] );
            
            return lr;
        }
        catch( ParseException e )
        {
            throw new IOException( "Failed to parse data - " + e );
        }
    }

    /**
     * Returns a CSV string representing the record
     *
     * @return CSV string
     */
    public String toDisplayString()
    {
        StringBuffer sb = new StringBuffer();
        sb.append( LocationRecord.getSqlDateString( date ) );
        sb.append( FileManager.CSV_SEPARATOR );
        sb.append( location.getDisplayString() );
        sb.append( FileManager.CSV_SEPARATOR );
        sb.append( FileManager.readString( getName() ) );
        if( getEndTime() != null )
        {
            sb.append( FileManager.CSV_SEPARATOR );
            sb.append( LocationRecord.getDateString( endTime ) );
        }
        return sb.toString();
    }

	@Override
	public boolean matches( LocationRecord t ) 
	{
		return t.date == date;
	}

	@Override
	public String generateCsvString() 
	{
        StringBuffer sb = new StringBuffer();
        sb.append( date.getTime() );
        sb.append( FileManager.CSV_SEPARATOR );
        sb.append( location.getDisplayString() );
        sb.append( FileManager.CSV_SEPARATOR );
        sb.append( FileManager.prepareString( name ) );
        sb.append(FileManager.CSV_SEPARATOR);
        if( endTime != null )
            sb.append( endTime.getTime() );
        return sb.toString();
	}
	
	public String toString()
    {
        StringBuffer sb = new StringBuffer();
        sb.append( "Date=" + getSqlDateString( date ) + "," );
        sb.append( "Location=[" + location + "]," );
        sb.append( "Name=" + name + "," );
        sb.append( "EndTime=" );
        sb.append( getSqlDateString( endTime ) );
        return sb.toString();
    }
}
